@@ -17,6 +17,13 @@
 # define CDECL extern "C"
 # define CDECL_BEGIN CDECL {
 # define CDECL_END   }
+template<typename T>
+static inline T replace_value(T &var,T new_value)
+{
+   T tmp=var;
+   var=new_value;
+   return tmp;
+}
 #else
 # define CDECL
 # define CDECL_BEGIN
@@ -32,8 +39,8 @@ CDECL const char *strerror(int);
 #endif
 
 #include <gettext.h>
-#define _(Text) gettext (Text)
-#define N_(Text) Text
+#define _(msgid) gettext (msgid)
+#define N_(msgid) msgid
 
 #ifndef PARAMS
 # if defined __STDC__ || defined __cplusplus
@@ -127,6 +134,7 @@ CDECL void *memmem(const void *haystack, size_t haystack_len,
 # define PRINTF_LIKE(n,m) __attribute__((format(__printf__,n,m)))
 #else
 # define PRINTF_LIKE(n,m)
+# define __attribute__(x)
 #endif
 
 #ifdef TRIO_REPLACE_STDIO
