@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: ConnectionSlot.h,v 1.4 2003/11/20 14:39:11 lav Exp $ */
+/* $Id: ConnectionSlot.h,v 1.7 2007/10/24 14:38:48 lav Exp $ */
 
 #ifndef CONNECTIONSLOT_H
 #define CONNECTIONSLOT_H
@@ -33,10 +33,9 @@ class ConnectionSlot : public KeyValueDB
    class SlotValue : public KeyValueDB::Pair
    {
    public:
-      FileAccess *session;
-      SlotValue(const char *n,FileAccess *s);
+      FileAccessRef session;
+      SlotValue(const char *n,const FileAccess *s);
       SlotValue(const char *n,const char *v);
-      ~SlotValue();
    };
 
    Pair *NewPair(const char *n,const char *v)
@@ -46,9 +45,9 @@ class ConnectionSlot : public KeyValueDB
 
 public:
    static ConnectionSlot::SlotValue *Find(const char *n);
-   static FileAccess *FindSession(const char *n);
-   static void Set(const char *n,FileAccess *s);
-   static void SetCwd(const char *n,const char *cwd);
+   static const FileAccess *FindSession(const char *n);
+   static void Set(const char *n,const FileAccess *s);
+   static void SetCwd(const char *n,const FileAccess::Path &cwd);
    static void Remove(const char *n);
    static char *Format();
    char *FormatThis();
