@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileAccess.h,v 1.93 2006/07/27 07:22:39 lav Exp $ */
+/* $Id: FileAccess.h,v 1.106 2008/02/17 08:19:23 lav Exp $ */
 
 #ifndef FILEACCESS_H
 #define FILEACCESS_H
@@ -35,18 +35,15 @@
 #else
 # include <poll.h>
 #endif
+#include <assert.h>
 
 #include "SMTask.h"
 #include "trio.h"
-#include "xmalloc.h"
+#include "xstring.h"
 #include "ResMgr.h"
 #include "FileSet.h"
-#include "LsCache.h"
+#include "ArgV.h"
 
-#define NO_SIZE	     ((off_t)-1L)
-#define NO_SIZE_YET  ((off_t)-2L)
-#define NO_DATE	     ((time_t)-1L)
-#define NO_DATE_YET  ((time_t)-2L)
 #define FILE_END     ((off_t)-1L)
 #define UNKNOWN_POS  ((off_t)-1L)
 
@@ -54,13 +51,13 @@ class ListInfo;
 class Glob;
 class NoGlob;
 class DirList;
-class ArgV;
+class FileAccessRef;
 
-class FileAccess : public SMTask
+class FileAccess : public SMTask, public ResClient
 {
    static bool class_inited;
 public:
-   static LsCache *cache;
+   static class LsCache *cache;
    enum open_mode
    {
       CLOSED,
@@ -95,23 +92,22 @@ public:
       void init();
    public:
       int   device_prefix_len;
-      char  *path;
+      xstring path;
       bool  is_file;
-      char  *url;
+      xstring url;
       Path() { init(); }
       Path(const Path *o) { init(); Set(o); }
       Path(const Path &o) { init(); Set(o); }
       Path(const char *new_path) { init(); Set(new_path); }
       Path(const char *new_path,bool new_is_file,const char *new_url=0,int new_device_prefix_len=0)
 	 { init(); Set(new_path,new_is_file,new_url,new_device_prefix_len); }
-      ~Path();
       void Set(const Path*);
       void Set(const Path &o) { Set(&o); }
       void Set(const char *new_path,bool new_is_file=false,const char *new_url=0,int device_prefix_len=0);
-      void SetURL(const char *u) { xfree(url); url=xstrdup(u); }
+      void SetURL(const char *u) { url.set(u); }
       void Change(const char *new_path,bool new_is_file=false,const char *new_path_enc=0,int device_prefix_len=0);
       void ExpandTilde(const Path &home);
-      static void Optimize(char *p,int dev_prefix=0);
+      static void Optimize(xstring& p,int dev_prefix=0);
       void Optimize() { Optimize(path,device_prefix_len); }
       const Path& operator=(const Path &o)
 	 {
@@ -124,32 +120,33 @@ public:
    };
 
 protected:
-   char	 *vproto;
-   char	 *hostname;
-   char	 *portname;
-   char  *user;
-   char  *pass;
+   xstring_c vproto;
+   xstring_c hostname;
+   xstring_c portname;
+   xstring_c user;
+   xstring_c pass;
    bool	 pass_open;
 
    const char *default_cwd;
    Path	 home;
    Path	 cwd;
    Path	 *new_cwd;
-   char  *file;
-   char	 *file_url;
-   char	 *file1;
+   xstring file;
+   xstring file_url;
+   xstring file1;
    int	 mode;
    off_t pos;
    off_t real_pos;
    off_t limit;
 
-   time_t *opt_date;
+   FileTimestamp *opt_date;
    off_t  *opt_size;
 
    static void NonBlock(int fd);
    static void CloseOnExec(int fd);
 
-   void  DebugPrint(const char *prefix,const char *str,int level=9);
+   static void Log2(int level,xstring& str);
+   static void Log3(int level,const char *prefix,const char *str);
 
    time_t try_time;
    int retries;
@@ -165,11 +162,10 @@ protected:
    void	 ExpandTildeInCWD();
    const char *ExpandTildeStatic(const char *s);
 
-   char *real_cwd;
+   xstring real_cwd;
    void set_real_cwd(const char *c)
       {
-	 xfree(real_cwd);
-	 real_cwd=xstrdup(c);
+	 real_cwd.set(c);
       }
    void set_home(const char *h)
       {
@@ -180,13 +176,10 @@ protected:
    off_t  entity_size; // size of file to be sent
    time_t entity_date; // date of file to be sent
 
-   char *closure;
+   xstring_c closure;
    const char *res_prefix;
-   ResValue Query(const char *name,const char *closure=0) const;
-   bool QueryBool(const char *name,const char *closure=0) const
-      {
-	 return Query(name,closure).to_bool();
-      }
+   const char *ResPrefix() const { return res_prefix?res_prefix:GetProto(); }
+   const char *ResClosure() const { return closure?closure.get():GetHostName(); }
 
    int chmod_mode;
    bool ascii;
@@ -199,14 +192,14 @@ protected:
    void ClearError();
    void SetError(int code,const char *mess=0);
    void Fatal(const char *mess);
-   char *error;
+   xstring error;
    int error_code;
-   char *location;
-   char *suggested_filename;
+   xstring_c location;
+   xstring_c suggested_filename;
    void SetSuggestedFileName(const char *fn);
 
-   char *entity_content_type;
-   char *entity_charset;
+   xstring_c entity_content_type;
+   xstring_c entity_charset;
 
    FileAccess *next;
    static FileAccess *chain;
@@ -215,16 +208,14 @@ protected:
 
    int device_prefix_len(const char *path);
 
-   virtual ~FileAccess();
-
 public:
    virtual const char *GetProto() const = 0; // http, ftp, file etc
    bool SameProtoAs(const FileAccess *fa) const { return !strcmp(GetProto(),fa->GetProto()); }
    virtual FileAccess *Clone() const = 0;
    virtual const char *ProtocolSubstitution(const char *host) { return 0; }
 
-   const char *GetVisualProto() const { return vproto?vproto:GetProto(); }
-   void SetVisualProto(const char *p) { xfree(vproto); vproto=xstrdup(p); }
+   const char *GetVisualProto() const { return vproto?vproto.get():GetProto(); }
+   void SetVisualProto(const char *p) { vproto.set(p); }
    const char  *GetHome() const { return home; }
    const char  *GetHostName() const { return hostname; }
    const char  *GetUser() const { return user; }
@@ -248,7 +239,7 @@ public:
    void SetLimit(off_t lim) { limit=lim; }
    void SetSize(off_t s) { entity_size=s; }
    void SetDate(time_t d) { entity_date=d; }
-   void WantDate(time_t *d) { opt_date=d; }
+   void WantDate(FileTimestamp *d) { opt_date=d; }
    void WantSize(off_t *s) { opt_size=s; }
    void AsciiTransfer() { ascii=true; }
    virtual void Close();
@@ -296,6 +287,7 @@ public:
    void Init();
    FileAccess() { Init(); }
    FileAccess(const FileAccess *);
+   virtual ~FileAccess();
 
    void	 DontSleep() { try_time=0; }
 
@@ -393,108 +385,13 @@ public:
 
    static void ClassInit();
    static void ClassCleanup();
+
+   static void LogError(int level,const char *fmt,...) PRINTF_LIKE(2,3);
+   static void LogNote(int level,const char *fmt,...) PRINTF_LIKE(2,3);
+   static void LogRecv(int level,const char *line);
+   static void LogSend(int level,const char *line);
 };
 
-class FileAccessOperation : public SMTask
-{
-protected:
-   bool done;
-   char *error_text;
-   void SetError(const char *);
-   void SetErrorCached(const char *);
-
-   bool use_cache;
-
-   virtual ~FileAccessOperation();
-
-public:
-   FileAccessOperation();
-
-   virtual int Do() = 0;
-   bool Done() { return done; }
-   bool Error() { return error_text!=0; }
-   const char *ErrorText() { return error_text; }
-
-   virtual const char *Status() = 0;
-
-   void UseCache(bool y=true) { use_cache=y; }
-};
-
-#include "FileSet.h"
-
-class PatternSet;
-
-class ListInfo : public FileAccessOperation
-{
-protected:
-   FileAccess *session;
-   FileAccess::Path saved_cwd;
-   FileSet *result;
-
-   const char *exclude_prefix;
-   PatternSet *exclude;
-
-   unsigned need;
-   bool follow_symlinks;
-
-   virtual ~ListInfo();
-
-public:
-   ListInfo(FileAccess *session,const char *path);
-
-   void SetExclude(const char *p,PatternSet *e);
-
-   // caller has to delete the resulting FileSet itself.
-   FileSet *GetResult() { return replace_value(result,(FileSet*)0); }
-
-   void Need(unsigned mask) { need|=mask; }
-   void NoNeed(unsigned mask) { need&=~mask; }
-   void FollowSymlinks() { follow_symlinks=true; }
-};
-
-#include "buffer.h"
-
-class LsOptions
-{
-public:
-   bool append_type:1;
-   bool multi_column:1;
-   bool show_all:1;
-   LsOptions()
-      {
-	 append_type=false;
-	 multi_column=false;
-	 show_all=false;
-      }
-};
-
-class DirList : public FileAccessOperation
-{
-protected:
-   Buffer *buf;
-   ArgV *args;
-   bool color;
-
-   ~DirList();
-public:
-   DirList(ArgV *a)
-      {
-	 buf=new Buffer();
-	 args=a;
-	 color=false;
-      }
-
-   virtual int Do() = 0;
-   virtual const char *Status() = 0;
-
-   int Size() { return buf->Size(); }
-   bool Eof() { return buf->Eof();  }
-   void Get(const char **b,int *size) { buf->Get(b,size); }
-   void Skip(int len) { buf->Skip(len); }
-   void UseColor(bool c=true) { color=c; }
-};
-
-
 // shortcut
 #define FA FileAccess
 
@@ -515,4 +412,143 @@ public:
    static void ClearAll();
 };
 
+class FileAccessRef : public SMTaskRef<FileAccess>
+{
+   FileAccessRef(const FileAccessRef&);  // disable cloning
+   void operator=(const FileAccessRef&);   // and assignment
+
+   void reuse() { if(ptr) { ptr->DecRefCount(); SessionPool::Reuse(ptr); ptr=0; } }
+
+public:
+   FileAccessRef() {}
+   FileAccessRef(FileAccess *p) : SMTaskRef<FileAccess>(p) {}
+   ~FileAccessRef() { reuse(); }
+   const FileAccessRef& operator=(FileAccess *p) { reuse(); ptr=SMTask::MakeRef(p); return *this; }
+
+   template<class T> const SMTaskRef<T>& Cast() const
+      { void(static_cast<T*>(this->ptr)); return *(const SMTaskRef<T>*)this; }
+
+   static const FileAccessRef null;
+};
+
+// constant ref (ref clone)
+class FileAccessRefC
+{
+   void close() { if(*ref) (*ref)->Close(); }
+
+protected:
+   const FileAccessRef *ref;
+
+public:
+   FileAccessRefC(const FileAccessRef& p) { ref=&p; }
+   ~FileAccessRefC() { close(); }
+   const FileAccessRef& operator=(const FileAccessRef& p) { close(); ref=&p; return p; }
+   operator const FileAccess*() const { return *ref; }
+   FileAccess *operator->() const { return ref->operator->(); }
+   operator const FileAccessRef&() { return *ref; }
+};
+
+// static reference
+class FileAccessRefS : public FileAccessRef
+{
+   FileAccessRefS(const FileAccessRefS&);  // disable cloning
+   void operator=(const FileAccessRefS&);   // and assignment
+
+public:
+   FileAccessRefS() {}
+   FileAccessRefS(FileAccess *p) { ptr=p; }
+   ~FileAccessRefS() { ptr=0; }
+   const FileAccessRefS& operator=(FileAccess *p) { ptr=p; return *this; }
+};
+
+class FileAccessOperation : public SMTask
+{
+protected:
+   FileAccessRefS session;
+   bool done;
+   xstring error_text;
+   void SetError(const char *);
+   void SetErrorCached(const char *);
+
+   bool use_cache;
+
+   ~FileAccessOperation() { if(session) session->Close(); }
+
+public:
+   FileAccessOperation(FileAccess *s) : session(s), done(false), use_cache(true) {}
+
+   virtual int Do() = 0;
+   bool Done() { return done; }
+   bool Error() { return error_text!=0; }
+   const char *ErrorText() { return error_text; }
+
+   virtual const char *Status() = 0;
+
+   void UseCache(bool y=true) { use_cache=y; }
+};
+
+#include "PatternSet.h"
+class ListInfo : public FileAccessOperation
+{
+protected:
+   FileAccess::Path saved_cwd;
+   Ref<FileSet> result;
+
+   const char *exclude_prefix;
+   const PatternSet *exclude;
+
+   unsigned need;
+   bool follow_symlinks;
+
+public:
+   ListInfo(FileAccess *session,const char *path);
+   ~ListInfo();
+
+   void SetExclude(const char *p,const PatternSet *e) { exclude_prefix=p; exclude=e; }
+
+   // caller has to delete the resulting FileSet itself.
+   FileSet *GetResult() { return result.borrow(); }
+
+   void Need(unsigned mask) { need|=mask; }
+   void NoNeed(unsigned mask) { need&=~mask; }
+   void FollowSymlinks() { follow_symlinks=true; }
+};
+
+#include "buffer.h"
+class LsOptions
+{
+public:
+   bool append_type:1;
+   bool multi_column:1;
+   bool show_all:1;
+   LsOptions()
+      {
+	 append_type=false;
+	 multi_column=false;
+	 show_all=false;
+      }
+};
+
+class DirList : public FileAccessOperation
+{
+protected:
+   Ref<Buffer> buf;
+   Ref<ArgV> args;
+   bool color;
+
+public:
+   DirList(FileAccess *s,ArgV *a);
+   ~DirList();
+
+   virtual int Do() = 0;
+   virtual const char *Status() = 0;
+
+   int Size() { return buf->Size(); }
+   bool Eof() { return buf->Eof();  }
+   void Get(const char **b,int *size) { buf->Get(b,size); }
+   void Skip(int len) { buf->Skip(len); }
+
+   void UseColor(bool c=true) { color=c; }
+};
+
 #endif /* FILEACCESS_H */
