@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileCopy.h,v 1.65 2006/05/13 05:47:15 lav Exp $ */
+/* $Id: FileCopy.h,v 1.78 2007/09/10 09:57:54 lav Exp $ */
 
 /*
   classes defined here:
@@ -27,7 +27,6 @@
    +FileCopyPeerFA
    +FileCopyPeerFDStream
    \FileCopyPeerList
-   +FileCopyPeerOutputJob
 */
 
 #ifndef FILECOPY_H
@@ -38,8 +37,7 @@
 #include "FileAccess.h"
 #include "Speedometer.h"
 #include "Timer.h"
-
-#define FILE_END     (-1L)
+#include "log.h"
 
 class FileCopyPeer : public IOBuffer
 {
@@ -49,7 +47,7 @@ protected:
    bool start_transfer;
    off_t size;
    off_t e_size;
-   time_t date;
+   FileTimestamp date;
 
    off_t seek_pos;
    bool can_seek;
@@ -66,24 +64,20 @@ protected:
 
    bool write_allowed;
 
-   char *suggested_filename;
+   xstring_c suggested_filename;
    bool auto_rename;
 
 public:
    off_t range_start; // NOTE: ranges are implemented only partially. (FIXME)
    off_t range_limit;
 
-protected:
-   ~FileCopyPeer();
-
-public:
    bool CanSeek() { return can_seek; }
    bool CanSeek0() { return can_seek0; }
    bool CanSeek(off_t p) { return p==0 ? CanSeek0() : CanSeek(); }
    off_t GetSeekPos() { return seek_pos; }
    virtual void Seek(off_t offs);
    virtual off_t GetRealPos() { return pos; }
-   virtual int Buffered() { return in_buffer; }
+   virtual int Buffered() { return Size(); }
    virtual bool IOReady() { return true; }
 
    virtual void WantDate() { want_date=true; date=NO_DATE_YET; }
@@ -91,16 +85,21 @@ public:
    time_t GetDate() { return date; }
    off_t  GetSize() { if(size>=0 && pos>size) WantSize(); return size; }
 
-   void SetDate(time_t d);
+   void SetDate(time_t d,int p=0);
    void SetSize(off_t s);
    void SetEntitySize(off_t s) { e_size=s; }
 
    void DontCopyDate() { do_set_date=false; }
    bool NeedDate() { return do_set_date; }
 
-   void SetRange(off_t s,off_t lim) { range_start=s; range_limit=lim; }
+   void SetRange(off_t s,off_t lim) {
+      range_start=s; range_limit=lim;
+      if(mode==PUT || range_start>GetPos()+0x4000);
+	 Seek(range_start);
+   }
 
    FileCopyPeer(dir_t m);
+   virtual ~FileCopyPeer() {}
 
    bool Done();
 
@@ -122,7 +121,12 @@ public:
 	 if(p==0)
 	    can_seek0=false;
       }
-   virtual FA *GetSession() { return 0; } // for fxp.
+
+   // for fxp:
+   virtual const FileAccessRef& GetSession() { return FileAccessRef::null; }
+   virtual void OpenSession() {}
+   virtual void SetFXP(bool) {}
+
    virtual void Fg() {}
    virtual void Bg() {}
 
@@ -137,12 +141,14 @@ public:
 
    const char *GetDescriptionForLog() { return 0; }
    virtual const char *GetURL() { return 0; }
+   virtual FileCopyPeer *Clone() { return 0; }
+   virtual const Ref<FDStream>& GetLocal() const { return Ref<FDStream>::null; }
 
    const char *GetSuggestedFileName() { return suggested_filename; }
    void SetSuggestedFileName(const char *f)
       {
 	 if(f && !suggested_filename)
-	    suggested_filename=xstrdup(f);
+	    suggested_filename.set(f);
       }
    void AutoRename(bool yes=true) { auto_rename=yes; }
 };
@@ -150,8 +156,8 @@ public:
 class FileCopy : public SMTask
 {
 public:
-   FileCopyPeer *get;
-   FileCopyPeer *put;
+   SMTaskRef<FileCopyPeer> get;
+   SMTaskRef<FileCopyPeer> put;
 
 protected:
    enum state_t
@@ -166,14 +172,13 @@ protected:
       } state;
 
 private:
-
    int max_buf;
    bool cont;
 
-   char *error_text;
+   xstring_c error_text;
 
-   Speedometer *rate;
-   Speedometer *rate_for_eta;
+   SMTaskRef<Speedometer> rate;
+   SMTaskRef<Speedometer> rate_for_eta;
    int put_buf;
    off_t put_eof_pos;
 
@@ -185,7 +190,7 @@ private:
    bool remove_source_later;
    bool remove_target_first;
 
-   Buffer *line_buffer;
+   Ref<Buffer> line_buffer;
    int  line_buffer_max;
 
 protected:
@@ -202,8 +207,6 @@ protected:
       }
    off_t bytes_count;
 
-   ~FileCopy();
-
 public:
    off_t GetPos();
    off_t GetSize();
@@ -223,9 +226,12 @@ public:
    off_t GetBytesCount() { return bytes_count; }
    double GetTimeSpent();
 
-   void SetDate(time_t t) { get->SetDate(t); }
+   void SetDate(time_t t,int p=0) { get->SetDate(t,p); }
+   void SetDate(const FileTimestamp &t) { SetDate(t.ts,t.ts_prec); }
    void SetSize(off_t  s) { get->SetSize(s); }
 
+   bool SetContinue(bool new_cont) { return replace_value(cont,new_cont); }
+
    bool Done() { return state==ALL_DONE; }
    bool Error() { return error_text!=0; }
    const char *ErrorText() { return error_text; }
@@ -236,13 +242,16 @@ public:
    void DontFailIfBroken() { fail_if_broken=false; }
    void FailIfCannotSeek() { fail_if_cannot_seek=true; }
    void SetRange(off_t s,off_t lim) { get->SetRange(s,lim); put->SetRange(s,lim); }
+   void SetRangeLimit(off_t lim) { get->range_limit=lim; }
+   off_t GetRangeStart() { return get->range_start; }
+   off_t GetRangeLimit() { return get->range_limit; }
    void RemoveSourceLater() { remove_source_later=true; }
    void RemoveTargetFirst() { remove_target_first=true; put->Resume(); put->RemoveFile(); }
    void LineBuffered(int size=0x1000);
    bool IsLineBuffered() const { return line_buffer; }
 
    FileCopy(FileCopyPeer *src,FileCopyPeer *dst,bool cont);
-   void Init();
+   ~FileCopy();
 
    int Do();
    void SuspendInternal();
@@ -250,20 +259,21 @@ public:
    void Fg();
    void Bg();
 
-   void Reconfig(const char *);
-
    static FileCopy *New(FileCopyPeer *src,FileCopyPeer *dst,bool cont);
    static FileCopy *(*fxp_create)(FileCopyPeer *src,FileCopyPeer *dst,bool cont);
 
    void AllowWrite(bool y=true) { if(put) put->AllowWrite(y); }
    bool WriteAllowed() { return !put || put->WriteAllowed(); }
    bool WritePending() { return put && put->WritePending(); }
+
+   void LogTransfer();
+   static SMTaskRef<Log> transfer_log;
 };
 
 class FileVerificator : public SMTask
 {
    bool done;
-   char *error_text;
+   xstring error_text;
    IOBufferFDStream *verify_buffer;
    InputFilter *verify_process;
    void Init0();
@@ -271,7 +281,7 @@ class FileVerificator : public SMTask
 public:
    FileVerificator(const char *f);
    FileVerificator(const FDStream *);
-   FileVerificator(FileAccess *,const char *f);
+   FileVerificator(const FileAccess *,const char *f);
    ~FileVerificator();
    int Do();
    bool Done() { return done; }
@@ -282,17 +292,18 @@ public:
 
 class FileCopyPeerFA : public FileCopyPeer
 {
-   char *file;
-   char *orig_url;
+   xstring_c file;
+   xstring orig_url;
+   FileAccessRef my_session;
+   FileAccessRefC session;
    int FAmode;
-   FileAccess *session;
 
    int Get_LL(int size);
    int Put_LL(const char *buf,int size);
+   int PutEOF_LL();
 
    FileAccess::fileinfo info;
 
-   bool reuse_later;
    bool fxp;   // FXP (ftp<=>ftp copy) active
 
    time_t try_time;
@@ -300,7 +311,7 @@ class FileCopyPeerFA : public FileCopyPeer
 
    int redirections;
 
-   FileVerificator *verify;
+   Ref<FileVerificator> verify;
 
 protected:
    ~FileCopyPeerFA();
@@ -308,45 +319,48 @@ protected:
 public:
    void Init();
    FileCopyPeerFA(FileAccess *s,const char *f,int m);
-   FileCopyPeerFA(class ParsedURL *u,int m);
+   FileCopyPeerFA(const FileAccessRef& s,const char *f,int m);
+   FileCopyPeerFA(const class ParsedURL *u,int m);
    int Do();
    bool IOReady();
    off_t GetRealPos();
    void Seek(off_t new_pos);
 
-   int Buffered() { return in_buffer+session->Buffered(); }
+   int Buffered() { return Size()+session->Buffered(); }
 
    void SuspendInternal();
    void ResumeInternal();
 
-   void DontReuseSession() { reuse_later=false; }
-
    const char *GetStatus();
-   const char *GetProto() { return session->GetProto(); }
+   const char *GetProto() const { return session->GetProto(); }
 
    bool NeedSizeDateBeforehand() { return session->NeedSizeDateBeforehand(); }
+   void WantSize();
    void RemoveFile();
 
-   static FileCopyPeerFA *New(FA *s,const char *url,int m,bool reuse=false);
+   static FileCopyPeerFA *New(FA *s,const char *url,int m);
+   static FileCopyPeerFA *New(const FileAccessRef& s,const char *url,int m);
 
    void OpenSession();
-   FA *GetSession() { return session; }
+   const FileAccessRef& GetSession() { return session; }
    void Fg() { session->SetPriority(1); }
    void Bg() { session->SetPriority(0); }
    void SetFXP(bool on) { fxp=on; }
 
    const char *GetDescriptionForLog()
       {
-	 return orig_url ? orig_url : session->GetFileURL(file);
+	 return orig_url ? orig_url.get() : session->GetFileURL(file);
       }
    const char *GetURL() { return GetDescriptionForLog(); }
+   FileCopyPeer *Clone();
 };
 
 class FileCopyPeerFDStream : public FileCopyPeer
 {
-   FDStream *stream;
+   Ref<FDStream> my_stream;
+   const Ref<FDStream>& stream;
    off_t seek_base;
-   Timer *put_ll_timer;
+   Ref<Timer> put_ll_timer;
 
    int Get_LL(int size);
    int Put_LL(const char *buf,int size);
@@ -354,16 +368,14 @@ class FileCopyPeerFDStream : public FileCopyPeer
 
    int getfd();
 
-   bool delete_stream;
    bool create_fg_data;
    bool need_seek;
 
-   FileVerificator *verify;
-
-protected:
-   ~FileCopyPeerFDStream();
+   Ref<FileVerificator> verify;
 
 public:
+   void Init();
+   FileCopyPeerFDStream(const Ref<FDStream>& o,dir_t m);
    FileCopyPeerFDStream(FDStream *o,dir_t m);
    int Do();
    bool IOReady();
@@ -372,9 +384,9 @@ public:
    pid_t GetProcGroup() { return stream->GetProcGroup(); }
    void Kill(int sig);
 
-   void DontDeleteStream() { delete_stream=false; }
    void DontCreateFgData() { create_fg_data=false; }
    void NeedSeek() { need_seek=true; }
+   void WantSize();
    void RemoveFile();
    void SetBase(off_t b) { seek_base=b; }
 
@@ -391,16 +403,15 @@ public:
       {
 	 return stream->full_name;
       }
+   const Ref<FDStream>& GetLocal() const { return stream; }
+   FileCopyPeer *Clone();
 };
 
 class FileCopyPeerDirList : public FileCopyPeer
 {
 private:
-   FileAccess *session;
-   DirList *dl;
-
-protected:
-   ~FileCopyPeerDirList();
+   FileAccessRef session;
+   SMTaskRef<DirList> dl;
 
 public:
    FileCopyPeerDirList(FA *s,ArgV *v); // consumes s and v.
@@ -413,24 +424,4 @@ public:
    void UseColor(bool c=true) { if(dl) dl->UseColor(c); }
 };
 
-class OutputJob;
-
-class FileCopyPeerOutputJob : public FileCopyPeer
-{
-   OutputJob *o;
-   int Put_LL(const char *buf,int len);
-
-public:
-   FileCopyPeerOutputJob(OutputJob *o);
-
-   int Do();
-   void Fg();
-   void Bg();
-
-   const char *GetDescriptionForLog()
-      {
-	 return "[pipe to other job]";
-      }
-};
-
 #endif
