@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileCopyFtp.cc,v 1.23 2006/04/19 09:48:03 lav Exp $ */
+/* $Id: FileCopyFtp.cc,v 1.25 2007/04/23 07:11:41 lav Exp $ */
 
 #include <config.h>
 
@@ -32,6 +32,9 @@
 #define passive_ssl_connect true
 #endif
 
+#define ftp_src get->GetSession().Cast<Ftp>()
+#define ftp_dst put->GetSession().Cast<Ftp>()
+
 void FileCopyFtp::Close()
 {
    ftp_src->Close();
@@ -52,7 +55,7 @@ int FileCopyFtp::Do()
 	 return m;
       if(ftp_dst->IsClosed())
       {
-	 ((FileCopyPeerFA*)put)->OpenSession();
+	 put->OpenSession();
 	 ftp_dst->SetCopyMode(Ftp::COPY_DEST,!passive_source,protect,
 	       passive_source^passive_ssl_connect,dst_retries,dst_try_time);
 	 m=MOVED;
@@ -68,14 +71,14 @@ int FileCopyFtp::Do()
 
    if(ftp_src->IsClosed())
    {
-      ((FileCopyPeerFA*)get)->OpenSession();
+      get->OpenSession();
       ftp_src->SetCopyMode(Ftp::COPY_SOURCE,passive_source,protect,
 	    !passive_source^passive_ssl_connect,src_retries,src_try_time);
       m=MOVED;
    }
    if(ftp_dst->IsClosed())
    {
-      ((FileCopyPeerFA*)put)->OpenSession();
+      put->OpenSession();
       ftp_dst->SetCopyMode(Ftp::COPY_DEST,!passive_source,protect,
 	    passive_source^passive_ssl_connect,dst_retries,dst_try_time);
       m=MOVED;
@@ -116,8 +119,8 @@ int FileCopyFtp::Do()
 	 Log::global->Write(0,_("**** FXP: giving up, reverting to plain copy\n"));
 	 Close();
 	 disable_fxp=true;
-	 ((FileCopyPeerFA*)get)->SetFXP(false);
-	 ((FileCopyPeerFA*)put)->SetFXP(false);
+	 get->SetFXP(false);
+	 put->SetFXP(false);
 
 	 if(ResMgr::QueryBool("ftp:fxp-force",ftp_src->GetHostName())
 	 || ResMgr::QueryBool("ftp:fxp-force",ftp_dst->GetHostName()))
@@ -204,7 +207,6 @@ int FileCopyFtp::Do()
 
 void FileCopyFtp::Init()
 {
-   ftp_src=ftp_dst=0;
    no_rest=false;
    orig_passive_source=passive_source=false;
    src_retries=dst_retries=0;
@@ -216,22 +218,15 @@ void FileCopyFtp::Init()
 #endif
 }
 
-FileCopyFtp::~FileCopyFtp()
-{
-}
-
 // s,d must be FileCopyPeerFA, s->GetSession(),d->GetSession() must be Ftp.
 FileCopyFtp::FileCopyFtp(FileCopyPeer *s,FileCopyPeer *d,bool c,bool rp)
    : super(s,d,c)
 {
    Init();
    passive_source=rp;
-   // not pretty.
-   ftp_src=(Ftp*)(s->GetSession());
-   ftp_dst=(Ftp*)(d->GetSession());
 
-   ((FileCopyPeerFA*)s)->SetFXP(true);
-   ((FileCopyPeerFA*)d)->SetFXP(true);
+   get->SetFXP(true);
+   put->SetFXP(true);
 
    if(ftp_src->IsPassive() && !ftp_dst->IsPassive())
       passive_source=true;
@@ -250,8 +245,8 @@ FileCopyFtp::FileCopyFtp(FileCopyPeer *s,FileCopyPeer *d,bool c,bool rp)
 
 FileCopy *FileCopyFtp::New(FileCopyPeer *s,FileCopyPeer *d,bool c)
 {
-   FA *s_s=s->GetSession();
-   FA *d_s=d->GetSession();
+   const FileAccessRef& s_s=s->GetSession();
+   const FileAccessRef& d_s=d->GetSession();
    if(!s_s || !d_s)
       return 0;
    if((strcmp(s_s->GetProto(),"ftp") && strcmp(s_s->GetProto(),"ftps"))
