@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileCopyFtp.h,v 1.5 2004/05/26 12:31:42 lav Exp $ */
+/* $Id: FileCopyFtp.h,v 1.7 2007/04/23 07:11:41 lav Exp $ */
 
 #ifndef FILECOPYFTP_H
 #define FILECOPYFTP_H
@@ -28,8 +28,6 @@
 
 class FileCopyFtp : public FileCopy
 {
-   Ftp *ftp_src;
-   Ftp *ftp_dst;
    bool no_rest;
    bool passive_source;
    bool orig_passive_source;
@@ -49,12 +47,10 @@ class FileCopyFtp : public FileCopy
 public:
    void Init();
    FileCopyFtp(FileCopyPeer *src,FileCopyPeer *dst,bool cont,bool rp);
-   ~FileCopyFtp();
 
    int Do();
 
    static FileCopy *New(FileCopyPeer *src,FileCopyPeer *dst,bool cont);
-/*   void ShowRunStatus(StatusLine *sl);*/
 };
 
 #endif // FILECOPYFTP_H
