@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileGlob.cc,v 1.3 2003/12/17 12:59:29 lav Exp $ */
+/* $Id: FileGlob.cc,v 1.11 2007/12/28 10:23:10 lav Exp $ */
 
 #include <config.h>
 #include "trio.h"
@@ -29,19 +29,13 @@
 #include <fnmatch.h>
 
 #include "FileGlob.h"
-#include "xmalloc.h"
 #include "misc.h"
 #include "url.h"
 
 // Glob implementation
-Glob::~Glob()
+Glob::Glob(FileAccess *s,const char *p)
+   : FileAccessOperation(s), pattern(p)
 {
-   xfree(pattern);
-}
-
-Glob::Glob(const char *p)
-{
-   pattern=xstrdup(p);
    dirs_only=false;
    files_only=false;
    match_period=true;
@@ -50,15 +44,11 @@ Glob::Glob(const char *p)
 
    if(pattern[0]=='~')
    {
-      char *slash=strchr(pattern,'/');
+      const char *slash=strchr(pattern,'/');
       if(slash)
-      {
-	 *slash=0;
-	 inhibit_tilde=HasWildcards(pattern);
-	 *slash='/';
-      }
+	 inhibit_tilde=HasWildcards(xstring::get_tmp(pattern,slash-pattern));
       else
-	 inhibit_tilde=HasWildcards(pattern);;
+	 inhibit_tilde=HasWildcards(pattern);
    }
    if(pattern[0] && !HasWildcards(pattern))
    {
@@ -69,6 +59,9 @@ Glob::Glob(const char *p)
       done=true;
    }
 }
+Glob::~Glob()
+{
+}
 
 void Glob::add_force(const FileInfo *info)
 {
@@ -86,7 +79,7 @@ void Glob::add(const FileInfo *info)
 	 return;
    }
 
-   char *s=info->name;
+   const char *s=info->name;
    if(s==0)
       return;
 
@@ -174,26 +167,22 @@ int NoGlob::Do()
    }
    return STALL;
 }
-NoGlob::NoGlob(const char *p) : Glob(p)
+NoGlob::NoGlob(const char *p) : Glob(0,p)
 {
 }
 
 void GlobURL::NewGlob(const char *p)
 {
-   SMTask::Delete(glob);
    glob=0;
-   if(session!=orig_session)
-      SessionPool::Reuse(session);
    session=orig_session;
-   xfree(url_prefix);
 
-   url_prefix=xstrdup(p);
-   url_prefix[url::path_index(p)]=0;
+   url_prefix.set(p);
+   url_prefix.truncate(url::path_index(p));
 
    ParsedURL p_url(p,true);
    if(p_url.proto && p_url.path)
    {
-      session=FA::New(&p_url);
+      session=my_session=FA::New(&p_url);
       if(session)
 	 glob=session->MakeGlob(p_url.path);
    }
@@ -209,22 +198,11 @@ void GlobURL::NewGlob(const char *p)
       glob->DirectoriesOnly();
 }
 
-GlobURL::GlobURL(FileAccess *s,const char *p,type_select t)
+GlobURL::GlobURL(const FileAccessRef& s,const char *p,type_select t)
+   : orig_session(s), session(orig_session), type(t)
 {
-   orig_session=session=s;
-   glob=0;
-   url_prefix=0;
-   type=t;
-
    NewGlob(p);
 }
-GlobURL::~GlobURL()
-{
-   SMTask::Delete(glob);
-   if(session!=orig_session)
-      SessionPool::Reuse(session);
-   xfree(url_prefix);
-}
 FileSet *GlobURL::GetResult()
 {
    FileSet &list=*glob->GetResult();
@@ -237,12 +215,9 @@ FileSet *GlobURL::GetResult()
 
 // GenericGlob implementation
 GenericGlob::GenericGlob(FileAccess *s,const char *n_pattern)
-   : Glob(n_pattern)
+   : Glob(s,n_pattern)
 {
-   session=s;
    dir_list=0;
-   updir_glob=0;
-   li=0;
    curr_dir=0;
 
    if(done)
@@ -259,17 +234,11 @@ GenericGlob::GenericGlob(FileAccess *s,const char *n_pattern)
 
    if(dir)
    {
-      updir_glob=new GenericGlob(session,dir);
+      updir_glob=new GenericGlob(s,dir);
       updir_glob->DirectoriesOnly();
    }
 }
 
-GenericGlob::~GenericGlob()
-{
-   Delete(li);
-   Delete(updir_glob);
-}
-
 int GenericGlob::Do()
 {
    int m=STALL;
@@ -281,7 +250,6 @@ int GenericGlob::Do()
    {
       if(updir_glob->Error())
       {
-	 Delete(updir_glob);
 	 updir_glob=0;
 	 done=true;
 	 return MOVED;
@@ -311,12 +279,17 @@ int GenericGlob::Do()
 	 set->rewind();
 	 for(FileInfo *info=set->curr(); info!=NULL; info=set->next())
 	 {
-	    info->SetName(dir_file(curr_dir,info->name));
+	    const char *name=info->name;
+	    if(name[0]=='.' && name[1]=='/')
+	       name+=2;
+	    if(curr_dir && curr_dir[0])
+	       name=dir_file(curr_dir,name);
+	    info->SetName(name);
 	    add(info);
 	 }
 	 delete set;
       }
-      Delete(li); li=0;
+      li=0;
 
       if(dir_list)
 	 dir_list->next();
@@ -356,8 +329,7 @@ const char *GenericGlob::Status()
    if(!curr_dir)
       return st;
 
-   static char *buf = 0;
-   if(buf) xfree(buf);
-   buf=xasprintf("%s: %s", curr_dir, st);
+   static xstring buf;
+   buf.vset(curr_dir,": ",st,NULL);
    return buf;
 }
