@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Http.cc,v 1.216 2006/07/26 13:15:26 lav Exp $ */
+/* $Id: Http.cc,v 1.231 2008/03/05 06:39:43 lav Exp $ */
 
 #include <config.h>
 #include "trio.h"
@@ -39,7 +39,6 @@
 #include "HttpDir.h"
 #include "misc.h"
 #include "buffer_ssl.h"
-#include "lftp_ssl.h"
 
 #include "ascii_ctype.h"
 
@@ -58,11 +57,15 @@ CDECL char *strptime(const char *buf, const char *format, struct tm *tm);
 /* Some status code validation macros: */
 #define H_20X(x)        (((x) >= 200) && ((x) < 300))
 #define H_PARTIAL(x)    ((x) == 206)
-#define H_REDIRECTED(x) (((x) == 301) || ((x) == 302))
+#define H_REDIRECTED(x) (((x) == 301) || ((x) == 302) || ((x) == 303) || ((x) == 307))
 #define H_EMPTY(x)	(((x) == 204) || ((x) == 205))
 #define H_CONTINUE(x)	((x) == 100 || (x) == 102)
 #define H_REQUESTED_RANGE_NOT_SATISFIABLE(x) ((x) == 416)
 
+#ifndef EINPROGRESS
+#define EINPROGRESS -1
+#endif
+
 void Http::Init()
 {
    state=DISCONNECTED;
@@ -72,8 +75,6 @@ void Http::Init()
    recv_buf=0;
    body_size=-1;
    bytes_received=0;
-   line=0;
-   status=0;
    status_code=0;
    status_consumed=0;
    proto_version=0x10;
@@ -105,7 +106,6 @@ void Http::Init()
    use_head=true;
 
    user_agent=0;
-   special_data=0;
    special=HTTP_NONE;
 }
 
@@ -117,14 +117,6 @@ Http::Http() : super()
 Http::Http(const Http *f) : super(f)
 {
    Init();
-   if(f->peer)
-   {
-      peer=(sockaddr_u*)xmemdup(f->peer,f->peer_num*sizeof(*peer));
-      peer_num=f->peer_num;
-      peer_curr=f->peer_curr;
-      if(peer_curr>=peer_num)
-	 peer_curr=0;
-   }
    Reconfig();
 }
 
@@ -136,10 +128,10 @@ Http::~Http()
 
 void Http::MoveConnectionHere(Http *o)
 {
-   send_buf=o->send_buf; o->send_buf=0;
-   recv_buf=o->recv_buf; o->recv_buf=0;
+   send_buf=o->send_buf.borrow();
+   recv_buf=o->recv_buf.borrow();
    sock=o->sock; o->sock=-1;
-   rate_limit=o->rate_limit; o->rate_limit=0;
+   rate_limit=o->rate_limit.borrow();
    last_method=o->last_method; o->last_method=0;
    timeout_timer.Reset(o->timeout_timer);
    state=CONNECTED;
@@ -149,18 +141,12 @@ void Http::MoveConnectionHere(Http *o)
 
 void Http::Disconnect()
 {
-   Delete(send_buf);
    send_buf=0;
-   Delete(recv_buf);
    recv_buf=0;
-   if(rate_limit)
-   {
-      delete rate_limit;
-      rate_limit=0;
-   }
+   rate_limit=0;
    if(sock!=-1)
    {
-      DebugPrint("---- ",_("Closing HTTP connection"),7);
+      LogNote(7,_("Closing HTTP connection"));
       close(sock);
       sock=-1;
    }
@@ -182,11 +168,9 @@ void Http::ResetRequestData()
    body_size=-1;
    bytes_received=0;
    real_pos=no_ranges?0:-1;
-   xfree(status);
-   status=0;
+   status.set(0);
    status_consumed=0;
-   xfree(line);
-   line=0;
+   line.set(0);
    sent_eot=false;
    keep_alive=false;
    keep_alive_max=-1;
@@ -203,12 +187,12 @@ void Http::Close()
    if(mode==CLOSED)
       return;
    if(recv_buf)
-      Roll(recv_buf);	// try to read any remaining data
+      recv_buf->Roll();	// try to read any remaining data
    if(sock!=-1 && keep_alive && (keep_alive_max>0 || keep_alive_max==-1)
    && mode!=STORE && !recv_buf->Eof() && (state==RECEIVING_BODY || state==DONE))
    {
       recv_buf->Resume();
-      Roll(recv_buf);
+      recv_buf->Roll();
       if(xstrcmp(last_method,"HEAD"))
       {
 	 // check if all data are in buffer
@@ -223,7 +207,6 @@ void Http::Close()
       // can reuse the connection.
       state=CONNECTED;
       ResetRequestData();
-      delete rate_limit;
       rate_limit=0;
    }
    else
@@ -237,45 +220,26 @@ void Http::Close()
    no_ranges=false;
    use_propfind_now=QueryBool("use-propfind",hostname);
    special=HTTP_NONE;
-   xfree(special_data);
-   special_data=0;
+   special_data.set(0);
    super::Close();
 }
 
 void Http::Send(const char *format,...)
 {
    va_list va;
-   char *str;
-
-   static int max_send=256;
-   for(;;)
-   {
-      va_start(va,format);
-      str=string_alloca(max_send);
-      int res=vsnprintf(str,max_send,format,va);
-      va_end(va);
-      if(res>=0 && res<max_send)
-      {
-	 if(res<max_send/16)
-	    max_send/=2;
-	 break;
-      }
-      max_send*=2;
-   }
-
-   DebugPrint("---> ",str,5);
+   va_start(va,format);
+   xstring& str=xstring::vformat(format,va);
+   va_end(va);
+   LogSend(5,str);
    send_buf->Put(str);
 }
 
 void Http::SendMethod(const char *method,const char *efile)
 {
-   char *ehost=string_alloca(strlen(hostname)*3+1+xstrlen(portname)*3+1);
-   url::encode_string(hostname,ehost,URL_HOST_UNSAFE);
-   if(portname)
-   {
-      strcat(ehost,":");
-      url::encode_string(portname,ehost+strlen(ehost),URL_PORT_UNSAFE);
-   }
+   const char *ehost=xstring::join(":",2,
+      url::encode(hostname,URL_HOST_UNSAFE),
+      url::encode(portname,URL_PORT_UNSAFE)
+   );
    if(!use_head && !strcmp(method,"HEAD"))
       method="GET";
    last_method=method;
@@ -343,10 +307,10 @@ void Http::SendMethod(const char *method,const char *efile)
       if(referer && referer[0])
 	 Send("Referer: %s%s\r\n",referer,slash);
 
-      char *cookie=MakeCookie(hostname,efile+(proxy?url::path_index(efile):0));
+      xstring cookie;
+      MakeCookie(cookie,hostname,efile+(proxy?url::path_index(efile):0));
       if(cookie && cookie[0])
-	 Send("Cookie: %s\r\n",cookie);
-      xfree(cookie);
+	 Send("Cookie: %s\r\n",cookie.get());
    }
 }
 
@@ -362,9 +326,7 @@ void Http::SendBasicAuth(const char *tag,const char *auth)
 void Http::SendBasicAuth(const char *tag,const char *user,const char *pass)
 {
    /* Basic scheme */
-   char *buf=string_alloca(strlen(user)+1+strlen(pass)+1);
-   sprintf(buf,"%s:%s",user,pass);
-   SendBasicAuth(tag,buf);
+   SendBasicAuth(tag,xstring::cat(user,":",pass,NULL));
 }
 
 void Http::SendAuth()
@@ -392,18 +354,9 @@ void Http::SendCacheControl()
 	     && (pos[cc_no_cache_len]==0 || pos[cc_no_cache_len]==' '))
 	 cc_no_cache=0, cc_no_cache_len=0;
    }
-   char *cc=string_alloca(xstrlen(cc_setting)+1+cc_no_cache_len+1);
-   cc[0]=0;
-   if(cc_no_cache)
-      strcpy(cc,cc_no_cache);
-   if(cc_setting)
-   {
-      if(cc[0])
-	 strcat(cc,",");
-      strcat(cc,cc_setting);
-   }
+   xstring& cc=xstring::join(",",2,cc_no_cache,cc_setting);
    if(*cc)
-      Send("Cache-Control: %s\r\n",cc);
+      Send("Cache-Control: %s\r\n",cc.get());
 }
 
 bool Http::ModeSupported()
@@ -437,8 +390,9 @@ bool Http::ModeSupported()
    abort(); // should not happen
 }
 
-void Http::DirFile(char *path_base,const char *ecwd,const char *efile)
+void Http::DirFile(xstring& path,const char *ecwd,const char *efile)
 {
+   int base=path.length();
    if(!strcmp(ecwd,"~") && !hftp)
       ecwd="";
    const char *sep=(last_char(ecwd)=='/'?"":"/");
@@ -446,96 +400,93 @@ void Http::DirFile(char *path_base,const char *ecwd,const char *efile)
       sep="";
    const char *pre=(ecwd[0]=='/'?"":"/");
    if(efile[0]=='/')
-      strcpy(path_base,efile);
+      path.append(efile);
    else if(efile[0]=='~')
-      sprintf(path_base,"/%s",efile);
+      path.vappend("/",efile,NULL);
    else
-      sprintf(path_base,"%s%s%s%s",pre,ecwd,sep,efile);
+      path.vappend(pre,ecwd,sep,efile,NULL);
 
-   if(path_base[1]=='~' && path_base[2]==0)
-      path_base[1]=0;
-   else if(path_base[1]=='~' && path_base[2]=='/')
-      memmove(path_base,path_base+2,strlen(path_base+2)+1);
+   if(path[base+1]=='~' && path[base+2]==0)
+      path.truncate(base+1);
+   else if(path[base+1]=='~' && path[base+2]=='/')
+      path.set_substr(base,2,"");
 }
 
 void Http::SendRequest(const char *connection,const char *f)
 {
-   char *efile=string_alloca(strlen(f)*3+1);
-   url::encode_string(f,efile,URL_PATH_UNSAFE);
-   if(mode==CHANGE_DIR && new_cwd && new_cwd->url)
-      efile=new_cwd->url+url::path_index(new_cwd->url);
-   char *ecwd;
-   int efile_len;
+   xstring efile;
+   xstring ecwd;
+   bool add_slash=true;
 
-   if(cwd.url)
+   if(mode==CHANGE_DIR && new_cwd && new_cwd->url)
    {
-      ecwd=cwd.url+url::path_index(cwd.url);
-      ecwd=alloca_strdup(ecwd);
+      const char *efile_c=new_cwd->url+url::path_index(new_cwd->url);
+      if(!*efile_c)
+	 efile_c="/";
+      efile.set(efile_c);
+      add_slash=false;
    }
+   else
+      efile.set(url::encode(f,URL_PATH_UNSAFE));
+
+   if(cwd.url)
+      ecwd.set(cwd.url+url::path_index(cwd.url));
    else
    {
-      ecwd=string_alloca(strlen(cwd)*3+3+1);
-      url::encode_string(cwd,ecwd,URL_PATH_UNSAFE);
+      ecwd.set(url::encode(cwd,URL_PATH_UNSAFE));
       if(hftp && ecwd[0]=='/' && ecwd[1]!='~')
       {
 	 // root directory in ftp urls needs special encoding. (/%2Fpath)
-	 memmove(ecwd+4,ecwd+1,strlen(ecwd+1)+1);
-	 memcpy(ecwd+1,"%2F",3);
+	 ecwd.set_substr(1,0,"%2F");
       }
    }
-   bool add_slash=true;
+
    if(cwd.is_file)
    {
       if(efile[0])
-	 dirname_modify(ecwd+(!strncmp(ecwd,"/~",2)));
-      else
-	 add_slash=false;
+	 ecwd.truncate(basename_ptr(ecwd+(!strncmp(ecwd,"/~",2)))-ecwd);
+      add_slash=false;
    }
-   if(mode==CHANGE_DIR && new_cwd)
+   if(mode==CHANGE_DIR && new_cwd && !new_cwd->url)
       add_slash=!new_cwd->is_file;
 
-   char *pfile=string_alloca(4+3+xstrlen(user)*6+3+xstrlen(pass)*3+1+
-			      strlen(hostname)*3+1+xstrlen(portname)*3+1+
-			      strlen(ecwd)+1+strlen(efile)+1+6+1);
-
    const char *allprop=	// PROPFIND request
       "<?xml version=\"1.0\" ?>"
       "<propfind xmlns=\"DAV:\">"
         "<allprop/>"
       "</propfind>\r\n";
 
+   xstring pfile;
    if(proxy && !https)
    {
       const char *proto="http";
       if(hftp)
 	 proto="ftp";
-      sprintf(pfile,"%s://",proto);
+      pfile.vset(proto,"://",NULL);
       if(hftp && user && pass)
       {
-	 url::encode_string(user,pfile+strlen(pfile),URL_USER_UNSAFE);
+	 pfile.append(url::encode(user,URL_USER_UNSAFE));
 	 if(!QueryBool("use-authorization",proxy))
 	 {
-	    strcat(pfile,":");
-	    url::encode_string(pass,pfile+strlen(pfile),URL_PASS_UNSAFE);
+	    pfile.append(':');
+	    pfile.append(url::encode(pass,URL_PASS_UNSAFE));
 	 }
-	 strcat(pfile,"@");
+	 pfile.append('@');
       }
-      url::encode_string(hostname,pfile+strlen(pfile),URL_HOST_UNSAFE);
+      pfile.append(url::encode(hostname,URL_HOST_UNSAFE));
       if(portname)
       {
-	 strcat(pfile,":");
-	 url::encode_string(portname,pfile+strlen(pfile),URL_PORT_UNSAFE);
+	 pfile.append(':');
+	 pfile.append(url::encode(portname,URL_PORT_UNSAFE));
       }
    }
    else
    {
-      pfile[0]=0;
+      pfile.set("");
    }
 
-   char *path_base=pfile+strlen(pfile);
-   DirFile(path_base,ecwd,efile);
-   efile=pfile;
-   efile_len=strlen(efile);
+   DirFile(pfile,ecwd,efile);
+   efile.set(pfile);
 
    if(pos==0)
       real_pos=0;
@@ -557,12 +508,12 @@ void Http::SendRequest(const char *connection,const char *f)
       switch(special)
       {
       case HTTP_POST:
-	 entity_size=xstrlen(special_data);
+	 entity_size=special_data.length();
 	 goto send_post;
       case HTTP_MOVE:
       case HTTP_COPY:
 	 SendMethod(special==HTTP_MOVE?"MOVE":"COPY",efile);
-	 Send("Destination: %s\r\n",special_data);
+	 Send("Destination: %s\r\n",special_data.get());
 	 break;
       case HTTP_NONE:
 	 abort(); // cannot happen
@@ -628,7 +579,7 @@ void Http::SendRequest(const char *connection,const char *f)
    case MP_LIST:
    case MAKE_DIR:
       if(last_char(efile)!='/' && add_slash)
-	 strcat(efile,"/");
+	 efile.append('/');
       if(mode==CHANGE_DIR)
       {
 	 if(use_propfind_now)
@@ -695,7 +646,7 @@ void Http::SendRequest(const char *connection,const char *f)
    if(special==HTTP_POST)
    {
       if(special_data)
-	 Send("%s",special_data);
+	 Send("%s",special_data.get());
       entity_size=NO_SIZE;
    }
    else if(mode==MP_LIST || (mode==CHANGE_DIR && use_propfind_now))
@@ -732,13 +683,12 @@ void Http::SendArrayInfoRequest()
 
 static const char *extract_quoted_header_value(const char *value)
 {
-   static char *value1;
-   xfree(value1);
+   static xstring value1;
    if(*value=='"')
    {
       value++;
-      value1=(char*)xmalloc(strlen(value));
-      char *store=value1;
+      value1.set(value);
+      char *store=value1.get_non_const();
       while(*value && *value!='"')
       {
 	 if(*value=='\\' && value[1])
@@ -750,8 +700,8 @@ static const char *extract_quoted_header_value(const char *value)
    else
    {
       int end=strcspn(value,"()<>@,;:\\\"/[]?={} \t");
-      value1=xstrdup(value);
-      value1[end]=0;
+      value1.set(value);
+      value1.truncate(end);
    }
    return value1;
 }
@@ -819,8 +769,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
    }
    if(!strcasecmp(name,"Location"))
    {
-      xfree(location);
-      location=xstrdup(value);
+      location.set(value);
       return;
    }
    if(!strcasecmp(name,"Keep-Alive"))
@@ -882,14 +831,12 @@ void Http::HandleHeaderLine(const char *name,const char *value)
    }
    if(!strcasecmp(name,"Content-Type"))
    {
-      xfree(entity_content_type);
-      entity_content_type=xstrdup(value);
+      entity_content_type.set(value);
       const char *cs=strstr(value,"charset=");
       if(cs)
       {
 	 cs=extract_quoted_header_value(cs+8);
-	 xfree(entity_charset);
-	 entity_charset=xstrdup(cs);
+	 entity_charset.set(cs);
       }
    }
 }
@@ -936,15 +883,6 @@ void Http::GetBetterConnection(int level)
 	 return;
       }
 
-      if(!peer && o->peer)
-      {
-	 // copy resolved address so that it would be possible to create
-	 // data connection.
-	 peer=(sockaddr_u*)xmemdup(o->peer,o->peer_num*sizeof(*o->peer));
-	 peer_num=o->peer_num;
-	 peer_curr=o->peer_curr;
-      }
-
       // so borrow the connection
       MoveConnectionHere(o);
       return;
@@ -961,7 +899,7 @@ int Http::Do()
    // check if idle time exceeded
    if(mode==CLOSED && sock!=-1 && idle_timer.Stopped())
    {
-      DebugPrint("---- ",_("Closing idle connection"),1);
+      LogNote(1,_("Closing idle connection"));
       Disconnect();
       return m;
    }
@@ -1023,16 +961,14 @@ int Http::Do()
 	    if(*scan!='/')
 	       strcat(url,"/");
 	    strcat(url,scan);
-	    char *path=xstrdup(url);
-	    char *space=strchr(path,' ');
-	    if(space)
-	       *space=0;
+
+	    file_url.set(url);
+	    file_url.truncate_at(' ');
+
 	    scan=strchr(scan,' ');
 	    while(scan && *scan==' ')
 	       scan++;
-	    special_data=xstrdup(scan);
-	    xfree(file_url);
-	    file_url=path;
+	    special_data.set(scan);
 	    return MOVED;
 	 }
 	 state=DONE;
@@ -1086,7 +1022,7 @@ int Http::Do()
       sock=SocketCreateTCP(peer[peer_curr].sa.sa_family);
       if(sock==-1)
       {
-	 if(peer_curr+1<peer_num)
+	 if(peer_curr+1<peer.count())
 	 {
 	    peer_curr++;
 	    retries--;
@@ -1108,14 +1044,10 @@ int Http::Do()
 
       SayConnectingTo();
       res=SocketConnect(sock,&peer[peer_curr]);
-      if(res==-1
-#ifdef EINPROGRESS
-      && errno!=EINPROGRESS
-#endif
-      )
+      if(res==-1 && errno!=EINPROGRESS)
       {
 	 NextPeer();
-	 Log::global->Format(0,"connect: %s\n",strerror(errno));
+	 LogError(0,"connect: %s\n",strerror(errno));
 	 Disconnect();
 	 if(NotSerious(errno))
 	    return MOVED;
@@ -1163,7 +1095,7 @@ int Http::Do()
 	 {
 	    // have to setup a tunnel.
 	    char *ehost=string_alloca(strlen(hostname)*3+1);
-	    const char *port_to_use=portname?portname:HTTPS_DEFAULT_PORT;
+	    const char *port_to_use=portname?portname.get():HTTPS_DEFAULT_PORT;
 	    char *eport=string_alloca(strlen(port_to_use)*3+1);
 	    url::encode_string(hostname,ehost,URL_HOST_UNSAFE);
 	    url::encode_string(port_to_use,eport,URL_PORT_UNSAFE);
@@ -1183,7 +1115,7 @@ int Http::Do()
 	 goto handle_quote_cmd;
       if(recv_buf->Eof())
       {
-	 DebugPrint("**** ",_("Peer closed connection"),0);
+	 LogError(0,_("Peer closed connection"));
 	 Disconnect();
 	 return MOVED;
       }
@@ -1199,7 +1131,7 @@ int Http::Do()
 	 state=DONE;
 	 return MOVED;
       }
-      DebugPrint("---- ",_("Sending request..."));
+      LogNote(9,_("Sending request..."));
       if(mode==ARRAY_INFO)
       {
 	 SendArrayInfoRequest();
@@ -1225,13 +1157,13 @@ int Http::Do()
       handle_buf_error:
 	 if(send_buf->Error())
 	 {
-	    DebugPrint("**** ",send_buf->ErrorText(),0);
+	    LogError(0,"send: %s",send_buf->ErrorText());
 	    if(send_buf->ErrorFatal())
 	       SetError(FATAL,send_buf->ErrorText());
 	 }
 	 if(recv_buf->Error())
 	 {
-	    DebugPrint("**** ",recv_buf->ErrorText(),0);
+	    LogError(0,"recv: %s",recv_buf->ErrorText());
 	    if(recv_buf->ErrorFatal())
 	       SetError(FATAL,recv_buf->ErrorText());
 	 }
@@ -1246,7 +1178,7 @@ int Http::Do()
       if(!buf)
       {
 	 // eof
-	 DebugPrint("**** ",_("Hit EOF while fetching headers"),0);
+	 LogError(0,_("Hit EOF while fetching headers"));
 	 // workaround some broken servers
 	 if(H_REDIRECTED(status_code) && location)
 	    goto pre_RECEIVING_BODY;
@@ -1262,7 +1194,7 @@ int Http::Do()
 	    // empty line indicates end of headers.
 	    if(eol==buf && status)
 	    {
-	       DebugPrint("<--- ","",4);
+	       LogRecv(4,"");
 	       recv_buf->Skip(eol_size);
 	       if(tunnel_state==TUNNEL_WAITING)
 	       {
@@ -1287,16 +1219,14 @@ int Http::Do()
 	       }
 	       if(H_CONTINUE(status_code))
 	       {
-		  xfree(status);
-		  status=0;
+		  status.set(0);
 		  status_code=0;
 		  return MOVED;
 	       }
 	       if(mode==ARRAY_INFO)
 	       {
 		  // we'll have to receive next header
-		  xfree(status);
-		  status=0;
+		  status.set(0);
 		  status_code=0;
 		  if(array_for_info[array_ptr].get_time)
 		     array_for_info[array_ptr].time=NO_DATE;
@@ -1331,7 +1261,7 @@ int Http::Do()
 		  if(H_20X(status_code))
 		  {
 		     // should never happen
-		     DebugPrint("**** ","Success, but did nothing??",0);
+		     LogError(0,"Success, but did nothing??");
 		     Disconnect();
 		     return MOVED;
 		  }
@@ -1340,21 +1270,17 @@ int Http::Do()
 	       goto pre_RECEIVING_BODY;
 	    }
 	    len=eol-buf;
-	    xfree(line);
-	    line=(char*)xmalloc(len+1);
-	    memcpy(line,buf,len);
-	    line[len]=0;
+	    line.nset(buf,len);
 
 	    recv_buf->Skip(len+eol_size);
 
-	    DebugPrint("<--- ",line,4);
+	    LogRecv(4,line);
 	    m=MOVED;
 
 	    if(status==0)
 	    {
 	       // it's status line
-	       status=line;
-	       line=0;
+	       status.set(line);
 	       int ver_major,ver_minor;
 	       if(3!=sscanf(status,"HTTP/%d.%d %n%d",&ver_major,&ver_minor,
 		     &status_consumed,&status_code))
@@ -1362,7 +1288,7 @@ int Http::Do()
 		  // simple 0.9 ?
 		  proto_version=0x09;
 		  status_code=200;
-		  DebugPrint("**** ",_("Could not parse HTTP status line"),0);
+		  LogError(0,_("Could not parse HTTP status line"));
 		  if(mode==STORE)
 		  {
 		     state=DONE;
@@ -1418,7 +1344,7 @@ int Http::Do()
 	    else
 	    {
 	       // header line.
-	       char *colon=strchr(line,':');
+	       char *colon=strchr(line.get_non_const(),':');
 	       if(colon)
 	       {
 		  *colon=0;
@@ -1494,11 +1420,10 @@ int Http::Do()
 		  char *slash=strrchr(new_location,'/');
 		  strcpy(slash+1,location);
 	       }
-	       xfree(location);
-	       location=xstrdup(new_location);
+	       location.set(new_location);
 	    }
-	    sprintf(err,"%s (%s -> %s)",status+status_consumed,file,
-				    location?location:"nowhere");
+	    sprintf(err,"%s (%s -> %s)",status+status_consumed,file.get(),
+				    location?location.get():"nowhere");
 	    code=FILE_MOVED;
 	 }
 	 else
@@ -1528,22 +1453,23 @@ int Http::Do()
 	    if(closure && closure[0])
 	       sprintf(err,"%s (%s)",status+status_consumed,closure);
 	    else
-	       sprintf(err,"%s (%s%s)",status+status_consumed,cwd.path,
+	       sprintf(err,"%s (%s%s)",status+status_consumed,cwd.path.get(),
 				       (last_char(cwd)=='/')?"":"/");
 	 }
 	 state=RECEIVING_BODY;
+	 LogErrorText();
 	 SetError(code,err);
 	 return MOVED;
       }
       if(mode==CHANGE_DIR)
       {
 	 cwd.Set(new_cwd);
+	 cache->SetDirectory(this, "", !cwd.is_file);
 	 state=DONE;
 	 return MOVED;
       }
 
-      DebugPrint("---- ",_("Receiving body..."));
-      assert(rate_limit==0);
+      LogNote(9,_("Receiving body..."));
       rate_limit=new RateLimit(hostname);
       if(real_pos<0) // assume Range: did not work
       {
@@ -1646,137 +1572,148 @@ int Http::Read(void *buf,int size)
       return 0;
    if(state==DONE)
       return 0;	  // eof
+   int res=DO_AGAIN;
    if(state==RECEIVING_BODY && real_pos>=0)
    {
-      const char *buf1;
-      int size1;
-   get_again:
-      if(recv_buf->Size()==0 && recv_buf->Error())
+      res=_Read(buf,size);
+      if(res>0)
       {
-	 DebugPrint("**** ",recv_buf->ErrorText(),0);
-	 if(recv_buf->ErrorFatal())
-	    SetError(FATAL,recv_buf->ErrorText());
+	 pos+=res;
+	 if(rate_limit)
+	    rate_limit->BytesGot(res);
+	 TrySuccess();
+      }
+   }
+   return res;
+}
+int Http::_Read(void *buf,int size)
+{
+   const char *buf1;
+   int size1;
+get_again:
+   if(recv_buf->Size()==0 && recv_buf->Error())
+   {
+      LogError(0,"recv: %s",recv_buf->ErrorText());
+      if(recv_buf->ErrorFatal())
+	 SetError(FATAL,recv_buf->ErrorText());
+      Disconnect();
+      return DO_AGAIN;
+   }
+   recv_buf->Get(&buf1,&size1);
+   if(buf1==0) // eof
+   {
+      LogNote(9,_("Hit EOF"));
+      if(bytes_received<body_size || chunked)
+      {
+	 LogError(0,_("Received not enough data, retrying"));
 	 Disconnect();
 	 return DO_AGAIN;
       }
-      recv_buf->Get(&buf1,&size1);
-      if(buf1==0) // eof
+      return 0;
+   }
+   if(!chunked)
+   {
+      if(body_size>=0 && bytes_received>=body_size)
       {
-	 DebugPrint("---- ",_("Hit EOF"));
-	 if(bytes_received<body_size || chunked)
-	 {
-	    DebugPrint("**** ",_("Received not enough data, retrying"),0);
-	    Disconnect();
-	    return DO_AGAIN;
-	 }
+	 LogNote(9,_("Received all"));
+	 return 0; // all received
+      }
+      if(entity_size>=0 && pos>=entity_size)
+      {
+	 LogNote(9,_("Received all (total)"));
 	 return 0;
       }
-      if(!chunked)
-      {
-	 if(body_size>=0 && bytes_received>=body_size)
-	 {
-	    DebugPrint("---- ",_("Received all"));
-	    return 0; // all received
-	 }
-	 if(entity_size>=0 && pos>=entity_size)
-	 {
-	    DebugPrint("---- ",_("Received all (total)"));
-	    return 0;
-	 }
-      }
-      if(size1==0)
+   }
+   if(size1==0)
+      return DO_AGAIN;
+   if(chunked)
+   {
+      if(chunked_trailer && state==RECEIVING_HEADER)
 	 return DO_AGAIN;
-      if(chunked)
+      const char *nl;
+      if(chunk_size==-1) // expecting first/next chunk
       {
-	 if(chunked_trailer && state==RECEIVING_HEADER)
-	    return DO_AGAIN;
-	 const char *nl;
-	 if(chunk_size==-1) // expecting first/next chunk
+	 nl=(const char*)memchr(buf1,'\n',size1);
+	 if(nl==0)  // not yet
 	 {
-	    nl=(const char*)memchr(buf1,'\n',size1);
-	    if(nl==0)  // not yet
-	    {
-	    not_yet:
-	       if(recv_buf->Eof())
-		  Disconnect();	 // connection closed too early
-	       return DO_AGAIN;
-	    }
-	    if(!is_ascii_xdigit(*buf1)
-	    || sscanf(buf1,"%lx",&chunk_size)!=1)
-	    {
-	       Fatal(_("chunked format violated"));
-	       return FATAL;
-	    }
-	    recv_buf->Skip(nl-buf1+1);
-	    chunk_pos=0;
-	    goto get_again;
-	 }
-	 if(chunk_size==0) // eof
-	 {
-	    DebugPrint("---- ",_("Received last chunk"));
-	    // headers may follow
-	    chunked_trailer=true;
-	    state=RECEIVING_HEADER;
-	    body_size=bytes_received;
+	 not_yet:
+	    if(recv_buf->Eof())
+	       Disconnect();	 // connection closed too early
 	    return DO_AGAIN;
 	 }
-	 if(chunk_pos==chunk_size)
+	 if(!is_ascii_xdigit(*buf1)
+	 || sscanf(buf1,"%lx",&chunk_size)!=1)
 	 {
-	    if(size1<2)
-	       goto not_yet;
-	    if(buf1[0]!='\r' || buf1[1]!='\n')
-	    {
-	       Fatal(_("chunked format violated"));
-	       return FATAL;
-	    }
-	    recv_buf->Skip(2);
-	    chunk_size=-1;
-	    goto get_again;
+	    Fatal(_("chunked format violated"));
+	    return FATAL;
 	 }
-	 // ok, now we may get portion of data
-	 if(size1>chunk_size-chunk_pos)
-	    size1=chunk_size-chunk_pos;
-      }
-      else
-      {
-	 // limit by body_size.
-	 if(body_size>=0 && size1+bytes_received>=body_size)
-	    size1=body_size-bytes_received;
-      }
-
-      int bytes_allowed=rate_limit->BytesAllowedToGet();
-      if(size1>bytes_allowed)
-	 size1=bytes_allowed;
-      if(size1==0)
-	 return DO_AGAIN;
-      if(norest_manual && real_pos==0 && pos>0)
-	 return DO_AGAIN;
-      if(real_pos<pos)
-      {
-	 off_t to_skip=pos-real_pos;
-	 if(to_skip>size1)
-	    to_skip=size1;
-	 recv_buf->Skip(to_skip);
-	 real_pos+=to_skip;
-	 bytes_received+=to_skip;
-	 if(chunked)
-	    chunk_pos+=to_skip;
+	 recv_buf->Skip(nl-buf1+1);
+	 chunk_pos=0;
 	 goto get_again;
       }
-      if(size>size1)
-	 size=size1;
-      memcpy(buf,buf1,size);
-      recv_buf->Skip(size);
-      pos+=size;
-      real_pos+=size;
-      bytes_received+=size;
-      if(chunked)
-	 chunk_pos+=size;
-      rate_limit->BytesGot(size);
-      TrySuccess();
-      return size;
+      if(chunk_size==0) // eof
+      {
+	 LogNote(9,_("Received last chunk"));
+	 // headers may follow
+	 chunked_trailer=true;
+	 state=RECEIVING_HEADER;
+	 body_size=bytes_received;
+	 return DO_AGAIN;
+      }
+      if(chunk_pos==chunk_size)
+      {
+	 if(size1<2)
+	    goto not_yet;
+	 if(buf1[0]!='\r' || buf1[1]!='\n')
+	 {
+	    Fatal(_("chunked format violated"));
+	    return FATAL;
+	 }
+	 recv_buf->Skip(2);
+	 chunk_size=-1;
+	 goto get_again;
+      }
+      // ok, now we may get portion of data
+      if(size1>chunk_size-chunk_pos)
+	 size1=chunk_size-chunk_pos;
    }
-   return DO_AGAIN;
+   else
+   {
+      // limit by body_size.
+      if(body_size>=0 && size1+bytes_received>=body_size)
+	 size1=body_size-bytes_received;
+   }
+
+   int bytes_allowed=0x10000000;
+   if(rate_limit)
+      bytes_allowed=rate_limit->BytesAllowedToGet();
+   if(size1>bytes_allowed)
+      size1=bytes_allowed;
+   if(size1==0)
+      return DO_AGAIN;
+   if(norest_manual && real_pos==0 && pos>0)
+      return DO_AGAIN;
+   if(real_pos<pos)
+   {
+      off_t to_skip=pos-real_pos;
+      if(to_skip>size1)
+	 to_skip=size1;
+      recv_buf->Skip(to_skip);
+      real_pos+=to_skip;
+      bytes_received+=to_skip;
+      if(chunked)
+	 chunk_pos+=to_skip;
+      goto get_again;
+   }
+   if(size>size1)
+      size=size1;
+   memcpy(buf,buf1,size);
+   recv_buf->Skip(size);
+   if(chunked)
+      chunk_pos+=size;
+   real_pos+=size;
+   bytes_received+=size;
+   return size;
 }
 
 int Http::Done()
@@ -1941,7 +1878,7 @@ void Http::Reconfig(const char *name)
    if(sock!=-1)
       SetSocketBuffer(sock);
    if(proxy && proxy_port==0)
-      proxy_port=xstrdup(HTTP_DEFAULT_PROXY_PORT);
+      proxy_port.set(HTTP_DEFAULT_PROXY_PORT);
 
    user_agent=ResMgr::Query("http:user-agent",c);
    use_propfind_now=(use_propfind_now && QueryBool("use-propfind",c));
@@ -1976,7 +1913,7 @@ void Http::ResetLocationData()
 
 DirList *Http::MakeDirList(ArgV *args)
 {
-   return new HttpDirList(args,this);
+   return new HttpDirList(this,args);
 }
 #include "FileGlob.h"
 Glob *Http::MakeGlob(const char *pattern)
@@ -2026,13 +1963,8 @@ bool Http::CookieClosureMatch(const char *closure_c,
    return false;
 }
 
-void Http::CookieMerge(char **all_p,const char *cookie_c)
+void Http::CookieMerge(xstring &all,const char *cookie_c)
 {
-   char *all=*all_p;
-   int all_len=xstrlen(all);
-   all=*all_p=(char*)xrealloc(all,all_len+2+xstrlen(cookie_c)+1);
-   all[all_len]=0;   // in case the buffer is freshly allocated.
-
    char *value=alloca_strdup(cookie_c);
 
    for(char *entry=strtok(value,";"); entry; entry=strtok(0,";"))
@@ -2056,7 +1988,7 @@ void Http::CookieMerge(char **all_p,const char *cookie_c)
 	 c_value=c_name, c_name=0;
       int c_name_len=xstrlen(c_name);
 
-      char *scan=all;
+      char *scan=all.get_non_const();
       for(;;)
       {
 	 while(*scan==' ') scan++;
@@ -2093,21 +2025,21 @@ void Http::CookieMerge(char **all_p,const char *cookie_c)
       // append cookie.
       int c_len=strlen(all);
       while(c_len>0 && all[c_len-1]==' ')
-	 all[--c_len]=0;  // trim
+	 c_len--; // trim
+      all.truncate(c_len);
       if(c_len>0 && all[c_len-1]!=';')
-	 all[c_len++]=';', all[c_len++]=' ';  // append '; '
+	 all.append("; ");
       if(c_name)
-	 sprintf(all+c_len,"%s=%s",c_name,c_value);
+	 all.vappend(c_name,"=",c_value,NULL);
       else
-	 strcpy(all+c_len,c_value);
+	 all.append(c_value);
    }
 }
 
-char *Http::MakeCookie(const char *hostname,const char *efile)
+void Http::MakeCookie(xstring &all_cookies,const char *hostname,const char *efile)
 {
    ResMgr::Resource *scan=0;
    const char *closure;
-   char *all_cookies=0;
    for(;;)
    {
       const char *cookie=ResMgr::QueryNext("http:cookie",&closure,&scan);
@@ -2115,9 +2047,8 @@ char *Http::MakeCookie(const char *hostname,const char *efile)
 	 break;
       if(!CookieClosureMatch(closure,hostname,efile))
 	 continue;
-      CookieMerge(&all_cookies,cookie);
+      CookieMerge(all_cookies,cookie);
    }
-   return all_cookies;
 }
 
 void Http::SetCookie(const char *value_const)
@@ -2175,14 +2106,9 @@ void Http::SetCookie(const char *value_const)
    if(secure)
       strcat(closure,";secure");
 
-   const char *old=Query("cookie",closure);
-
-   char *c=xstrdup(old,2+strlen(value_const));
-   CookieMerge(&c,value_const);
-
+   xstring c(Query("cookie",closure));
+   CookieMerge(c,value_const);
    ResMgr::Set("http:cookie",closure,c);
-
-   xfree(c);
 }
 
 #if USE_SSL
@@ -2206,14 +2132,10 @@ FileAccess *Https::New(){ return new Https();}
 
 void Http::MakeSSLBuffers()
 {
-   Delete(send_buf);
-   Delete(recv_buf);
-
-   lftp_ssl *ssl=new lftp_ssl(sock,lftp_ssl::CLIENT,hostname);
+   ssl=new lftp_ssl(sock,lftp_ssl::CLIENT,hostname);
    ssl->load_keys();
    IOBufferSSL *send_buf_ssl=new IOBufferSSL(ssl,IOBuffer::PUT);
    IOBufferSSL *recv_buf_ssl=new IOBufferSSL(ssl,IOBuffer::GET);
-   recv_buf_ssl->CloseLater();
    send_buf=send_buf_ssl;
    recv_buf=recv_buf_ssl;
 }
@@ -2267,6 +2189,24 @@ void Http::CleanupThis()
    Disconnect();
 }
 
+void Http::LogErrorText()
+{
+   if(!recv_buf)
+      return;
+   recv_buf->Roll();
+   size_t size=recv_buf->Size();
+   if(size==0)
+      return;
+   char *buf=string_alloca(size+1);
+   size=_Read(buf,size);
+   if(size<0)
+      return;
+   buf[size]=0;
+   remove_tags(buf);
+   for(char *line=strtok(buf,"\n"); line; line=strtok(0,"\n"))
+      if(*line)
+	 Log::global->Format(4,"<--* %s\n",line);
+}
 
 
 /* The functions http_atotm and check_end are taken from wget */
