@@ -18,15 +18,17 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: LocalDir.h,v 1.2 2002/01/29 07:20:26 lav Exp $ */
+/* $Id: LocalDir.h,v 1.3 2007/04/11 17:50:01 lav Exp $ */
 
 #ifndef LOCALDIR_H
 #define LOCALDIR_H
 
+#include "xstring.h"
+
 class LocalDirectory
 {
    int fd;
-   char *name;
+   xstring_c name;
 
 public:
    LocalDirectory();
