@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2005 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2007 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,20 +18,20 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: MirrorJob.cc,v 1.139.2.1 2007/01/06 06:39:33 lav Exp $ */
+/* $Id: MirrorJob.cc,v 1.161 2008/02/26 15:46:57 lav Exp $ */
 
 #include <config.h>
 
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <errno.h>
+#include <fcntl.h>
 #include <unistd.h>
 #include <assert.h>
 #include <mbswidth.h>
 #include "MirrorJob.h"
 #include "CmdExec.h"
 #include "rmJob.h"
-#include "mkdirJob.h"
 #include "ChmodJob.h"
 #include "misc.h"
 #include "plural.h"
@@ -43,6 +43,7 @@
 
 #define set_state(s) do { state=(s); \
    Log::global->Format(11,"mirror(%p) enters state %s\n", this, #s); } while(0)
+#define waiting_num waiting.count()
 
 void  MirrorJob::PrintStatus(int v,const char *tab)
 {
@@ -58,32 +59,33 @@ void  MirrorJob::PrintStatus(int v,const char *tab)
    case(TARGET_REMOVE_OLD):
    case(TARGET_REMOVE_OLD_FIRST):
    case(TARGET_CHMOD):
+   case(LAST_EXEC):
       break;
 
    case(MAKE_TARGET_DIR):
-      printf("\tmkdir `%s' [%s]\n",target_dir,target_session->CurrentStatus());
+      printf("\tmkdir `%s' [%s]\n",target_dir.get(),target_session->CurrentStatus());
       break;
 
    case(CHANGING_DIR_SOURCE):
    case(CHANGING_DIR_TARGET):
       if(target_session->IsOpen())
-	 printf("\tcd `%s' [%s]\n",target_dir,target_session->CurrentStatus());
+	 printf("\tcd `%s' [%s]\n",target_dir.get(),target_session->CurrentStatus());
       if(source_session->IsOpen())
-	 printf("\tcd `%s' [%s]\n",source_dir,source_session->CurrentStatus());
+	 printf("\tcd `%s' [%s]\n",source_dir.get(),source_session->CurrentStatus());
       break;
 
    case(GETTING_LIST_INFO):
       if(target_list_info)
       {
 	 if(target_relative_dir)
-	    printf("\t%s: %s\n",target_relative_dir,target_list_info->Status());
+	    printf("\t%s: %s\n",target_relative_dir.get(),target_list_info->Status());
 	 else
 	    printf("\t%s\n",target_list_info->Status());
       }
       if(source_list_info)
       {
 	 if(source_relative_dir)
-	    printf("\t%s: %s\n",source_relative_dir,source_list_info->Status());
+	    printf("\t%s: %s\n",source_relative_dir.get(),source_list_info->Status());
 	 else
 	    printf("\t%s\n",source_list_info->Status());
       }
@@ -117,7 +119,7 @@ final:
 	       tab,stats.error_count);
 }
 
-void  MirrorJob::ShowRunStatus(StatusLine *s)
+void  MirrorJob::ShowRunStatus(const SMTaskRef<StatusLine>& s)
 {
    int w=s->GetWidthDelayed();
    switch(state)
@@ -132,19 +134,20 @@ void  MirrorJob::ShowRunStatus(StatusLine *s)
    case(TARGET_CHMOD):
    case(FINISHING):
    case(DONE):
+   case(LAST_EXEC):
       Job::ShowRunStatus(s);
       break;
 
    case(MAKE_TARGET_DIR):
-      s->Show("mkdir `%s' [%s]",target_dir,target_session->CurrentStatus());
+      s->Show("mkdir `%s' [%s]",target_dir.get(),target_session->CurrentStatus());
       break;
 
    case(CHANGING_DIR_SOURCE):
    case(CHANGING_DIR_TARGET):
       if(target_session->IsOpen() && (!source_session->IsOpen() || now%4>=2))
-	 s->Show("cd `%s' [%s]",target_dir,target_session->CurrentStatus());
+	 s->Show("cd `%s' [%s]",target_dir.get(),target_session->CurrentStatus());
       else if(source_session->IsOpen())
-	 s->Show("cd `%s' [%s]",source_dir,source_session->CurrentStatus());
+	 s->Show("cd `%s' [%s]",source_dir.get(),source_session->CurrentStatus());
       break;
 
    case(GETTING_LIST_INFO):
@@ -228,7 +231,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    && (old->defined&file->TYPE) && old->filetype==old->NORMAL
 	    && (flags&IGNORE_TIME ||
 	    	((file->defined&file->DATE) && (old->defined&old->DATE)
-	    	&& file->date + file->date_prec < old->date - old->date_prec))
+	    	&& file->date + file->date.ts_prec < old->date - old->date.ts_prec))
 	    && (file->defined&file->SIZE) && (old->defined&old->SIZE)
 	    && file->size >= old->size
 	    && !use_pget)
@@ -254,6 +257,12 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    else
 	       stats.new_files++;
 	 }
+	 else if(flags&ONLY_EXISTING)
+	 {
+	    Report(_("Skipping file `%s' (only-existing)"),
+		     dir_file(source_relative_dir,file->name));
+	    goto skip;
+	 }
 	 else
 	    stats.new_files++;
 
@@ -272,56 +281,47 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	       args.Append("-c");
 	    if(remove_target)
 	       args.Append("-e");
+	    if(FlagSet(ASCII))
+	       args.Append("-a");
 	    args.Append("-O");
-	    args.Append(target_is_local?target_dir
+	    args.Append(target_is_local?target_dir.get()
 			:target_session->GetConnectURL());
 	    args.Append(source_session->GetFileURL(file->name));
-	    char *cmd=args.CombineQuoted();
-	    fprintf(script,"%s\n",cmd);
-	    xfree(cmd);
+	    xstring_ca cmd(args.CombineQuoted());
+	    fprintf(script,"%s\n",cmd.get());
 	    if(script_only)
 	       goto skip;
 	 }
 
-	 if(!use_pget)
-	 {
-	    FileCopyPeerFA *src_peer=
-	       new FileCopyPeerFA(source_session->Clone(),file->name,FA::RETRIEVE);
-	    FileCopyPeerFA *dst_peer=
-	       new FileCopyPeerFA(target_session->Clone(),file->name,FA::STORE);
+	 FileCopyPeer *src_peer=0;
+	 if(source_is_local)
+	    src_peer=new FileCopyPeerFDStream(new FileStream(source_name,O_RDONLY),FileCopyPeer::GET);
+	 else
+	    src_peer=new FileCopyPeerFA(source_session->Clone(),file->name,FA::RETRIEVE);
+
+	 FileCopyPeer *dst_peer=0;
+	 if(target_is_local)
+	    dst_peer=new FileCopyPeerFDStream(new FileStream(target_name,O_WRONLY|O_CREAT|(cont_this?0:O_TRUNC)),FileCopyPeer::PUT);
+	 else
+	    dst_peer=new FileCopyPeerFA(target_session,file->name,FA::STORE);
+
+	 FileCopy *c=FileCopy::New(src_peer,dst_peer,cont_this);
+	 if(remove_source_files)
+	    c->RemoveSourceLater();
+	 if(remove_target)
+	    c->RemoveTargetFirst();
+	 if(FlagSet(ASCII))
+	    c->Ascii();
+	 CopyJob *cp=(use_pget ? new pgetJob(c,file->name,pget_n) : new CopyJob(c,file->name,"mirror"));
+	 if(file->defined&file->DATE)
+	    cp->SetDate(file->date);
+	 if(file->defined&file->SIZE)
+	    cp->SetSize(file->size);
+	 AddWaiting(cp);
+	 transfer_count++;
+	 cp->SetParentFg(this);
+	 cp->cmdline.vset("\\transfer ",file->name.get(),NULL);
 
-	    FileCopy *c=FileCopy::New(src_peer,dst_peer,cont_this);
-	    if(remove_source_files)
-	       c->RemoveSourceLater();
-	    if(remove_target)
-	       c->RemoveTargetFirst();
-	    CopyJob *cp=
-	       new CopyJob(c,file->name,"mirror");
-	    if((file->defined&file->DATE) && file->date_prec<=1)
-	       cp->SetDate(file->date);
-	    if(file->defined&file->SIZE)
-	       cp->SetSize(file->size);
-	    AddWaiting(cp);
-	    transfer_count++;
-	    cp->SetParentFg(this);
-	    cp->cmdline=xasprintf("\\transfer %s",file->name);
-	 }
-	 else // pget
-	 {
-	    ArgV *get_args=new ArgV("mirror");
-	    get_args->Append(file->name);
-	    get_args->Append(dir_file(target_dir,file->name));
-	    pgetJob *cp=new pgetJob(source_session->Clone(),get_args,cont_this);
-	    cp->SetMaxConn(pget_n);
-	    if(remove_source_files)
-	       cp->RemoveSourceLater();
-	    if(remove_target)
-	       cp->RemoveTargetFirst();
-	    AddWaiting(cp);
-	    transfer_count++;
-	    cp->SetParentFg(this);
-	    cp->cmdline=xasprintf("\\transfer %s",file->name);
-	 }
 	 set_state(WAITING_FOR_TRANSFER);
 	 break;
       }
@@ -331,9 +331,22 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    goto skip;
 
 	 bool create_target_dir=true;
-	 FileInfo *old=target_set->FindByName(file->name);
-	 if(old && old->defined&old->TYPE && old->filetype==old->DIRECTORY)
+	 const FileInfo *old=0;
+	 if(target_set)
+	    old=target_set->FindByName(file->name);
+	 if(!old)
+	 {
+	    if(flags&ONLY_EXISTING)
+	    {
+	       Report(_("Skipping directory `%s' (only-existing)"),
+			dir_file(target_relative_dir,file->name));
+	       goto skip;
+	    }
+	 }
+	 else if(old->defined&old->TYPE && old->filetype==old->DIRECTORY)
+	 {
 	    create_target_dir=false;
+	 }
 	 if(target_is_local && !script_only)
 	 {
 	    if(lstat(target_name,&st)!=-1)
@@ -363,7 +376,7 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    source_name,target_name);
 	 AddWaiting(mj);
 	 mj->SetParentFg(this);
-	 mj->cmdline=xasprintf("\\mirror %s",file->name);
+	 mj->cmdline.vset("\\mirror ",file->name.get(),NULL);
 
 	 // inherit flags and other things
 	 mj->SetFlags(flags,1);
@@ -371,10 +384,8 @@ void  MirrorJob::HandleFile(FileInfo *file)
 
 	 mj->SetExclude(exclude);
 
-	 mj->source_relative_dir=
-	       xstrdup(dir_file(source_relative_dir,file->name));
-	 mj->target_relative_dir=
-	       xstrdup(dir_file(target_relative_dir,file->name));
+	 mj->source_relative_dir.set(dir_file(source_relative_dir,file->name));
+	 mj->target_relative_dir.set(dir_file(target_relative_dir,file->name));
 
 	 mj->verbose_report=verbose_report;
 	 mj->newer_than=newer_than;
@@ -390,13 +401,13 @@ void  MirrorJob::HandleFile(FileInfo *file)
 
 	 mj->script=script;
 	 mj->script_needs_closing=false;
-	 mj->script_name=xstrdup(script_name);
+	 mj->script_name.set(script_name);
 	 mj->script_only=script_only;
 
 	 mj->max_error_count=max_error_count;
 
 	 if(verbose_report>=3)
-	    Report(_("Mirroring directory `%s'"),mj->target_relative_dir);
+	    Report(_("Mirroring directory `%s'"),mj->target_relative_dir.get());
 
 	 break;
       }
@@ -421,9 +432,8 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    args.Append("-sf");
 	    args.Append(shell_encode(file->symlink));
 	    args.Append(shell_encode(target_name));
-	    char *cmd=args.CombineQuoted();
-	    fprintf(script,"%s\n",cmd);
-	    xfree(cmd);
+	    xstring_ca cmd(args.CombineQuoted());
+	    fprintf(script,"%s\n",cmd.get());
 	    if(script_only)
 	       goto skip;
 	 }
@@ -445,10 +455,16 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    }
 	    else
 	    {
+	       if(flags&ONLY_EXISTING)
+	       {
+		  Report(_("Skipping symlink `%s' (only-existing)"),
+			   dir_file(target_relative_dir,file->name));
+		  goto skip;
+	       }
 	       stats.new_symlinks++;
 	    }
 	    Report(_("Making symbolic link `%s' to `%s'"),
-		     dir_file(target_relative_dir,file->name),file->symlink);
+		     dir_file(target_relative_dir,file->name),file->symlink.get());
 	    res=symlink(file->symlink,target_name);
 	    if(res==-1)
 	       eprintf("mirror: symlink(%s): %s\n",target_name,strerror(errno));
@@ -463,7 +479,7 @@ skip:
    return;
 }
 
-void  MirrorJob::InitSets(FileSet *source,FileSet *dest)
+void  MirrorJob::InitSets(const FileSet *source,const FileSet *dest)
 {
    source->Count(NULL,&stats.tot_files,&stats.tot_symlinks,&stats.tot_files);
 
@@ -486,10 +502,10 @@ void  MirrorJob::InitSets(FileSet *source,FileSet *dest)
 
    same->SubtractAny(to_transfer);
 
-   if(newer_than!=(time_t)-1)
-      to_transfer->SubtractOlderThan(newer_than);
-   if(older_than!=(time_t)-1)
-      to_transfer->SubtractNewerThan(older_than);
+   if(newer_than!=NO_DATE)
+      to_transfer->SubtractNotNewerThan(newer_than);
+   if(older_than!=NO_DATE)
+      to_transfer->SubtractNotOlderThan(older_than);
    if(size_range)
       to_transfer->SubtractSizeOutside(size_range);
 
@@ -514,7 +530,7 @@ void  MirrorJob::InitSets(FileSet *source,FileSet *dest)
    to_transfer->SortByPatternList(ResMgr::Query("mirror:order",0));
 }
 
-void MirrorJob::HandleChdir(FileAccess * &session, int &redirections)
+void MirrorJob::HandleChdir(FileAccessRef& session, int &redirections)
 {
    if(!session->IsOpen())
       return;
@@ -543,7 +559,6 @@ void MirrorJob::HandleChdir(FileAccess * &session, int &redirections)
 	       session->Chdir(loc);
 	       return;
 	    }
-	    SessionPool::Reuse(session);
 	    session=FA::New(&u);
 	    session->Chdir(u.path);
 	    return;
@@ -569,7 +584,7 @@ void MirrorJob::HandleChdir(FileAccess * &session, int &redirections)
    if(res==FA::OK)
       session->Close();
 }
-void MirrorJob::HandleListInfoCreation(FileAccess * &session,ListInfo * &list_info,const char *relative_dir)
+void MirrorJob::HandleListInfoCreation(const FileAccessRef& session,Ref<ListInfo>& list_info,const char *relative_dir)
 {
    if(state!=GETTING_LIST_INFO)
       return;
@@ -600,10 +615,10 @@ void MirrorJob::HandleListInfoCreation(FileAccess * &session,ListInfo * &list_in
       list_info->FollowSymlinks();
 
    list_info->SetExclude(relative_dir,exclude);
-   Roll(list_info);
+   list_info->Roll();
 }
 
-void MirrorJob::HandleListInfo(ListInfo * &list_info, FileSet * &set)
+void MirrorJob::HandleListInfo(Ref<ListInfo>& list_info, Ref<FileSet>& set)
 {
    if(!list_info)
       return;
@@ -615,14 +630,11 @@ void MirrorJob::HandleListInfo(ListInfo * &list_info, FileSet * &set)
       stats.error_count++;
       transfer_count-=root_transfer_count;
       set_state(FINISHING);
-      Delete(source_list_info);
       source_list_info=0;
-      Delete(target_list_info);
       target_list_info=0;
       return;
    }
    set=list_info->GetResult();
-   Delete(list_info);
    list_info=0;
    set->ExcludeDots(); // don't need .. and .
 }
@@ -639,7 +651,7 @@ int   MirrorJob::Do()
    case(INITIAL_STATE):
       source_session->Chdir(source_dir);
       source_redirections=0;
-      Roll(source_session);
+      source_session->Roll();
       set_state(CHANGING_DIR_SOURCE);
       m=MOVED;
       /*fallthrough*/
@@ -650,10 +662,13 @@ int   MirrorJob::Do()
       if(source_session->IsOpen())
 	 return m;
 
-      xfree(source_dir);
-      source_dir=xstrdup(source_session->GetCwd());
+      source_dir.set(source_session->GetCwd());
 
-      if(!create_target_dir || !strcmp(target_dir,".") || !strcmp(target_dir,".."))
+   pre_MAKE_TARGET_DIR:
+   {
+      if(!strcmp(target_dir,".") || !strcmp(target_dir,".."))
+	 create_target_dir=false;
+      if(!create_target_dir)
 	 goto pre_CHANGING_DIR_TARGET;
       if(target_is_local)
       {
@@ -664,55 +679,61 @@ int   MirrorJob::Do()
 	    {
 	       if(!script_only)
 		  chmod(target_dir,st.st_mode|0700);
+	       create_target_dir=false;
 	       goto pre_CHANGING_DIR_TARGET;
 	    }
 	    else
 	    {
-	       Report(_("Removing old local file `%s'"),target_dir);
+	       Report(_("Removing old local file `%s'"),target_dir.get());
 	       if(script)
 	       {
 		  ArgV args("rm");
 		  args.Append(target_session->GetFileURL(target_dir));
-		  char *cmd=args.CombineQuoted();
-		  fprintf(script,"%s\n",cmd);
-		  xfree(cmd);
+		  xstring_ca cmd(args.CombineQuoted());
+		  fprintf(script,"%s\n",cmd.get());
 	       }
 	       if(!script_only)
 	       {
 		  if(remove(target_dir)==-1)
-		     eprintf("mirror: remove(%s): %s\n",target_dir,strerror(errno));
+		     eprintf("mirror: remove(%s): %s\n",target_dir.get(),strerror(errno));
 	       }
 	    }
 	 }
       }
+
+      if(FlagSet(DEPTH_FIRST))
+	 goto pre_GETTING_LIST_INFO;
+
       if(target_relative_dir)
-	 Report(_("Making directory `%s'"),target_relative_dir);
+	 Report(_("Making directory `%s'"),target_relative_dir.get());
+      bool mkdir_p=(parent_mirror==0 || parent_mirror->create_target_dir);
       if(script)
       {
 	 ArgV args("mkdir");
-	 if(parent_mirror==0)
+	 if(mkdir_p)
 	    args.Append("-p");
 	 args.Append(target_session->GetFileURL(target_dir));
-	 char *cmd=args.CombineQuoted();
-	 fprintf(script,"%s\n",cmd);
-	 xfree(cmd);
+	 xstring_ca cmd(args.CombineQuoted());
+	 fprintf(script,"%s\n",cmd.get());
 	 if(script_only)
 	    goto pre_CHANGING_DIR_TARGET;
       }
-      target_session->Mkdir(target_dir,(parent_mirror==0));
+      target_session->Mkdir(target_dir,mkdir_p);
       set_state(MAKE_TARGET_DIR);
       m=MOVED;
+   }
       /*fallthrough*/
    case(MAKE_TARGET_DIR):
       res=target_session->Done();
       if(res==FA::IN_PROGRESS)
 	 return m;
       target_session->Close();
+      create_target_dir=false;
 
    pre_CHANGING_DIR_TARGET:
       target_session->Chdir(target_dir);
       target_redirections=0;
-      Roll(target_session);
+      target_session->Roll();
       set_state(CHANGING_DIR_TARGET);
       m=MOVED;
       /*fallthrough*/
@@ -722,19 +743,21 @@ int   MirrorJob::Do()
 	 return MOVED;
       if(target_session->IsOpen())
 	 return m;
+      create_target_dir=false;
 
-      xfree(target_dir);
-      target_dir=xstrdup(target_session->GetCwd());
+      target_dir.set(target_session->GetCwd());
 
    pre_GETTING_LIST_INFO:
       set_state(GETTING_LIST_INFO);
       m=MOVED;
-      HandleListInfoCreation(source_session,source_list_info,source_relative_dir);
-      HandleListInfoCreation(target_session,target_list_info,target_relative_dir);
+      if(!source_set)
+	 HandleListInfoCreation(source_session,source_list_info,source_relative_dir);
+      if(!target_set && !create_target_dir && (!FlagSet(DEPTH_FIRST) || FlagSet(ONLY_EXISTING)))
+	 HandleListInfoCreation(target_session,target_list_info,target_relative_dir);
       if(state!=GETTING_LIST_INFO)
       {
-	 Delete(source_list_info); source_list_info=0;
-	 Delete(target_list_info); target_list_info=0;
+	 source_list_info=0;
+	 target_list_info=0;
       }
       return m;	  // give time to other tasks
    case(GETTING_LIST_INFO):
@@ -745,6 +768,16 @@ int   MirrorJob::Do()
       if(source_list_info || target_list_info)
 	 return m;
 
+      transfer_count-=root_transfer_count; // leave room for transfers.
+
+      if(FlagSet(DEPTH_FIRST) && source_set && !target_set)
+      {
+	 // transfer directories first
+	 to_transfer=new FileSet(source_set);
+	 to_transfer->SubtractNotDirs();
+	 goto pre_WAITING_FOR_TRANSFER;
+      }
+
       // now we have both target and source file sets.
       stats.dirs++;
 
@@ -753,7 +786,6 @@ int   MirrorJob::Do()
       to_rm->rewind();
       to_rm_mismatched->Count(&stats.del_dirs,&stats.del_files,&stats.del_symlinks,&stats.del_files);
       to_rm_mismatched->rewind();
-      transfer_count-=root_transfer_count; // leave room for transfers.
       set_state(TARGET_REMOVE_OLD_FIRST);
       goto TARGET_REMOVE_OLD_FIRST_label;
 
@@ -783,6 +815,21 @@ int   MirrorJob::Do()
 	 {
 	    if(waiting_num>0)
 	       break;
+	    if(FlagSet(DEPTH_FIRST))
+	    {
+	       // we have been in the depth, don't go there again
+	       SetFlags(DEPTH_FIRST,false);
+	       SetFlags(NO_RECURSION,true);
+
+	       // if we have not created any subdirs and there are only subdirs,
+	       // then the directory would be empty - skip it.
+	       if(FlagSet(NO_EMPTY_DIRS) && stats.dirs==0
+	       && source_set->get_fnum()==to_transfer->get_fnum())
+		  goto pre_FINISHING;
+
+	       transfer_count+=root_transfer_count;
+	       goto pre_MAKE_TARGET_DIR;
+	    }
 	    goto pre_TARGET_REMOVE_OLD;
 	 }
 	 HandleFile(file);
@@ -802,6 +849,8 @@ int   MirrorJob::Do()
    TARGET_REMOVE_OLD_FIRST_label:
       while((j=FindDoneAwaitedJob())!=0)
       {
+	 if(j->ExitCode()!=0)
+	    stats.error_count++;
 	 RemoveWaiting(j);
 	 Delete(j);
 	 transfer_count--;
@@ -854,9 +903,8 @@ int   MirrorJob::Do()
 		  args.Append("-r");
 	    }
 	    args.Append(target_session->GetFileURL(file->name));
-	    char *cmd=args.CombineQuoted();
-	    fprintf(script,"%s\n",cmd);
-	    xfree(cmd);
+	    xstring_ca cmd(args.CombineQuoted());
+	    fprintf(script,"%s\n",cmd.get());
 	 }
 	 if(!script_only)
 	 {
@@ -865,7 +913,7 @@ int   MirrorJob::Do()
 	    args->seek(1);
 	    rmJob *j=new rmJob(target_session->Clone(),args);
 	    j->SetParentFg(this);
-	    j->cmdline=args->Combine();
+	    j->cmdline.set_allocated(args->Combine());
 	    AddWaiting(j);
 	    transfer_count++;
 	    if(file->defined&file->TYPE && file->filetype==file->DIRECTORY)
@@ -941,9 +989,8 @@ int   MirrorJob::Do()
 	    sprintf(m,"%03lo",(unsigned long)(file->mode&~mode_mask));
 	    args.Append(m);
 	    args.Append(target_session->GetFileURL(file->name));
-	    char *cmd=args.CombineQuoted();
-	    fprintf(script,"%s\n",cmd);
-	    xfree(cmd);
+	    xstring_ca cmd(args.CombineQuoted());
+	    fprintf(script,"%s\n",cmd.get());
 	 }
 	 if(!script_only)
 	 {
@@ -955,7 +1002,7 @@ int   MirrorJob::Do()
 	    AddWaiting(cj);
 	    transfer_count++;
 	    cj->SetParentFg(this);
-	    cj->cmdline=a->Combine();
+	    cj->cmdline.set_allocated(a->Combine());
 	    cj->BeQuiet(); // chmod is not supported on all servers; be quiet.
 	    m=MOVED;
 	 }
@@ -968,9 +1015,9 @@ int   MirrorJob::Do()
 	 to_transfer->LocalUtime(target_dir,/*only_dirs=*/true);
 	 if(flags&ALLOW_CHOWN)
 	    to_transfer->LocalChown(target_dir);
-	 if(!(flags&NO_PERMS))
+	 if(!FlagSet(NO_PERMS) && same)
 	    same->LocalChmod(target_dir,get_mode_mask());
-	 if(flags&ALLOW_CHOWN)
+	 if(FlagSet(ALLOW_CHOWN) && same)
 	    same->LocalChown(target_dir);
       }
       set_state(FINISHING);
@@ -992,24 +1039,44 @@ int   MirrorJob::Do()
 	 parent_mirror->stats.Add(stats);
       else
       {
-	 if((flags&LOOP) && stats.HaveSomethingDone(flags) && !stats.error_count)
+	 if(stats.HaveSomethingDone(flags) && on_change)
 	 {
-	    set_state(DONE);
-	    PrintStatus(0,"");
-	    printf(_("Retrying mirror...\n"));
-	    stats.Reset();
-	    goto pre_GETTING_LIST_INFO;
+	    CmdExec *exec=new CmdExec(source_session->Clone(),0);
+	    exec->SetParentFg(this);
+	    AddWaiting(exec);
+	    exec->FeedCmd(on_change);
+	    exec->FeedCmd("\n");
+	    set_state(LAST_EXEC);
+	    break;
 	 }
       }
+      goto pre_DONE;
+   case(LAST_EXEC):
+      while((j=FindDoneAwaitedJob())!=0)
+      {
+	 RemoveWaiting(j);
+	 Delete(j);
+	 m=MOVED;
+      }
+      if(waiting_num>0)
+	 break;
+   pre_DONE:
       set_state(DONE);
       m=MOVED;
+      if(!parent_mirror && (flags&LOOP) && stats.HaveSomethingDone(flags) && !stats.error_count)
+      {
+	 PrintStatus(0,"");
+	 printf(_("Retrying mirror...\n"));
+	 stats.Reset();
+	 goto pre_GETTING_LIST_INFO;
+      }
       /*fallthrough*/
    case(DONE):
-      return m;
+      break;
    }
    // give direct parent priority over grand-parents.
    if(transfer_count<parallel && parent_mirror)
-      m|=Roll(parent_mirror);
+      m|=parent_mirror->Roll();
    return m;
 }
 
@@ -1017,6 +1084,7 @@ MirrorJob::MirrorJob(MirrorJob *parent,
    FileAccess *source,FileAccess *target,
    const char *new_source_dir,const char *new_target_dir)
  :
+   source_dir(new_source_dir), target_dir(new_target_dir),
    root_transfer_count(0),
    transfer_count(parent?parent->transfer_count:root_transfer_count)
 {
@@ -1029,18 +1097,8 @@ MirrorJob::MirrorJob(MirrorJob *parent,
    source_is_local=!strcmp(source_session->GetProto(),"file");
    target_is_local=!strcmp(target_session->GetProto(),"file");
 
-   source_dir=xstrdup(new_source_dir);
-   target_dir=xstrdup(new_target_dir);
-   source_relative_dir=0;
-   target_relative_dir=0;
-
-   to_transfer=to_rm=to_rm_mismatched=same=0;
-   source_set=target_set=0;
-   new_files_set=old_files_set=0;
    create_target_dir=true;
    no_target_dir=false;
-   source_list_info=0;
-   target_list_info=0;
 
    flags=0;
    max_error_count=0;
@@ -1049,11 +1107,10 @@ MirrorJob::MirrorJob(MirrorJob *parent,
 
    set_state(INITIAL_STATE);
 
-   newer_than=(time_t)-1;
-   older_than=(time_t)-1;
+   newer_than=NO_DATE;
+   older_than=NO_DATE;
    size_range=0;
 
-   script_name=0;
    script=0;
    script_only=false;
    script_needs_closing=false;
@@ -1082,29 +1139,6 @@ MirrorJob::MirrorJob(MirrorJob *parent,
 
 MirrorJob::~MirrorJob()
 {
-   xfree(source_dir);
-   xfree(target_dir);
-   xfree(source_relative_dir);
-   xfree(target_relative_dir);
-   delete source_set;
-   delete target_set;
-   delete to_transfer;
-   delete to_rm;
-   delete to_rm_mismatched;
-   delete same;
-   delete new_files_set;
-   delete old_files_set;
-   Delete(source_list_info);
-   Delete(target_list_info);
-   // session disposal should be done after ListInfo deletion.
-   SessionPool::Reuse(source_session);
-   SessionPool::Reuse(target_session);
-   if(!parent_mirror)
-   {
-      // common pointers, delete only on top level.
-      delete exclude;
-      delete size_range;
-   }
    if(script && script_needs_closing)
       fclose(script);
 }
@@ -1246,7 +1280,7 @@ bool MirrorJob::Statistics::HaveSomethingDone(int flags)
 
 char *MirrorJob::SetScriptFile(const char *n)
 {
-   script_name=xstrdup(n);
+   script_name.set(n);
    if(strcmp(n,"-"))
    {
       script=fopen(n,"w");
@@ -1262,6 +1296,11 @@ char *MirrorJob::SetScriptFile(const char *n)
    return xasprintf("%s: %s",n,strerror(errno));
 }
 
+void MirrorJob::SetOnChange(const char *oc)
+{
+   on_change.set(oc);
+}
+
 CMD(mirror)
 {
 #define args (parent->args)
@@ -1278,6 +1317,7 @@ CMD(mirror)
       OPT_NO_UMASK,
       OPT_OLDER_THAN,
       OPT_ONLY_MISSING,
+      OPT_ONLY_EXISTING,
       OPT_PERMS,
       OPT_REMOVE_SOURCE_FILES,
       OPT_SCRIPT,
@@ -1286,6 +1326,10 @@ CMD(mirror)
       OPT_USE_CACHE,
       OPT_USE_PGET_N,
       OPT_SKIP_NOACCESS,
+      OPT_ON_CHANGE,
+      OPT_NO_EMPTY_DIRS,
+      OPT_DEPTH_FIRST,
+      OPT_ASCII,
    };
    static const struct option mirror_opts[]=
    {
@@ -1315,6 +1359,7 @@ CMD(mirror)
       {"ignore-time",no_argument,0,OPT_IGNORE_TIME},
       {"ignore-size",no_argument,0,OPT_IGNORE_SIZE},
       {"only-missing",no_argument,0,OPT_ONLY_MISSING},
+      {"only-existing",no_argument,0,OPT_ONLY_EXISTING},
       {"log",required_argument,0,OPT_SCRIPT},
       {"script",    required_argument,0,OPT_SCRIPT_ONLY},
       {"just-print",optional_argument,0,OPT_SCRIPT_ONLY},
@@ -1325,6 +1370,10 @@ CMD(mirror)
       {"loop",no_argument,0,OPT_LOOP},
       {"max-errors",required_argument,0,OPT_MAX_ERRORS},
       {"skip-noaccess",no_argument,0,OPT_SKIP_NOACCESS},
+      {"on-change",required_argument,0,OPT_ON_CHANGE},
+      {"no-empty-dirs",no_argument,0,OPT_NO_EMPTY_DIRS},
+      {"depth-first",no_argument,0,OPT_DEPTH_FIRST},
+      {"ascii",no_argument,0,OPT_ASCII},
       {0}
    };
 
@@ -1334,13 +1383,13 @@ CMD(mirror)
 
    bool use_cache=false;
 
-   FileAccess *source_session=0;
-   FileAccess *target_session=0;
+   FileAccessRef source_session;
+   FileAccessRef target_session;
 
    int	 verbose=0;
    const char *newer_than=0;
    const char *older_than=0;
-   Range *size_range=0;
+   Ref<Range> size_range;
    bool  remove_source_files=false;
    bool	 skip_noaccess=ResMgr::QueryBool("mirror:skip-noaccess",0);
    int	 parallel=ResMgr::Query("mirror:parallel-transfer-count",0);
@@ -1348,8 +1397,9 @@ CMD(mirror)
    bool	 reverse=false;
    bool	 script_only=false;
    const char *script_file=0;
+   const char *on_change=0;
 
-   PatternSet *exclude=0;
+   Ref<PatternSet> exclude;
    const char *default_exclude=ResMgr::Query("mirror:exclude-regex",0);
    const char *default_include=ResMgr::Query("mirror:include-regex",0);
 
@@ -1400,15 +1450,14 @@ CMD(mirror)
 	 PatternSet::Pattern *pattern=0;
 	 if(opt=='x' || opt=='i')
 	 {
-	    PatternSet::Regex *rx=new PatternSet::Regex(optarg);
+	    Ref<PatternSet::Regex> rx(new PatternSet::Regex(optarg));
 	    if(rx->Error())
 	    {
 	       eprintf(_("%s: regular expression `%s': %s\n"),
 		  args->a0(),optarg,rx->ErrorText());
-	       delete rx;
 	       goto no_job;
 	    }
-	    pattern=rx;
+	    pattern=rx.borrow();
 	 }
 	 else // X or I
 	 {
@@ -1519,15 +1568,24 @@ CMD(mirror)
       case(OPT_SKIP_NOACCESS):
 	 skip_noaccess=true;
 	 break;
+      case(OPT_ON_CHANGE):
+	 on_change=optarg;
+	 break;
+      case(OPT_ONLY_EXISTING):
+	 flags|=MirrorJob::ONLY_EXISTING;
+	 break;
+      case(OPT_NO_EMPTY_DIRS):
+	 flags|=MirrorJob::NO_EMPTY_DIRS|MirrorJob::DEPTH_FIRST;
+	 break;
+      case(OPT_DEPTH_FIRST):
+	 flags|=MirrorJob::DEPTH_FIRST;
+	 break;
+      case(OPT_ASCII):
+	 flags|=MirrorJob::ASCII|MirrorJob::IGNORE_SIZE;
+	 break;
       case('?'):
 	 eprintf(_("Try `help %s' for more information.\n"),args->a0());
       no_job:
-	 delete exclude;
-	 delete size_range;
-	 if(source_session)
-	    SMTask::Delete(source_session);
-	 if(target_session)
-	    SMTask::Delete(target_session);
 	 return 0;
       }
    }
@@ -1556,7 +1614,7 @@ CMD(mirror)
 	 source_session=FileAccess::New(&source_url);
 	 if(!source_session)
 	 {
-	    eprintf("%s: %s%s\n",args->a0(),source_url.proto,
+	    eprintf("%s: %s%s\n",args->a0(),source_url.proto.get(),
 		     _(" - not supported protocol"));
 	    goto no_job;
 	 }
@@ -1572,7 +1630,7 @@ CMD(mirror)
 	    target_session=FileAccess::New(&target_url);
 	    if(!target_session)
 	    {
-	       eprintf("%s: %s%s\n",args->a0(),target_url.proto,
+	       eprintf("%s: %s%s\n",args->a0(),target_url.proto.get(),
 			_(" - not supported protocol"));
 	       goto no_job;
 	    }
@@ -1582,7 +1640,7 @@ CMD(mirror)
 	 {
 	    // user wants source dir name appended.
 	    const char *base=basename_ptr(source_dir);
-	    if(base[0]!='/')
+	    if(base[0]!='/' && strcmp(base,basename_ptr(arg)))
 	    {
 	       char *target_dir1=alloca_strdup2(target_dir,strlen(base));
 	       strcat(target_dir1,base);
@@ -1615,17 +1673,17 @@ CMD(mirror)
 	 target_session=parent->session->Clone();
    }
 
-   MirrorJob *j=new MirrorJob(0,source_session,target_session,source_dir,target_dir);
+   Ref<MirrorJob> j(new MirrorJob(0,source_session.borrow(),target_session.borrow(),source_dir,target_dir));
    j->SetFlags(flags,1);
    j->SetVerbose(verbose);
-   j->SetExclude(exclude);
+   j->SetExclude(exclude.borrow());
 
    if(newer_than)
       j->SetNewerThan(newer_than);
    if(older_than)
       j->SetOlderThan(older_than);
    if(size_range)
-      j->SetSizeRange(size_range);
+      j->SetSizeRange(size_range.borrow());
    j->UseCache(use_cache);
    if(remove_source_files)
       j->RemoveSourceFiles();
@@ -1637,17 +1695,15 @@ CMD(mirror)
       parallel=64;   // a (in)sane limit.
    if(parallel)
       j->SetParallel(parallel);
-   if(use_pget>1)
+   if(use_pget>1 && !(flags&MirrorJob::ASCII))
       j->SetPGet(use_pget);
 
    if(script_file)
    {
-      char *err=j->SetScriptFile(script_file);
+      xstring_ca err(j->SetScriptFile(script_file));
       if(err)
       {
-	 eprintf("%s: %s\n",args->a0(),err);
-	 xfree(err);
-	 delete j;
+	 eprintf("%s: %s\n",args->a0(),err.get());
 	 return 0;
       }
    }
@@ -1658,8 +1714,10 @@ CMD(mirror)
 	 j->SetScriptFile("-");
    }
    j->SetMaxErrorCount(max_error_count);
+   if(on_change)
+      j->SetOnChange(on_change);
 
-   return j;
+   return j.borrow();
 
 #undef args
 }
