@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: NetAccess.cc,v 1.87 2006/07/27 07:44:54 lav Exp $ */
+/* $Id: NetAccess.cc,v 1.104 2008/03/05 06:36:04 lav Exp $ */
 
 #include <config.h>
 
@@ -63,25 +63,17 @@ void NetAccess::Init()
    socket_buffer=0;
    socket_maxseg=0;
 
-   peer=0;
-   peer_num=0;
    peer_curr=0;
 
    reconnect_interval=30;  // retry with 30 second interval
    reconnect_interval_multiplier=1.2;
    reconnect_interval_max=300;
 
-   proxy=proxy_proto=0;
-   proxy_port=0;
-   proxy_user=proxy_pass=0;
-
    rate_limit=0;
 
    connection_limit=0;	// no limit.
    connection_takeover=false;
 
-   home_auto=0;
-
    Reconfig(0);
 }
 
@@ -92,22 +84,18 @@ NetAccess::NetAccess()
 NetAccess::NetAccess(const NetAccess *o) : super(o)
 {
    Init();
-   xfree(home_auto);
-   home_auto=xstrdup(o->home_auto);
+   if(o->peer)
+   {
+      peer.set(o->peer);
+      peer_curr=o->peer_curr;
+      if(peer_curr>=peer.count())
+	 peer_curr=0;
+   }
+   home_auto.set(o->home_auto);
 }
 NetAccess::~NetAccess()
 {
-   Delete(resolver);
-   if(rate_limit)
-      delete rate_limit;
    ClearPeer();
-
-   xfree(proxy);
-   xfree(proxy_port);
-   xfree(proxy_user);
-   xfree(proxy_pass);
-   xfree(proxy_proto);
-   xfree(home_auto);
 }
 
 void NetAccess::Reconfig(const char *name)
@@ -141,17 +129,15 @@ void NetAccess::KeepAlive(int sock)
 }
 void NetAccess::MinimizeLatency(int sock)
 {
-   int tos;
 #ifdef IP_TOS
-   tos = IPTOS_LOWDELAY;
+   int tos = IPTOS_LOWDELAY;
    setsockopt(sock, IPPROTO_IP, IP_TOS, (char *)&tos, sizeof(int));
 #endif
 }
 void NetAccess::MaximizeThroughput(int sock)
 {
-   int tos;
 #ifdef IP_TOS
-   tos = IPTOS_THROUGHPUT;
+   int tos = IPTOS_THROUGHPUT;
    setsockopt(sock, IPPROTO_IP, IP_TOS, (char *)&tos, sizeof(int));
 #endif
 }
@@ -164,9 +150,9 @@ void NetAccess::SetSocketBuffer(int sock)
    if(socket_buffer==0)
       return;
    if(-1==setsockopt(sock,SOL_SOCKET,SO_SNDBUF,(char*)&socket_buffer,sizeof(socket_buffer)))
-      Log::global->Format(1,"setsockopt(SO_SNDBUF,%d): %s\n",socket_buffer,strerror(errno));
+      LogError(1,"setsockopt(SO_SNDBUF,%d): %s",socket_buffer,strerror(errno));
    if(-1==setsockopt(sock,SOL_SOCKET,SO_RCVBUF,(char*)&socket_buffer,sizeof(socket_buffer)))
-      Log::global->Format(1,"setsockopt(SO_RCVBUF,%d): %s\n",socket_buffer,strerror(errno));
+      LogError(1,"setsockopt(SO_RCVBUF,%d): %s",socket_buffer,strerror(errno));
 }
 void NetAccess::SetSocketMaxseg(int sock)
 {
@@ -177,48 +163,10 @@ void NetAccess::SetSocketMaxseg(int sock)
    if(socket_maxseg==0)
       return;
    if(-1==setsockopt(sock,SOL_TCP,TCP_MAXSEG,(char*)&socket_maxseg,sizeof(socket_maxseg)))
-      Log::global->Format(1,"setsockopt(TCP_MAXSEG,%d): %s\n",socket_maxseg,strerror(errno));
+      LogError(1,"setsockopt(TCP_MAXSEG,%d): %s",socket_maxseg,strerror(errno));
 #endif
 }
 
-const char *NetAccess::SocketNumericAddress(const sockaddr_u *u)
-{
-#ifdef HAVE_GETNAMEINFO
-   static char buf[NI_MAXHOST];
-   if(getnameinfo(&u->sa,SocketAddrLen(u),buf,sizeof(buf),0,0,NI_NUMERICHOST)<0)
-      return "????";
-   return buf;
-#else
-   static char buf[256];
-   if(u->sa.sa_family!=AF_INET)
-      return "????";
-   unsigned char *a=(unsigned char *)&u->in.sin_addr;
-   sprintf(buf,"%u.%u.%u.%u",a[0],a[1],a[2],a[3]);
-   return buf;
-#endif
-}
-int NetAccess::SocketPort(const sockaddr_u *u)
-{
-   if(u->sa.sa_family==AF_INET)
-      return ntohs(u->in.sin_port);
-#if INET6
-   if(u->sa.sa_family==AF_INET6)
-      return ntohs(u->in6.sin6_port);
-#endif
-   return 0;
-}
-
-socklen_t NetAccess::SocketAddrLen(const sockaddr_u *u)
-{
-   if(u->sa.sa_family==AF_INET)
-      return sizeof(u->in);
-#if INET6
-   if(u->sa.sa_family==AF_INET6)
-      return sizeof(u->in6);
-#endif
-   return sizeof(*u);
-}
-
 #if HAVE_INET_ATON && !HAVE_DECL_INET_ATON
 CDECL int inet_aton(const char *,struct in_addr *);
 #endif
@@ -248,9 +196,9 @@ int NetAccess::SocketCreate(int af,int type,int proto)
 #endif
    if(b)
    {
-      int res=bind(s,&bind_addr.sa,sizeof(bind_addr.sa));
+      int res=bind_addr.bind_to(s);
       if(res==-1)
-	 Log::global->Format(0,"**** bind(socket, %s): %s\n",b,strerror(errno));
+	 LogError(0,"bind(socket, %s): %s",b,strerror(errno));
    }
    return s;
 }
@@ -329,7 +277,6 @@ int NetAccess::CheckHangup(const struct pollfd *pfd,int num)
    for(int i=0; i<num; i++)
    {
 #ifdef SO_ERROR
-      char  str[256];
       int   s_errno=0;
       socklen_t len;
 
@@ -346,15 +293,14 @@ int NetAccess::CheckHangup(const struct pollfd *pfd,int num)
 	 return 0;
       if(errno!=0 || s_errno!=0)
       {
-	 sprintf(str,_("Socket error (%s) - reconnecting"),
+	 LogError(0,_("Socket error (%s) - reconnecting"),
 				    strerror(errno?errno:s_errno));
-	 DebugPrint("**** ",str,0);
 	 return 1;
       }
 #endif /* SO_ERROR */
       if(pfd[i].revents&POLLERR)
       {
-	 DebugPrint("**** ","POLLERR",0);
+	 LogError(0,"POLLERR on fd %d",pfd[i].fd);
 	 return 1;
       }
    } /* end for */
@@ -378,22 +324,21 @@ int NetAccess::Poll(int fd,int ev)
 
 void NetAccess::SayConnectingTo()
 {
-   assert(peer_curr<peer_num);
+   assert(peer_curr<peer.count());
    const char *h=(proxy?proxy:hostname);
-   char *str=string_alloca(256+strlen(h));
-   sprintf(str,_("Connecting to %s%s (%s) port %u"),proxy?"proxy ":"",
+   LogNote(1,_("Connecting to %s%s (%s) port %u"),proxy?"proxy ":"",
       h,SocketNumericAddress(&peer[peer_curr]),SocketPort(&peer[peer_curr]));
-   DebugPrint("---- ",str,1);
 }
 
 void NetAccess::SetProxy(const char *px)
 {
    bool was_proxied=(proxy!=0);
 
-   xfree(proxy); proxy=0;
-   xfree(proxy_port); proxy_port=0;
-   xfree(proxy_user); proxy_user=0;
-   xfree(proxy_pass); proxy_pass=0;
+   proxy.set(0);
+   proxy_port.set(0);
+   proxy_user.set(0);
+   proxy_pass.set(0);
+   proxy_proto.set(0);
 
    if(!px)
       px="";
@@ -406,11 +351,11 @@ void NetAccess::SetProxy(const char *px)
       return;
    }
 
-   proxy=xstrdup(url.host);
-   proxy_port=xstrdup(url.port);
-   proxy_user=xstrdup(url.user);
-   proxy_pass=xstrdup(url.pass);
-   proxy_proto=xstrdup(url.proto);
+   proxy.set(url.host);
+   proxy_port.set(url.port);
+   proxy_user.set(url.user);
+   proxy_pass.set(url.pass);
+   proxy_proto.set(url.proto);
    ClearPeer();
 }
 
@@ -437,7 +382,7 @@ bool NetAccess::NoProxy(const char *hostname)
 
 void NetAccess::HandleTimeout()
 {
-   DebugPrint("**** ",_("Timeout - reconnecting"),0);
+   LogError(0,_("Timeout - reconnecting"));
    Disconnect();
    timeout_timer.Reset();
 }
@@ -454,15 +399,14 @@ bool NetAccess::CheckTimeout()
 
 void NetAccess::ClearPeer()
 {
-   xfree(peer);
-   peer=0;
-   peer_curr=peer_num=0;
+   peer.unset();
+   peer_curr=0;
 }
 
 void NetAccess::NextPeer()
 {
    peer_curr++;
-   if(peer_curr>=peer_num)
+   if(peer_curr>=peer.count())
       peer_curr=0;
    else
       try_time=0;	// try next address immediately
@@ -489,16 +433,14 @@ int NetAccess::Resolve(const char *defp,const char *ser,const char *pr)
 
    if(!resolver)
    {
-      xfree(peer);
-      peer=0;
-      peer_num=0;
+      peer.unset();
       if(proxy)
 	 resolver=new Resolver(proxy,proxy_port,defp);
       else
 	 resolver=new Resolver(hostname,portname,defp,ser,pr);
-      Roll(resolver);
       if(!resolver)
 	 return MOVED;
+      resolver->Roll();
       m=MOVED;
    }
 
@@ -511,121 +453,14 @@ int NetAccess::Resolve(const char *defp,const char *ser,const char *pr)
       return(MOVED);
    }
 
-   xfree(peer);
-   peer=(sockaddr_u*)xmalloc(resolver->GetResultSize());
-   peer_num=resolver->GetResultNum();
-   resolver->GetResult(peer);
-   if(peer_curr>=peer_num)
+   peer.set(resolver->Result());
+   if(peer_curr>=peer.count())
       peer_curr=0;
 
-   Delete(resolver);
    resolver=0;
    return MOVED;
 }
 
-
-// RateLimit class implementation.
-int RateLimit::total_xfer_number;
-RateLimit::BytesPool RateLimit::total[2];
-bool RateLimit::total_reconfig_needed=true;
-
-RateLimit::RateLimit(const char *c)
-{
-   if(total_xfer_number==0)
-   {
-      total[GET].Reset();
-      total[PUT].Reset();
-   }
-   total_xfer_number++;
-   Reconfig(0,c);
-}
-RateLimit::~RateLimit()
-{
-   total_xfer_number--;
-}
-
-#define LARGE 0x10000000
-#define DEFAULT_MAX_COEFF 2
-void RateLimit::BytesPool::AdjustTime()
-{
-   double dif=TimeDiff(SMTask::now,t);
-
-   if(dif>0)
-   {
-      // prevent overflow
-      if((LARGE-pool)/dif < rate)
-	 pool = pool_max>0 ? pool_max : rate*DEFAULT_MAX_COEFF;
-      else
-	 pool += int(dif*rate+0.5);
-
-      if(pool>pool_max && pool_max>0)
-	 pool=pool_max;
-      if(pool_max==0 && pool>rate*DEFAULT_MAX_COEFF)
-	 pool=rate*DEFAULT_MAX_COEFF;
-
-      t=SMTask::now;
-   }
-}
-
-int RateLimit::BytesAllowed(dir_t dir)
-{
-   if(total_reconfig_needed)
-      ReconfigTotal();
-
-   if(one[dir].rate==0 && total[dir].rate==0) // unlimited
-      return LARGE;
-
-   one  [dir].AdjustTime();
-   total[dir].AdjustTime();
-
-   int ret=LARGE;
-   if(total[dir].rate>0)
-      ret=total[dir].pool/total_xfer_number;
-   if(one[dir].rate>0 && ret>one[dir].pool)
-      ret=one[dir].pool;
-   return ret;
-}
-
-void RateLimit::BytesPool::Used(int bytes)
-{
-   if(pool<bytes)
-      pool=0;
-   else
-      pool-=bytes;
-}
-
-void RateLimit::BytesUsed(int bytes,dir_t dir)
-{
-   total[dir].Used(bytes);
-   one  [dir].Used(bytes);
-}
-
-void RateLimit::BytesPool::Reset()
-{
-   pool=rate;
-   t=SMTask::now;
-}
-void RateLimit::Reconfig(const char *name,const char *c)
-{
-   if(name && strncmp(name,"net:limit-",10))
-      return;
-   ResMgr::Query("net:limit-rate",c).ToNumberPair(one[GET].rate,one[PUT].rate);
-   ResMgr::Query("net:limit-max",c) .ToNumberPair(one[GET].pool_max,one[PUT].pool_max);
-   one[GET].Reset(); // to cut bytes_pool.
-   one[PUT].Reset();
-
-   if(name && !strncmp(name,"net:limit-total-",16))
-      total_reconfig_needed=true;
-}
-void RateLimit::ReconfigTotal()
-{
-   ResMgr::Query("net:limit-total-rate",0).ToNumberPair(total[GET].rate,total[PUT].rate);
-   ResMgr::Query("net:limit-total-max",0) .ToNumberPair(total[GET].pool_max,total[PUT].pool_max);
-   total[GET].Reset();
-   total[PUT].Reset();
-   total_reconfig_needed = false;
-}
-
 bool NetAccess::ReconnectAllowed()
 {
    if(max_retries>0 && retries>=max_retries)
@@ -655,11 +490,8 @@ const char *NetAccess::DelayingMessage()
 
 bool NetAccess::NextTry()
 {
-   if(max_retries>0 && retries>=max_retries)
-   {
-      Fatal(_("max-retries exceeded"));
+   if(!CheckRetries())
       return false;
-   }
    if(retries==0)
       reconnect_interval_current=reconnect_interval;
    else if(reconnect_interval_multiplier>1)
@@ -668,8 +500,17 @@ bool NetAccess::NextTry()
       if(reconnect_interval_current>reconnect_interval_max)
 	 reconnect_interval_current=reconnect_interval_max;
    }
-   try_time=now;
    retries++;
+   return CheckRetries();
+}
+bool NetAccess::CheckRetries()
+{
+   if(max_retries>0 && retries>max_retries)
+   {
+      Fatal(_("max-retries exceeded"));
+      return false;
+   }
+   try_time=now;
    return true;
 }
 void NetAccess::TrySuccess()
@@ -685,10 +526,7 @@ void NetAccess::Close()
       idle_timer.Reset();
 
    TrySuccess();
-
-   Delete(resolver);
    resolver=0;
-
    super::Close();
 }
 
@@ -712,7 +550,7 @@ void NetAccess::PropagateHomeAuto()
       NetAccess *o=(NetAccess*)fo; // we are sure it is NetAccess.
       if(!o->home_auto)
       {
-	 o->home_auto=xstrdup(home_auto);
+	 o->home_auto.set(home_auto);
 	 if(!o->home)
 	    o->set_home(home_auto);
       }
@@ -736,7 +574,6 @@ int GenericParseListInfo::Do()
 #define need_size (need&FileInfo::SIZE)
 #define need_time (need&FileInfo::DATE)
 
-   FA::fileinfo *cur;
    FileInfo *file;
    int res;
    int m=STALL;
@@ -775,8 +612,8 @@ do_again:
 	 ubuf->SetSpeedometer(new Speedometer());
 	 if(FileAccess::cache->IsEnabled(session->GetHostName()))
 	    ubuf->Save(FileAccess::cache->SizeLimit());
-	 Roll(session);
-	 Roll(ubuf);
+	 session->Roll();
+	 ubuf->Roll();
       }
       m=MOVED;
    }
@@ -788,13 +625,11 @@ do_again:
 	 if(mode==FA::MP_LIST)
 	 {
 	    mode=FA::LONG_LIST;
-	    delete ubuf;
 	    ubuf=0;
 	    m=MOVED;
 	    goto do_again;
 	 }
 	 SetError(ubuf->ErrorText());
-	 Delete(ubuf);
 	 ubuf=0;
 	 return MOVED;
       }
@@ -802,8 +637,6 @@ do_again:
       if(!ubuf->Eof())
 	 return m;
 
-      FileAccess::cache->Add(session,"",mode,FA::OK,ubuf);
-
       // now we have all the index in ubuf; parse it.
       const char *b;
       int len;
@@ -811,7 +644,7 @@ do_again:
 
       int old_mode=mode;
 
-      FileSet *set=Parse(b,len);
+      Ref<FileSet> set(Parse(b,len));
       if(set)
       {
 	 bool need_resort=false;
@@ -821,25 +654,20 @@ do_again:
 	    // tilde is special.
 	    if(file->name[0]=='~')
 	    {
-	       int len=strlen(file->name);
-	       file->name=(char*)xrealloc(file->name,len+3);
-	       memmove(file->name+2,file->name,len+1);
-	       memcpy(file->name,"./",2);
+	       file->name.set_substr(0,1,"./");
 	       need_resort=true;
 	    }
 	 }
 	 if(need_resort && !result)
 	    result=new FileSet; // Merge will sort the names
 	 if(result)
-	 {
 	    result->Merge(set);
-	    delete set;
-	 }
 	 else
-	    result=set;
+	    result=set.borrow();
       }
 
-      Delete(ubuf);
+      FileAccess::cache->Add(session,"",old_mode,FA::OK,ubuf,result);
+
       ubuf=0;
       m=MOVED;
 
@@ -852,27 +680,17 @@ do_again:
 
       result->ExcludeDots();
       result->Exclude(exclude_prefix,exclude);
-
-      get_info_cnt=result->get_fnum();
-      if(get_info_cnt==0)
-      {
-	 done=true;
-	 return m;
-      }
-
-      get_info=(FA::fileinfo*)xmalloc(sizeof(*get_info)*get_info_cnt);
-      cur=get_info;
-
-      get_info_cnt=0;
       result->rewind();
       for(file=result->curr(); file!=0; file=result->next())
       {
-	 cur->size=NO_SIZE;
-	 cur->time=NO_DATE;
-	 cur->get_size = need_size && !(file->defined & file->SIZE);
-	 cur->get_time = need_time && (!(file->defined & file->DATE)
-				 || (file->date_prec>0 && can_get_prec_time));
-	 cur->file=0;
+	 FA::fileinfo add;
+	 add.size=NO_SIZE;
+	 add.time=NO_DATE;
+	 add.get_size = need_size && !(file->defined & file->SIZE);
+	 add.get_time = need_time && (!(file->defined & file->DATE)
+				 || (file->date.ts_prec>0 && can_get_prec_time));
+	 add.is_dir=false;
+	 add.file=0;
 
 	 if(file->defined & file->TYPE)
 	 {
@@ -880,44 +698,42 @@ do_again:
 	    {
 	       //file->filetype=file->NORMAL;
 	       file->defined &= ~(file->SIZE|file->SYMLINK_DEF|file->MODE|file->DATE|file->TYPE);
-	       cur->get_size=true;
-	       cur->get_time=true;
+	       add.get_size=true;
+	       add.get_time=true;
 	    }
 
 	    if(file->filetype==file->SYMLINK)
 	    {
 	       // don't need these for symlinks
-	       cur->get_size=false;
-	       cur->get_time=false;
+	       add.get_size=false;
+	       add.get_time=false;
 	    }
 	    else if(file->filetype==file->DIRECTORY)
 	    {
 	       if(!get_time_for_dirs)
 		  continue;
 	       // don't need size for directories
-	       cur->get_size=false;
+	       add.get_size=false;
+	       add.is_dir=true;
 	    }
 	 }
 
-	 if(cur->get_size || cur->get_time)
+	 if(add.get_size || add.get_time)
 	 {
-	    cur->file=file->name;
-	    if(!cur->get_size)
-	       cur->size=-1;
-	    if(!cur->get_time)
-	       cur->time=(time_t)-1;
-	    cur++;
-	    get_info_cnt++;
+	    add.file=file->name;
+	    if(!add.get_size)
+	       add.size=NO_SIZE;
+	    if(!add.get_time)
+	       add.time=NO_DATE;
+	    get_info.append(add);
 	 }
       }
-      if(get_info_cnt==0)
+      if(get_info.count()==0)
       {
-	 xfree(get_info);
-	 get_info=0;
 	 done=true;
 	 return m;
       }
-      session->GetInfoArray(get_info,get_info_cnt);
+      session->GetInfoArray(get_info.get_non_const(),get_info.count());
    }
    if(get_info)
    {
@@ -927,15 +743,14 @@ do_again:
       if(res==FA::IN_PROGRESS)
 	 return m;
       session->Close();
-      for(cur=get_info; get_info_cnt-->0; cur++)
+      for(int i=0; i<get_info.count(); i++)
       {
-	 if(cur->time!=NO_DATE)
-	    result->SetDate(cur->file,cur->time,0);
-	 if(cur->size!=NO_SIZE)
-	    result->SetSize(cur->file,cur->size);
+	 if(get_info[i].time!=NO_DATE)
+	    result->SetDate(get_info[i].file,get_info[i].time,0);
+	 if(get_info[i].size!=NO_SIZE)
+	    result->SetSize(get_info[i].file,get_info[i].size);
       }
-      xfree(get_info);
-      get_info=0;
+      get_info.unset();
       done=true;
       m=MOVED;
    }
@@ -945,39 +760,20 @@ do_again:
 GenericParseListInfo::GenericParseListInfo(FileAccess *s,const char *p)
    : ListInfo(s,p)
 {
-   get_info=0;
-   get_info_cnt=0;
-
-   ubuf=0;
-
    get_time_for_dirs=true;
    can_get_prec_time=true;
-
    mode=FA::MP_LIST;
 }
 
-GenericParseListInfo::~GenericParseListInfo()
-{
-   xfree(get_info);
-   Delete(ubuf);
-}
-
 const char *GenericParseListInfo::Status()
 {
-   static char s[256];
    if(ubuf && !ubuf->Eof() && session->IsOpen())
-   {
-      sprintf(s,_("Getting directory contents (%lld) %s[%s]"),
+      return xstring::format("%s (%lld) %s[%s]",_("Getting directory contents"),
 		     (long long)session->GetPos(),
 		     ubuf->GetRateStrS(),session->CurrentStatus());
-      return s;
-   }
    if(get_info)
-   {
-      sprintf(s,_("Getting files information (%d%%) [%s]"),
+      return xstring::format("%s (%d%%) [%s]",_("Getting files information"),
 		     session->InfoArrayPercentDone(),
 		     session->CurrentStatus());
-      return s;
-   }
    return "";
 }
