@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 2002 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 2002-2005 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -91,8 +91,7 @@ void OutputJob::InitCopy()
 	 return;
       }
 
-      FileCopyPeerFA *dst_peer = FileCopyPeerFA::New(fa, fa_path, FA::STORE, true);
-      fa=0;
+      FileCopyPeerFA *dst_peer = FileCopyPeerFA::New(fa.borrow(), fa_path, FA::STORE);
 
       /* Status only for remote outputs. */
       if(!strcmp(dst_peer->GetProto(), "file"))
@@ -111,8 +110,7 @@ void OutputJob::InitCopy()
       pipe_output->CloseFD();
       output_fd->CloseFD();
 
-      xfree(fa_path);
-      fa_path=0;
+      fa_path.set(0);
    }
 
    initialized=true;
@@ -132,29 +130,19 @@ void OutputJob::InitCopy()
    if(filter)
    {
       /* Create the global filter: */
-      OutputFilter *global = new OutputFilter(filter, output_fd);
-      global->DeleteSecondaryStream();
+      OutputFilter *global = new OutputFilter(filter, output_fd.borrow());
       output_fd=global;
    }
 
    /* Use a FileCopy to buffer our output to the filter: */
-   FileCopyPeerFDStream *out = new FileCopyPeerFDStream(output_fd, FileCopyPeer::PUT);
+   FileCopyPeerFDStream *out = new FileCopyPeerFDStream(output_fd.borrow(), FileCopyPeer::PUT);
    FileCopy *input_fc = FileCopy::New(new FileCopyPeer(FileCopyPeer::GET), out, false);
 
-   /* out now owns output_fd, and will delete it when it's finished, so
-    * we can't keep it around. */
-   output_fd=0;
-
-   // I don't think we need to do this; the CopyJob picks up the output_fd's
-   // FgData now, since we're not telling it not to delete it.
-   // fg_data=new FgData(output_fd->GetProcGroup(),fg);
-
    if(!fail_if_broken)
       input_fc->DontFailIfBroken();
 
-   char *buf = xasprintf(_("%s (filter)"), a0);
+   xstring_ca buf(xasprintf(_("%s (filter)"), a0.get()));
    input=new CopyJob(input_fc, buf, filter?filter:a0);
-   xfree(buf);
 
    if(!output)
       output=input;
@@ -186,17 +174,12 @@ void OutputJob::InitCopy()
 void OutputJob::Init(const char *_a0)
 {
    input=output=0;
-   filter=0;
    initialized=false;
    error=false;
    no_status=false;
-   a0=xstrdup(_a0);
+   a0.set(_a0);
    is_stdout=false;
    fail_if_broken=true;
-   output_fd=0;
-   fa=0;
-   fa_path=0;
-   tmp_buf=0;
    is_a_tty=false;
    width=-1;
    statusbar_redisplay=true;
@@ -204,14 +187,11 @@ void OutputJob::Init(const char *_a0)
 
 /* Local (fd) output. */
 OutputJob::OutputJob(FDStream *output_, const char *a0)
+   : output_fd(output_ ? output_ : new FDStream(1,"<stdout>"))
 {
    Init(a0);
 
-   output_fd=output_;
-
-   if(!output_fd)
-      output_fd=new FDStream(1,"<stdout>");
-   else
+   if(output_)
    {
       /* We don't want to produce broken pipe when we're actually
        * piping, since some legitimate uses produce broken pipe, eg
@@ -237,24 +217,15 @@ OutputJob::OutputJob(FDStream *output_, const char *a0)
     * a file.) */
    if(output_fd->getfd()==-1 && output_fd->error())
    {
-      eprintf("%s: %s\n", a0, output_fd->error_text);
+      eprintf("%s: %s\n", a0, output_fd->error_text.get());
       error=true;
    }
 }
 
 OutputJob::OutputJob(const char *path, const char *a0, FileAccess *fa0)
+   : fa(fa0 ? fa0->Clone() : FileAccess::New("file")), fa_path(path)
 {
    Init(a0);
-
-   if(fa0)
-      fa=fa0->Clone();
-   else
-   {
-      fa=FileAccess::New("file");
-      if(!fa)
-	 fa=new DummyNoProto("file");
-   }
-   fa_path=xstrdup(path);
 }
 
 OutputJob::~OutputJob()
@@ -265,17 +236,10 @@ OutputJob::~OutputJob()
    Delete(input);
    if(input != output)
       Delete(output);
-   delete output_fd;
-   SessionPool::Reuse(fa);
-   xfree(fa_path);
-   delete tmp_buf;
-
-   xfree(a0);
-   xfree(filter);
 }
 
 /* This is called to ask us "permission" to display a status line. */
-bool OutputJob::ShowStatusLine(StatusLine *s)
+bool OutputJob::ShowStatusLine(const SMTaskRef<StatusLine>& s)
 {
    /* If our output file is gone, or isn't stdout, we don't care. */
    if(!output || !is_stdout)
@@ -362,20 +326,9 @@ void OutputJob::PutEOF()
 void OutputJob::PreFilter(const char *newfilter)
 {
    if(!filter)
-   {
-      SetFilter(newfilter);
-      return;
-   }
-
-   char *newstr = xasprintf("%s | %s", newfilter, filter);
-   SetFilter(newstr);
-   xfree(newstr);
-}
-
-void OutputJob::SetFilter(const char *newfilter)
-{
-   xfree(filter);
-   filter=xstrdup(newfilter);
+      filter.set(newfilter);
+   else
+      filter.vset(newfilter," | ",filter.get(),NULL);
 }
 
 /* Return the width of the output.  If there's a filter, we can either
@@ -397,16 +350,22 @@ bool OutputJob::IsTTY() const
    return is_a_tty;
 }
 
+static const SMTaskRef<FileCopyPeer> null_peer;
+
 /* Get the input FileCopyPeer; this is the buffer we write to. */
-FileCopyPeer *OutputJob::InputPeer() const
+const SMTaskRef<FileCopyPeer>& OutputJob::InputPeer() const
 {
-   return input? input->GetGet():0;
+   if(input)
+      return input->GetGet();
+   return null_peer;
 }
 
 /* Get the output FileCopyPeer (the FileCopyPeer that's doing the final output). */
-FileCopyPeer *OutputJob::OutputPeer() const
+const SMTaskRef<FileCopyPeer>& OutputJob::OutputPeer() const
 {
-   return output? output->GetPut():0;
+   if(output)
+      return output->GetPut();
+   return null_peer;
 }
 
 /* We're done if the output is finished, or on error. */
@@ -464,22 +423,20 @@ void OutputJob::Bg()
    super::Bg();
 }
 
-void OutputJob::Suspend()
+void OutputJob::SuspendInternal()
 {
    if(input)
-      input->Suspend();
+      input->SuspendSlave();
    if(output && input != output)
-      output->Suspend();
-   super::Suspend();
+      output->SuspendSlave();
 }
 
-void OutputJob::Resume()
+void OutputJob::ResumeInternal()
 {
    if(input)
-      input->Resume();
+      input->ResumeSlave();
    if(output && input != output)
-      output->Resume();
-   super::Resume();
+      output->ResumeSlave();
 }
 
 bool OutputJob::Full()
@@ -527,8 +484,7 @@ void OutputJob::Put(const char *buf,int size)
    // InputPeer was inited, flush tmp_buf.
    if(InputPeer() && tmp_buf)
    {
-      Buffer *saved_buf=tmp_buf;
-      tmp_buf=0;
+      Ref<Buffer> saved_buf(tmp_buf.borrow());
       const char *b=0;
       int s=0;
       saved_buf->Get(&b,&s);
@@ -536,7 +492,6 @@ void OutputJob::Put(const char *buf,int size)
 	 Put(b,s);
       if(saved_buf->Eof())
 	 PutEOF();
-      delete saved_buf;
    }
 
    update_timer.SetResource("cmd:status-interval",0);
@@ -550,10 +505,9 @@ void OutputJob::Format(const char *f,...)
 {
    va_list v;
    va_start(v,f);
-   char *str=xvasprintf(f,v);
+   xstring_ca str(xvasprintf(f,v));
    va_end(v);
    Put(str);
-   xfree(str);
 }
 
 /* Propagate signals down to our child processes. */
