@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: PtyShell.h,v 1.1 2001/10/12 07:05:57 lav Exp $ */
+/* $Id: PtyShell.h,v 1.4 2007/04/11 05:00:16 lav Exp $ */
 
 #ifndef PTYSHELL_H
 #define PTYSHELL_H
@@ -31,9 +31,12 @@ class PtyShell : public FDStream
    ProcWait *w;
    pid_t pg;
 
-   char *oldcwd;
+   xstring_c oldcwd;
 
    bool closed;
+   bool use_pipes;
+   int pipe_in;
+   int pipe_out;
 
    void Init();
 
@@ -43,12 +46,15 @@ public:
    ~PtyShell();
 
    void SetCwd(const char *);
+   void UsePipes() { use_pipes=true; }
 
    int getfd();
+   int getfd_pipe_in()  { return pipe_in; }
+   int getfd_pipe_out() { return pipe_out; }
    bool Done();
 
    void Kill(int sig=SIGTERM) { if(w) w->Kill(sig); }
-   pid_t GetProcGroup() { return pg; }
+   pid_t GetProcGroup() const { return pg; }
 
    bool broken();
 };
