@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: ResMgr.cc,v 1.64.2.2 2007/06/15 07:16:39 lav Exp $ */
+/* $Id: ResMgr.cc,v 1.69 2007/08/08 06:23:54 lav Exp $ */
 
 #include <config.h>
 
@@ -37,8 +37,8 @@ CDECL_BEGIN
 CDECL_END
 #include "ResMgr.h"
 #include "SMTask.h"
-#include "xmalloc.h"
 #include "misc.h"
+#include "StringSet.h"
 
 ResMgr::Resource  *ResMgr::chain=0;
 ResType		  *ResMgr::type_chain=0;
@@ -143,20 +143,13 @@ const char *ResMgr::Set(const char *name,const char *cclosure,const char *cvalue
    if(msg)
       return msg;
 
-   char *value=xstrdup(cvalue);
+   xstring_c value(cvalue);
    if(value && type->val_valid && (msg=(*type->val_valid)(&value))!=0)
-   {
-      xfree(value);
       return msg;
-   }
 
-   char *closure=xstrdup(cclosure);
+   xstring_c closure(cclosure);
    if(closure && type->closure_valid && (msg=(*type->closure_valid)(&closure))!=0)
-   {
-      xfree(closure);
-      xfree(value);
       return msg;
-   }
 
    Resource **scan;
    // find the old value
@@ -171,29 +164,22 @@ const char *ResMgr::Set(const char *name,const char *cclosure,const char *cvalue
    if(*scan)
    {
       if(value)
-      {
-	 xfree((*scan)->value);
-	 (*scan)->value=value;
-	 xfree(closure);
-      }
+	 (*scan)->value.set(value);
       else
       {
 	 Resource *to_free=*scan;
 	 *scan=(*scan)->next;
 	 delete to_free;
-	 xfree(closure);
       }
-      SMTask::ReconfigAll(type->name);
+      ResClient::ReconfigAll(type->name);
    }
    else
    {
       if(value)
       {
 	 chain=new Resource(chain,type,closure,value);
-	 SMTask::ReconfigAll(type->name);
+	 ResClient::ReconfigAll(type->name);
       }
-      else
-	 xfree(closure);
    }
    return 0;
 }
@@ -221,8 +207,6 @@ int ResMgr::VResourceCompare(const void *a,const void *b)
 
 char *ResMgr::Format(bool with_defaults,bool only_defaults)
 {
-   char *res;
-
    Resource *scan;
    ResType  *dscan;
 
@@ -250,8 +234,7 @@ char *ResMgr::Format(bool with_defaults,bool only_defaults)
       }
    }
 
-   res=(char*)xmalloc(size+1);
-   char *store=res;
+   xstring res("");
 
    Resource **created=(Resource**)alloca((dn+1)*sizeof(Resource*));
    Resource **c_store=created;
@@ -287,54 +270,50 @@ char *ResMgr::Format(bool with_defaults,bool only_defaults)
 
    for(i=0; i<n; i++)
    {
-      sprintf(store,"set %s",arr[i]->type->name);
-      store+=strlen(store);
-      char *s=arr[i]->closure;
+      res.appendf("set %s",arr[i]->type->name);
+      const char *s=arr[i]->closure;
       if(s)
       {
-	 *store++='/';
+	 res.append('/');
 	 bool par=false;
 	 if(strcspn(s," \t>|;&")!=strlen(s))
 	    par=true;
 	 if(par)
-	    *store++='"';
+	    res.append('"');
 	 while(*s)
 	 {
 	    if(strchr("\"\\",*s))
-	       *store++='\\';
-	    *store++=*s++;
+	       res.append('\\');
+	    res.append(*s++);
 	 }
 	 if(par)
-	    *store++='"';
+	    res.append('"');
       }
-      *store++=' ';
+      res.append(' ');
       s=arr[i]->value;
 
       bool par=false;
       if(*s==0 || strcspn(s," \t>|;&")!=strlen(s))
 	 par=true;
       if(par)
-	 *store++='"';
+	 res.append('"');
       while(*s)
       {
 	 if(strchr("\"\\",*s))
-	    *store++='\\';
-	 *store++=*s++;
+	    res.append('\\');
+	 res.append(*s++);
       }
       if(par)
-	 *store++='"';
-      *store++='\n';
+	 res.append('"');
+      res.append('\n');
    }
-   *store=0;
    for(i=0; i<dn; i++)
       delete created[i];
-   return res;
+   return res.borrow();
 }
 
 char **ResMgr::Generator(void)
 {
-   char **res;
-
    Resource *scan;
    ResType  *dscan;
 
@@ -345,8 +324,7 @@ char **ResMgr::Generator(void)
    for(dscan=type_chain; dscan; dscan=dscan->next)
       dn++;
 
-   res=(char**)xmalloc((n + dn + 1) * sizeof(char*));
-   char **store = res;
+   StringSet res;
 
    Resource **created=(Resource**)alloca((dn+1)*sizeof(Resource*));
    Resource **c_store=created;
@@ -373,18 +351,16 @@ char **ResMgr::Generator(void)
    qsort(arr,n,sizeof(*arr),&ResMgr::VResourceCompare);
 
    for(i=0; i<n; i++)
-      *(store++) = xstrdup(arr[i]->type->name);
-
-   *store=0;
+      res.Append(arr[i]->type->name);
 
    for(i=0; i<dn; i++)
       delete created[i];
-   return res;
+   return res.borrow();
 }
 
-const char *ResMgr::BoolValidate(char **value)
+const char *ResMgr::BoolValidate(xstring_c *value)
 {
-   char *v=*value;
+   const char *v=*value;
    const char *newval=0;
 
    switch(v[0])
@@ -407,21 +383,18 @@ const char *ResMgr::BoolValidate(char **value)
       return _("invalid boolean value");
    }
    if(strcmp(v,newval))
-   {
-      xfree(v);
-      *value=xstrdup(newval);
-   }
+      value->set(newval);
 
    return 0;
 }
 
-const char *ResMgr::TriBoolValidate(char **value)
+const char *ResMgr::TriBoolValidate(xstring_c *value)
 {
    if(!BoolValidate(value))
       return 0;
 
    /* not bool */
-   char *v=*value;
+   const char *v=*value;
    const char *newval=0;
 
    switch(v[0])
@@ -433,17 +406,14 @@ const char *ResMgr::TriBoolValidate(char **value)
    }
 
    if(strcmp(v,newval))
-   {
-      xfree(v);
-      *value=xstrdup(newval);
-   }
+      value->set(newval);
 
    return 0;
 }
 
-const char *ResMgr::NumberValidate(char **value)
+const char *ResMgr::NumberValidate(xstring_c *value)
 {
-   char *v=*value;
+   const char *v=*value;
 
    v+=strspn(v," \t");
 
@@ -455,14 +425,14 @@ const char *ResMgr::NumberValidate(char **value)
    if(n==0)
       return _("invalid number");
 
-   v[n]=0;
+   value->truncate(n+(v-*value));
 
    return 0;
 }
 
-const char *ResMgr::FloatValidate(char **value)
+const char *ResMgr::FloatValidate(xstring_c *value)
 {
-   char *v=*value;
+   const char *v=*value;
 
    int n=0;
 
@@ -470,17 +440,17 @@ const char *ResMgr::FloatValidate(char **value)
    if(1>sscanf(v,"%lf%n",&f,&n))
       return _("invalid floating point number");
 
-   v[n]=0;
+   value->truncate(n);
 
    return 0;
 }
 
-const char *ResMgr::UNumberValidate(char **value)
+const char *ResMgr::UNumberValidate(xstring_c *value)
 {
-   char *v=*value;
+   const char *v=*value;
 
    v+=strspn(v," \t");
-   memmove(*value,v,strlen(v)+1);   // drop leading space
+   value->set(v); // drop leading space
 
    v=*value;
    if(!strncasecmp(v,"0x",2))
@@ -491,7 +461,7 @@ const char *ResMgr::UNumberValidate(char **value)
    if(n==0)
       return _("invalid number");
 
-   v[n]=0;
+   value->truncate(n+(v-*value));
 
    return 0;
 }
@@ -686,7 +656,7 @@ void TimeIntervalR::init(const char *s)
    TimeDiff::Set(interval);
 }
 
-const char *ResMgr::TimeIntervalValidate(char **s)
+const char *ResMgr::TimeIntervalValidate(xstring_c *s)
 {
    TimeIntervalR t(*s);
    if(t.Error())
@@ -773,7 +743,7 @@ long long Range::Random()
    return start + (long long)((end-start+1)*mult);
 }
 
-const char *ResMgr::RangeValidate(char **s)
+const char *ResMgr::RangeValidate(xstring_c *s)
 {
    Range r(*s);
    if(r.Error())
@@ -781,7 +751,7 @@ const char *ResMgr::RangeValidate(char **s)
    return 0;
 }
 
-const char *ResMgr::ERegExpValidate(char **s)
+const char *ResMgr::ERegExpValidate(xstring_c *s)
 {
    if(**s==0)
       return 0;
@@ -801,7 +771,7 @@ const char *ResMgr::ERegExpValidate(char **s)
 CDECL int inet_aton(const char *,struct in_addr *);
 #endif
 
-const char *ResMgr::IPv4AddrValidate(char **value)
+const char *ResMgr::IPv4AddrValidate(xstring_c *value)
 {
    if(!**value)
       return 0;
@@ -812,7 +782,7 @@ const char *ResMgr::IPv4AddrValidate(char **value)
 }
 
 #if INET6
-const char *ResMgr::IPv6AddrValidate(char **value)
+const char *ResMgr::IPv6AddrValidate(xstring_c *value)
 {
    if(!**value)
       return 0;
@@ -823,16 +793,16 @@ const char *ResMgr::IPv6AddrValidate(char **value)
 }
 #endif
 
-const char *ResMgr::FileAccessible(char **value,int mode,int want_dir)
+const char *ResMgr::FileAccessible(xstring_c *value,int mode,int want_dir)
 {
    if(!**value)
       return 0;
    const char *f=expand_home_relative(*value);
-   char *cwd=0;
+   xstring_c cwd;
    const char *error=0;
    if(f[0]!='/')
    {
-      cwd=xgetcwd();
+      cwd.set_allocated(xgetcwd());
       if(cwd)
 	 f=dir_file(cwd,f);
    }
@@ -844,36 +814,28 @@ const char *ResMgr::FileAccessible(char **value,int mode,int want_dir)
    else if(access(f,mode)<0)
       error=strerror(errno);
    else
-   {
-      if(f!=*value)
-      {
-	 xfree(*value);
-	 *value=xstrdup(f);
-      }
-   }
-   xfree(cwd);
+      value->set(f);
    return error;
 }
-const char *ResMgr::FileReadable(char **value)
+const char *ResMgr::FileReadable(xstring_c *value)
 {
    return FileAccessible(value,R_OK);
 }
-const char *ResMgr::FileExecutable(char **value)
+const char *ResMgr::FileExecutable(xstring_c *value)
 {
    return FileAccessible(value,X_OK);
 }
-const char *ResMgr::DirReadable(char **value)
+const char *ResMgr::DirReadable(xstring_c *value)
 {
    return FileAccessible(value,R_OK|X_OK,1);
 }
 
 #ifdef HAVE_ICONV
 CDECL_BEGIN
-#include <iconv.h>
+# include <iconv.h>
 CDECL_END
 #endif
-
-const char *ResMgr::CharsetValidate(char **value)
+const char *ResMgr::CharsetValidate(xstring_c *value)
 {
    if(!**value)
       return 0;
@@ -888,26 +850,52 @@ const char *ResMgr::CharsetValidate(char **value)
 #endif
 }
 
-const char *ResMgr::NoClosure(char **)
+const char *ResMgr::NoClosure(xstring_c *)
 {
    return _("no closure defined for this setting");
 }
 
-const char *ResMgr::UNumberPairValidate(char **value)
+const char *ResMgr::UNumberPairValidate(xstring_c *value)
 {
-   char *v=*value;
-
    int n=0;
 
    unsigned a,b;
-   if(2>sscanf(v,"%u%*c%u%n",&a,&b,&n))
+   if(2>sscanf(value->get(),"%u%*c%u%n",&a,&b,&n))
    {
       if(UNumberValidate(value))
 	 return _("invalid pair of numbers");
       return 0;
    }
 
-   v[n]=0;
+   value->truncate(n);
 
    return 0;
 }
+
+ResClient *ResClient::chain;
+ResValue ResClient::Query(const char *name,const char *closure) const
+{
+   if(!strchr(name,':'))
+   {
+      const char *prefix=ResPrefix();
+      char *fullname=(char*)alloca(3+strlen(prefix)+1+strlen(name)+1);
+      sprintf(fullname,"%s:%s",prefix,name);
+      name=fullname;
+   }
+   if(!closure)
+      closure=ResClosure();
+   return ResMgr::Query(name,closure);
+}
+ResClient::ResClient()
+{
+   ListAdd(ResClient,chain,this,next);
+}
+ResClient::~ResClient()
+{
+   ListDel(ResClient,chain,this,next);
+}
+void ResClient::ReconfigAll(const char *r)
+{
+   ListScan(ResClient,chain,next)
+      scan->Reconfig(r);
+}
