@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 2003-2005 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 2003-2008 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: SFtp.cc,v 1.59.2.1 2007/06/15 06:28:51 lav Exp $ */
+/* $Id: SFtp.cc,v 1.77 2008/02/26 15:46:57 lav Exp $ */
 
 #include <config.h>
 #include "SFtp.h"
@@ -31,10 +31,11 @@
 
 #include <assert.h>
 #include <errno.h>
+#include <stddef.h>
 
 #define max_buf 0x10000
 
-#define super NetAccess
+#define super SSH_Access
 
 bool SFtp::GetBetterConnection(int level,bool limit_reached)
 {
@@ -77,7 +78,7 @@ int SFtp::Do()
    // check if idle time exceeded
    if(mode==CLOSED && send_buf && idle_timer.Stopped())
    {
-      DebugPrint("---- ",_("Closing idle connection"),1);
+      LogNote(1,_("Closing idle connection"));
       Disconnect();
       return m;
    }
@@ -90,10 +91,12 @@ int SFtp::Do()
 
    if(send_buf && send_buf->Error())
    {
+      LogError(0,"send: %s",send_buf->ErrorText());
       Disconnect();
       return MOVED;
    }
-   m|=HandleReplies();
+   if(state!=CONNECTING_1)
+      m|=HandleReplies();
 
    if(Error())
       return m;
@@ -163,12 +166,7 @@ int SFtp::Do()
 	 received_greeting=true;
       }
       else
-      {
-	 const char *greeting="echo SFTP: >&2";
-	 char *init1=string_alloca(strlen(init)+strlen(greeting)+2);
-	 sprintf(init1,"%s;%s",greeting,init);
-	 init=init1;
-      }
+	 init=xstring::cat("echo SFTP: >&2;",init,NULL);
       if(user)
       {
 	 cmd->Add("-l");
@@ -182,9 +180,8 @@ int SFtp::Do()
       cmd->Add(hostname);
       if(init[0])
 	 cmd->Add(init);
-      char *cmd_str=cmd->Combine(0);
-      Log::global->Format(9,"---- %s (%s)\n",_("Running connect program"),cmd_str);
-      xfree(cmd_str);
+      xstring_ca cmd_str(cmd->Combine(0));
+      LogNote(9,"%s (%s)",_("Running connect program"),cmd_str.get());
       ssh=new PtyShell(cmd);
       ssh->UsePipes();
       state=CONNECTING;
@@ -204,19 +201,15 @@ int SFtp::Do()
 	 TimeoutS(1);
 	 return m;
       }
-      int pipe_in =ssh->getfd_pipe_in();
-      int pipe_out=ssh->getfd_pipe_out();
-      ssh->Kill(SIGCONT);
-      pty_send_buf=new IOBufferFDStream(ssh,IOBuffer::PUT);
-      ssh=0;
-      pty_recv_buf=new IOBufferFDStream(new FDStream(fd,"pseudo-tty"),IOBuffer::GET);
-      send_buf=new IOBufferFDStream(new FDStream(pipe_out,"pipe-out"),IOBuffer::PUT);
-      recv_buf=new IOBufferFDStream(new FDStream(pipe_in,"pipe-in"),IOBuffer::GET);
+      MakePtyBuffers();
       set_real_cwd("~");
       state=CONNECTING_1;
       m=MOVED;
    }
    case CONNECTING_1:
+      m|=HandleSSHMessage();
+      if(state!=CONNECTING_1)
+	 return MOVED;
       if(!received_greeting)
 	 return m;
       SendRequest(new Request_INIT(Query("protocol-version",hostname)),Expect::FXP_VERSION);
@@ -224,6 +217,9 @@ int SFtp::Do()
       return MOVED;
 
    case CONNECTING_2:
+      m|=HandleSSHMessage();
+      if(state!=CONNECTING_2)
+	 return MOVED;
       if(protocol_version==0)
 	 return m;
       if(home_auto==0)
@@ -262,12 +258,16 @@ int SFtp::Do()
    case FILE_SEND:
       // pack data from file_buf.
       file_buf->Get(&b,&s);
-      if(s<size_write && !eof)
+      if(s==0 && !eof)
+	 return m;
+      if(s<size_write && !eof && !flush_timer.Stopped())
 	 return m;   // wait for more data before sending.
+      if(RespQueueSize()>max_packets_in_flight)
+	 return m;
       if(s==0)
       {
 	 // no more data, set attributes and close the file.
-	 Request_FSETSTAT *req=new Request_FSETSTAT(handle,handle_len,protocol_version);
+	 Request_FSETSTAT *req=new Request_FSETSTAT(handle,protocol_version);
 	 req->attrs.mtime=entity_date;
 	 req->attrs.flags|=SSH_FILEXFER_ATTR_MODIFYTIME;
 	 SendRequest(req,Expect::IGNORE);
@@ -276,11 +276,12 @@ int SFtp::Do()
 	 m=MOVED;
 	 break;
       }
-      if(RespQueueSize()>max_packets_in_flight)
-	 return m;
-      SendRequest(new Request_WRITE(handle,handle_len,request_pos,b,s),Expect::WRITE_STATUS);
+      if(s>size_write)
+	 s=size_write;
+      SendRequest(new Request_WRITE(handle,request_pos,b,s),Expect::WRITE_STATUS);
       file_buf->Skip(s);
       request_pos+=s;
+      flush_timer.Reset();
       m=MOVED;
       break;
    case WAITING:
@@ -295,14 +296,11 @@ int SFtp::Do()
 
 void SFtp::MoveConnectionHere(SFtp *o)
 {
+   super::MoveConnectionHere(o);
    protocol_version=o->protocol_version;
-   recv_translate=o->recv_translate; o->recv_translate=0;
-   send_translate=o->send_translate; o->send_translate=0;
-   send_buf=o->send_buf; o->send_buf=0;
-   recv_buf=o->recv_buf; o->recv_buf=0;
-   pty_send_buf=o->pty_send_buf; o->pty_send_buf=0;
-   pty_recv_buf=o->pty_recv_buf; o->pty_recv_buf=0;
-   rate_limit=o->rate_limit; o->rate_limit=0;
+   recv_translate=o->recv_translate.borrow();
+   send_translate=o->send_translate.borrow();
+   rate_limit=o->rate_limit.borrow();
    expect_queue_size=o->expect_queue_size; o->expect_queue_size=0;
    expect_chain=o->expect_chain; o->expect_chain=0;
    expect_chain_end=o->expect_chain_end;
@@ -320,39 +318,23 @@ void SFtp::MoveConnectionHere(SFtp *o)
 
 void SFtp::Disconnect()
 {
-   if(send_buf)
-      DebugPrint("---- ",_("Disconnecting"),9);
-   xfree(handle); handle=0; handle_len=0;
-   Delete(send_buf); send_buf=0;
-   Delete(recv_buf); recv_buf=0;
-   Delete(pty_send_buf); pty_send_buf=0;
-   Delete(pty_recv_buf); pty_recv_buf=0;
-   delete file_buf; file_buf=0;
-   delete ssh; ssh=0;
+   super::Disconnect();
+   handle.set(0);
+   file_buf=0;
    EmptyRespQueue();
    state=DISCONNECTED;
    if(mode==STORE)
       SetError(STORE_FAILED);
-   received_greeting=false;
-   password_sent=0;
    protocol_version=0;
-   delete send_translate; send_translate=0;
-   delete recv_translate; recv_translate=0;
+   send_translate=0;
+   recv_translate=0;
    ssh_id=0;
-   xfree(home_auto); home_auto=0;
-   home_auto=xstrdup(FindHomeAuto());
+   home_auto.set(FindHomeAuto());
 }
 
 void SFtp::Init()
 {
    state=DISCONNECTED;
-   send_buf=0;
-   recv_buf=0;
-   pty_send_buf=0;
-   pty_recv_buf=0;
-   file_buf=0;
-   file_set=0;
-   ssh=0;
    ssh_id=0;
    eof=false;
    received_greeting=false;
@@ -364,15 +346,15 @@ void SFtp::Init()
    protocol_version=0;
    send_translate=0;
    recv_translate=0;
-   handle=0;
-   handle_len=0;
    max_packets_in_flight=16;
    max_packets_in_flight_slow_start=1;
    size_read=0x8000;
    size_write=0x8000;
+   use_full_path=false;
+   flush_timer.Set(0,500);
 }
 
-SFtp::SFtp()
+SFtp::SFtp() : SSH_Access("SFTP:")
 {
    Init();
    Reconfig(0);
@@ -403,17 +385,15 @@ void SFtp::Packet::PackString(Buffer *b,const char *str,int len)
    b->PackUINT32BE(len);
    b->Put(str,len);
 }
-SFtp::unpack_status_t SFtp::Packet::UnpackString(Buffer *b,int *offset,int limit,char **str_out,int *len_out)
+SFtp::unpack_status_t SFtp::Packet::UnpackString(const Buffer *b,int *offset,int limit,xstring *str_out)
 {
-   assert(str_out && *str_out==0);
-
    if(limit-*offset<4)
       return b->Eof()?UNPACK_PREMATURE_EOF:UNPACK_NO_DATA_YET;
 
    int len=b->UnpackUINT32BE(*offset);
    if(len>limit-*offset-4)
    {
-      Log::global->Write(2,"**** bad string in reply (invalid length field)");
+      LogError(2,"bad string in reply (invalid length field)");
       return UNPACK_WRONG_FORMAT;
    }
    *offset+=4;
@@ -422,19 +402,14 @@ SFtp::unpack_status_t SFtp::Packet::UnpackString(Buffer *b,int *offset,int limit
    int data_len;
    b->Get(&data,&data_len);
 
-   char *string=(char*)xmalloc(len+1);
-   memcpy(string,data+*offset,len);
-   string[len]=0;
+   str_out->nset(data+*offset,len);
 
    *offset+=len;
-   *str_out=string;
-   if(len_out)
-      *len_out=len;
 
    return UNPACK_SUCCESS;
 }
 
-SFtp::unpack_status_t SFtp::Packet::Unpack(Buffer *b)
+SFtp::unpack_status_t SFtp::Packet::Unpack(const Buffer *b)
 {
    unpacked=0;
    if(b->Size()<4)
@@ -520,10 +495,10 @@ SFtp::unpack_status_t SFtp::UnpackPacket(Buffer *b,SFtp::Packet **p)
    case SSH_FXP_BLOCK:
    case SSH_FXP_UNBLOCK:
    case SSH_FXP_EXTENDED:
-      DebugPrint("**** ","request in reply??",0);
+      LogError(0,"request in reply??");
       return UNPACK_WRONG_FORMAT;
    case SSH_FXP_EXTENDED_REPLY:
-      DebugPrint("**** ","unexpected SSH_FXP_EXTENDED_REPLY",0);
+      LogError(0,"unexpected SSH_FXP_EXTENDED_REPLY");
       return UNPACK_WRONG_FORMAT;
    }
    res=pp->Unpack(b);
@@ -532,10 +507,10 @@ SFtp::unpack_status_t SFtp::UnpackPacket(Buffer *b,SFtp::Packet **p)
       switch(res)
       {
       case UNPACK_PREMATURE_EOF:
-	 DebugPrint("**** ","premature eof",0);
+	 LogError(0,"premature eof");
 	 break;
       case UNPACK_WRONG_FORMAT:
-	 DebugPrint("**** ","wrong packet format",0);
+	 LogError(0,"wrong packet format");
 	 break;
       case UNPACK_NO_DATA_YET:
       case UNPACK_SUCCESS:
@@ -554,7 +529,7 @@ void SFtp::SendRequest(Packet *request,Expect::expect_t tag,int i)
    request->ComputeLength();
    Log::global->Format(9,"---> sending a packet, length=%d, type=%d(%s), id=%u\n",
       request->GetLength(),request->GetPacketType(),request->GetPacketTypeText(),request->GetID());
-   request->Pack(send_buf);
+   request->Pack(send_buf.get_non_const());
    PushExpect(new Expect(request,tag,i));
 }
 
@@ -566,7 +541,7 @@ const char *SFtp::SkipHome(const char *path)
       return ".";
    if(!home)
       return path;
-   int home_len=strlen(home);
+   int home_len=home.path.length();
    if(strncmp(home,path,home_len))
       return path;
    if(path[home_len]=='/' && path[home_len+1] && path[home_len+1]!='/')
@@ -577,8 +552,10 @@ const char *SFtp::SkipHome(const char *path)
 }
 const char *SFtp::WirePath(const char *path)
 {
-   path=SkipHome(dir_file(cwd,path));
-   Log::global->Format(9,"---- path on wire is `%s'\n",path);
+   path=dir_file(cwd,path);
+   if(!use_full_path || path[0]=='~')
+      path=SkipHome(path);
+   LogNote(9,"path on wire is `%s'",path);
    return lc_to_utf8(path);
 }
 
@@ -604,8 +581,13 @@ void SFtp::SendRequest()
       state=WAITING;
       break;
    case STORE:
-      SendRequest(new Request_OPEN(WirePath(file),SSH_FXF_WRITE|SSH_FXF_CREAT,
-	 ACE4_WRITE_DATA|ACE4_WRITE_ATTRIBUTES,SSH_FXF_OPEN_OR_CREATE,protocol_version),Expect::HANDLE);
+      SendRequest(
+	 new Request_OPEN(WirePath(file),
+	    SSH_FXF_WRITE|SSH_FXF_CREAT|(pos==0?SSH_FXF_TRUNC:0),
+	    ACE4_WRITE_DATA|ACE4_WRITE_ATTRIBUTES,
+	    pos==0?SSH_FXF_CREATE_TRUNCATE:SSH_FXF_OPEN_OR_CREATE,
+	    protocol_version),
+	 Expect::HANDLE);
       state=WAITING;
       break;
    case ARRAY_INFO:
@@ -673,8 +655,8 @@ void SFtp::CloseHandle(Expect::expect_t c)
 {
    if(handle)
    {
-      SendRequest(new Request_CLOSE(handle,handle_len),c);
-      xfree(handle); handle=0; handle_len=0;
+      SendRequest(new Request_CLOSE(handle),c);
+      handle.set(0);
    }
 }
 
@@ -697,8 +679,8 @@ void SFtp::Close()
    CloseExpectQueue();
    state=(recv_buf?CONNECTED:DISCONNECTED);
    eof=false;
-   delete file_buf; file_buf=0;
-   delete file_set; file_set=0;
+   file_buf=0;
+   file_set=0;
    CloseHandle(Expect::IGNORE);
    super::Close();
    // don't need these out-of-order packets anymore
@@ -720,39 +702,10 @@ int SFtp::HandlePty()
    const char *eol=(const char*)memchr(b,'\n',s);
    if(!eol)
    {
-      const char *p="password:";
-      const char *y="(yes/no)?";
-      int p_len=strlen(p);
-      int y_len=strlen(y);
-      if(s>0 && b[s-1]==' ')
-	 s--;
-      if(s>=p_len && !strncasecmp(b+s-p_len,p,p_len)
-      || s>10 && !strncmp(b+s-2,"':",2))
-      {
-	 if(!pass)
-	 {
-	    SetError(LOGIN_FAILED,_("Password required"));
-	    return MOVED;
-	 }
-	 if(password_sent>1)
-	 {
-	    SetError(LOGIN_FAILED,_("Login incorrect"));
-	    return MOVED;
-	 }
-	 pty_recv_buf->Put("XXXX");
-	 pty_send_buf->Put(pass);
-	 pty_send_buf->Put("\n");
-	 password_sent++;
-	 return m;
-      }
-      if(s>=y_len && !strncasecmp(b+s-y_len,y,y_len))
-      {
-	 pty_recv_buf->Put("yes\n");
-	 pty_send_buf->Put("yes\n");
-	 return m;
-      }
       if(pty_recv_buf->Eof())
-	 DebugPrint("**** ",_("Peer closed connection"),0);
+	 LogError(0,_("Peer closed connection"));
+      if(pty_recv_buf->Error())
+	 LogError(0,"pty read: %s",pty_recv_buf->ErrorText());
       if(pty_recv_buf->Eof() || pty_recv_buf->Error())
       {
 	 Disconnect();
@@ -767,23 +720,21 @@ int SFtp::HandlePty()
    line[s-1]=0;
    pty_recv_buf->Skip(s);
 
-   DebugPrint("<--- ",line,4);
-   if(!received_greeting && !strcmp(line,"SFTP:"))
-      received_greeting=true;
+   LogRecv(4,line);
 
    return m;
 }
 
 void SFtp::HandleExpect(Expect *e)
 {
-   Packet *reply=e->reply;
+   const Packet *reply=e->reply;
    switch(e->tag)
    {
    case Expect::FXP_VERSION:
       if(reply->TypeIs(SSH_FXP_VERSION))
       {
 	 protocol_version=((Reply_VERSION*)reply)->GetVersion();
-	 Log::global->Format(9,"---- protocol version set to %d\n",protocol_version);
+	 LogNote(9,"protocol version set to %d",protocol_version);
 	 const char *charset=0;
 	 if(protocol_version>=4)
 	    charset="UTF-8";
@@ -810,8 +761,8 @@ void SFtp::HandleExpect(Expect *e)
 	 const NameAttrs *a=r->GetNameAttrs(0);
 	 if(a && !home_auto)
 	 {
-	    home_auto=xstrdup(utf8_to_lc(a->name));
-	    Log::global->Format(9,"---- home set to %s\n",home_auto);
+	    home_auto.set(utf8_to_lc(a->name));
+	    LogNote(9,"home set to %s",home_auto.get());
 	    PropagateHomeAuto();
 	    if(!home)
 	       set_home(home_auto);
@@ -842,23 +793,23 @@ void SFtp::HandleExpect(Expect *e)
    case Expect::HANDLE:
       if(reply->TypeIs(SSH_FXP_HANDLE))
       {
-	 handle=((Reply_HANDLE*)reply)->GetHandle(&handle_len);
+	 handle.set(((Reply_HANDLE*)reply)->GetHandle());
 	 state=(mode==STORE?FILE_SEND:FILE_RECV);
-	 assert(!file_buf);
 	 file_buf=new Buffer;
-	 Log::global->Write(9,"---- got file handle ");
+	 xstring handle_x("");
+	 int handle_len=handle.length();
 	 for(int i=0; i<handle_len; i++)
-	    Log::global->Format(9,"%02X",handle[i]);
-	 Log::global->Format(9," (%d)\n",handle_len);
+	    handle_x.appendf("%02X",(unsigned char)handle[i]);
+	 LogNote(9,"got file handle %s (%d)",handle_x.get(),handle_len);
 	 request_pos=real_pos=pos;
 	 if(mode==RETRIEVE)
-	    SendRequest(new Request_FSTAT(handle,handle_len,
+	    SendRequest(new Request_FSTAT(handle,
 	       SSH_FILEXFER_ATTR_SIZE|SSH_FILEXFER_ATTR_MODIFYTIME|SSH_FILEXFER_ATTR_PERMISSIONS,
 	       protocol_version),Expect::INFO);
-	 else if(mode==STORE)
+	 else if(mode==STORE && pos>0)
 	 {
 	    // truncate the file at write position.
-	    Request_FSETSTAT *req=new Request_FSETSTAT(handle,handle_len,protocol_version);
+	    Request_FSETSTAT *req=new Request_FSETSTAT(handle,protocol_version);
 	    req->attrs.size=pos;
 	    req->attrs.flags|=SSH_FILEXFER_ATTR_SIZE;
 	    SendRequest(req,Expect::IGNORE);
@@ -871,10 +822,8 @@ void SFtp::HandleExpect(Expect *e)
       if(reply->TypeIs(SSH_FXP_HANDLE))
       {
 	 // close the handle immediately.
-	 int h_len;
-	 char *handle=((Reply_HANDLE*)reply)->GetHandle(&h_len);
-	 SendRequest(new Request_CLOSE(handle,h_len),Expect::IGNORE);
-	 xfree(handle);
+	 const xstring &handle=((Reply_HANDLE*)reply)->GetHandle();
+	 SendRequest(new Request_CLOSE(handle),Expect::IGNORE);
       }
       break;
    case Expect::DATA:
@@ -882,13 +831,13 @@ void SFtp::HandleExpect(Expect *e)
 	 max_packets_in_flight_slow_start++;
       if(reply->TypeIs(SSH_FXP_DATA))
       {
-	 Request_READ *r=(Request_READ*)e->request;
+	 const Request_READ *r=e->request.Cast<Request_READ>();
 	 Reply_DATA *d=(Reply_DATA*)reply;
 	 if(r->pos==pos+file_buf->Size())
 	 {
 	    const char *b; int s;
 	    d->GetData(&b,&s);
-	    Log::global->Format(9,"---- data packet: pos=%lld, size=%d\n",(long long)r->pos,s);
+	    LogNote(9,"data packet: pos=%lld, size=%d",(long long)r->pos,s);
 	    file_buf->Put(b,s);
 	    if(d->Eof())
 	       goto eof;
@@ -899,14 +848,14 @@ void SFtp::HandleExpect(Expect *e)
 	       if(r->pos+r->len==request_pos)
 		  request_pos=r->pos+s;
 	       else
-		  SendRequest(new Request_READ(handle,handle_len,r->pos+s,r->len-s),Expect::DATA);
+		  SendRequest(new Request_READ(handle,r->pos+s,r->len-s),Expect::DATA);
 	    }
 	 }
 	 else
 	 {
 	    if(e->next!=ooo_chain)
-	       Log::global->Format(9,"---- put a packet with id=%d on out-of-order chain (need_pos=%lld packet_pos=%lld)\n",
-		  reply->GetID(),pos+file_buf->Size(),r->pos);
+	       LogNote(9,"put a packet with id=%d on out-of-order chain (need_pos=%lld packet_pos=%lld)",
+		  reply->GetID(),(long long)(pos+file_buf->Size()),(long long)r->pos);
 	    e->next=ooo_chain;
 	    ooo_chain=e;
 	    return;
@@ -915,7 +864,7 @@ void SFtp::HandleExpect(Expect *e)
       else if(reply->TypeIs(SSH_FXP_NAME))
       {
 	 Reply_NAME *r=(Reply_NAME*)reply;
-	 Log::global->Format(9,"---- file name count=%d\n",r->GetCount());
+	 LogNote(9,"file name count=%d",r->GetCount());
 	 for(int i=0; i<r->GetCount(); i++)
 	 {
 	    const NameAttrs *a=r->GetNameAttrs(i);
@@ -957,7 +906,7 @@ void SFtp::HandleExpect(Expect *e)
 	    {
 	    eof:
 	       if(!eof)
-		  Log::global->Write(9,"---- eof\n");
+		  LogNote(9,"eof");
 	       eof=true;
 	       state=DONE;
 	       if(file_buf && !ooo_chain)
@@ -986,7 +935,7 @@ void SFtp::HandleExpect(Expect *e)
 	    array_for_info[e->i].get_time=false;
 	    break;
 	 }
-	 Log::global->Format(9,"---- file info: size=%lld, date=%s",(long long)entity_size,ctime(&entity_date));
+	 LogNote(9,"file info: size=%lld, date=%s",(long long)entity_size,ctime(&entity_date));
 	 if(opt_size)
 	    *opt_size=entity_size;
 	 if(opt_date)
@@ -1022,19 +971,22 @@ void SFtp::RequestMoreData()
 {
    if(mode==RETRIEVE) {
       int req_len=size_read;
-      SendRequest(new Request_READ(handle,handle_len,request_pos,req_len),Expect::DATA);
+      SendRequest(new Request_READ(handle,request_pos,req_len),Expect::DATA);
       request_pos+=req_len;
    } else if(mode==LIST || mode==LONG_LIST) {
-      SendRequest(new Request_READDIR(handle,handle_len),Expect::DATA);
+      SendRequest(new Request_READDIR(handle),Expect::DATA);
    }
 }
 
 int SFtp::HandleReplies()
 {
-   int m=HandlePty();
+   int m=STALL;
    if(recv_buf==0)
       return m;
 
+   if(state!=CONNECTING_2)
+      m|=HandlePty();
+
    int i=0;
    Expect *ooo_scan=ooo_chain;
    while(ooo_scan)
@@ -1045,7 +997,7 @@ int SFtp::HandleReplies()
       ooo_scan=next;
       if(++i>64)
       {
-	 DebugPrint("**** ","Too many out-of-order packets");
+	 LogError(0,"Too many out-of-order packets");
 	 Disconnect();
 	 return MOVED;
       }
@@ -1057,12 +1009,13 @@ int SFtp::HandleReplies()
    {
       if(recv_buf->Error())
       {
+	 LogError(0,"receive: %s",recv_buf->ErrorText());
 	 Disconnect();
 	 return MOVED;
       }
       if(recv_buf->Eof() && pty_recv_buf->Size()==0)
       {
-	 DebugPrint("**** ",_("Peer closed connection"),0);
+	 LogError(0,_("Peer closed connection"));
 	 Disconnect();
 	 m=MOVED;
       }
@@ -1073,21 +1026,21 @@ int SFtp::HandleReplies()
       return m;
 
    Packet *reply=0;
-   unpack_status_t st=UnpackPacket(recv_buf,&reply);
+   unpack_status_t st=UnpackPacket(recv_buf.get_non_const(),&reply);
    if(st==UNPACK_NO_DATA_YET)
       return m;
    if(st!=UNPACK_SUCCESS)
    {
-      DebugPrint("**** ",_("invalid server response format"),2);
+      LogError(2,_("invalid server response format"));
       Disconnect();
       return MOVED;
    }
 
-   reply->DropData(recv_buf);
+   reply->DropData(recv_buf.get_non_const());
    Expect *e=FindExpectExclusive(reply);
    if(e==0)
    {
-      DebugPrint("**** ",_("extra server response"),3);
+      LogError(3,_("extra server response"));
       delete reply;
       return MOVED;
    }
@@ -1240,7 +1193,7 @@ int SFtp::Write(const void *buf,int size)
       size=entity_size-pos;
    if(size<=0)
       return 0;
-   file_buf->Put((char*)buf,size);
+   file_buf->Put(static_cast<const char*>(buf),size);
    TrySuccess();
    rate_limit->BytesPut(size);
    pos+=size;
@@ -1389,11 +1342,12 @@ void SFtp::Reconfig(const char *name)
       size_read=16;
    if(size_write<16)
       size_write=16;
+   use_full_path=QueryBool("use-full-path",c);
    if(!xstrcmp(name,"sftp:charset") && protocol_version && protocol_version<4)
    {
       if(!IsSuspended())
 	 cache->TreeChanged(this,"/");
-      const char *charset=ResMgr::Query("fish:charset",hostname);
+      const char *charset=ResMgr::Query("sftp:charset",hostname);
       if(charset && *charset)
       {
 	 if(!send_translate)
@@ -1403,6 +1357,11 @@ void SFtp::Reconfig(const char *name)
 	 send_translate->SetTranslation(charset,false);
 	 recv_translate->SetTranslation(charset,true);
       }
+      else
+      {
+	 send_translate=0;
+	 recv_translate=0;
+      }
    }
 }
 
@@ -1415,7 +1374,7 @@ FileAccess *SFtp::New() { return new SFtp(); }
 
 DirList *SFtp::MakeDirList(ArgV *args)
 {
-  return new SFtpDirList(args,this);
+  return new SFtpDirList(this,args);
 }
 
 struct code_text { int code; const char *text; };
@@ -1542,21 +1501,17 @@ void SFtp::SetError(int code,const Packet *reply)
 #define PACK32_SIGNED(data)	b->PackINT32BE(data)
 #define PACK64_SIGNED(data)	b->PackINT64BE(data)
 
-SFtp::unpack_status_t SFtp::PacketSTRING::Unpack(Buffer *b)
+SFtp::unpack_status_t SFtp::PacketSTRING::Unpack(const Buffer *b)
 {
    unpack_status_t res;
    res=Packet::Unpack(b);
    if(res!=UNPACK_SUCCESS)
       return res;
-   res=UnpackString(b,&unpacked,length+4,&string,&string_len);
+   res=UnpackString(b,&unpacked,length+4,&string);
    return res;
 }
-SFtp::PacketSTRING::~PacketSTRING()
-{
-   xfree(string);
-}
 
-SFtp::unpack_status_t SFtp::Reply_NAME::Unpack(Buffer *b)
+SFtp::unpack_status_t SFtp::Reply_NAME::Unpack(const Buffer *b)
 {
    unpack_status_t res=Packet::Unpack(b);
    if(res!=UNPACK_SUCCESS)
@@ -1575,7 +1530,7 @@ SFtp::unpack_status_t SFtp::Reply_NAME::Unpack(Buffer *b)
       UNPACK8(eof);
    return UNPACK_SUCCESS;
 }
-SFtp::unpack_status_t SFtp::Reply_DATA::Unpack(Buffer *b)
+SFtp::unpack_status_t SFtp::Reply_DATA::Unpack(const Buffer *b)
 {
    unpack_status_t res=PacketSTRING::Unpack(b);
    if(res!=UNPACK_SUCCESS)
@@ -1586,7 +1541,7 @@ SFtp::unpack_status_t SFtp::Reply_DATA::Unpack(Buffer *b)
       UNPACK8(eof);
    return UNPACK_SUCCESS;
 }
-SFtp::unpack_status_t SFtp::NameAttrs::Unpack(Buffer *b,int *offset,int limit,int protocol_version)
+SFtp::unpack_status_t SFtp::NameAttrs::Unpack(const Buffer *b,int *offset,int limit,int protocol_version)
 {
    unpack_status_t res;
 
@@ -1606,7 +1561,7 @@ SFtp::unpack_status_t SFtp::NameAttrs::Unpack(Buffer *b,int *offset,int limit,in
    return UNPACK_SUCCESS;
 }
 
-SFtp::unpack_status_t SFtp::FileAttrs::Unpack(Buffer *b,int *offset,int limit,int protocol_version)
+SFtp::unpack_status_t SFtp::FileAttrs::Unpack(const Buffer *b,int *offset,int limit,int protocol_version)
 {
    unpack_status_t res;
 
@@ -1834,13 +1789,11 @@ int SFtp::FileAttrs::ComputeLength(int protocol_version)
 }
 SFtp::FileAttrs::~FileAttrs()
 {
-   xfree(owner); xfree(group);
-   xfree(mime_type); xfree(untranslated_name);
    delete[] extended_attrs;
    delete[] ace;
 }
 
-SFtp::unpack_status_t SFtp::FileAttrs::FileACE::Unpack(Buffer *b,int *offset,int limit)
+SFtp::unpack_status_t SFtp::FileAttrs::FileACE::Unpack(const Buffer *b,int *offset,int limit)
 {
    UNPACK32(ace_type);
    UNPACK32(ace_flag);
@@ -1855,7 +1808,7 @@ void SFtp::FileAttrs::FileACE::Pack(Buffer *b)
    Packet::PackString(b,who);
 }
 
-SFtp::unpack_status_t SFtp::FileAttrs::ExtFileAttr::Unpack(Buffer *b,int *offset,int limit)
+SFtp::unpack_status_t SFtp::FileAttrs::ExtFileAttr::Unpack(const Buffer *b,int *offset,int limit)
 {
    unpack_status_t res;
    res=Packet::UnpackString(b,offset,limit,&extended_type);
@@ -1872,7 +1825,7 @@ void SFtp::FileAttrs::ExtFileAttr::Pack(Buffer *b)
    Packet::PackString(b,extended_data);
 }
 
-SFtp::unpack_status_t SFtp::Reply_ATTRS::Unpack(Buffer *b)
+SFtp::unpack_status_t SFtp::Reply_ATTRS::Unpack(const Buffer *b)
 {
    unpack_status_t res=Packet::Unpack(b);
    if(res!=UNPACK_SUCCESS)
@@ -1880,7 +1833,7 @@ SFtp::unpack_status_t SFtp::Reply_ATTRS::Unpack(Buffer *b)
    return attrs.Unpack(b,&unpacked,length+4,protocol_version);
 }
 
-SFtp::unpack_status_t SFtp::Reply_STATUS::Unpack(Buffer *b)
+SFtp::unpack_status_t SFtp::Reply_STATUS::Unpack(const Buffer *b)
 {
    unpack_status_t res=Packet::Unpack(b);
    if(res!=UNPACK_SUCCESS)
@@ -1909,6 +1862,7 @@ void SFtp::Request_WRITE::Pack(Buffer *b)
 {
    PacketSTRING::Pack(b);
    PACK64(pos);
+   int len=data.length();
    PACK32(len);
    b->Put(data,len);
 }
@@ -1970,13 +1924,13 @@ FileInfo *SFtp::MakeFileInfo(const NameAttrs *na)
       return 0;
    if(name[0]=='~')
       name=dir_file(".",name);
-   FileInfo *fi=new FileInfo(name);
+   Ref<FileInfo> fi(new FileInfo(name));
    switch(a->type)
    {
    case SSH_FILEXFER_TYPE_REGULAR:  fi->SetType(fi->NORMAL);    break;
    case SSH_FILEXFER_TYPE_DIRECTORY:fi->SetType(fi->DIRECTORY); break;
    case SSH_FILEXFER_TYPE_SYMLINK:  fi->SetType(fi->SYMLINK);   break;
-   default: delete fi; return 0;
+   default: return 0;
    }
    if(na->longname)
       fi->SetLongName(utf8_to_lc(na->longname));
@@ -1998,7 +1952,7 @@ FileInfo *SFtp::MakeFileInfo(const NameAttrs *na)
    else if(fi->longname)
    {
       // try to extract owner/group from long name.
-      FileInfo *ls=FileInfo::parse_ls_line(fi->longname,0);
+      Ref<FileInfo> ls(FileInfo::parse_ls_line(fi->longname,0));
       if(ls)
       {
 	 if(ls->user)
@@ -2008,13 +1962,12 @@ FileInfo *SFtp::MakeFileInfo(const NameAttrs *na)
 	 if(ls->nlinks>0)
 	    fi->SetNlink(ls->nlinks);
       }
-      delete ls;
    }
    if(a->flags&SSH_FILEXFER_ATTR_PERMISSIONS)
       fi->SetMode(a->permissions&07777);
    if(a->flags&SSH_FILEXFER_ATTR_MODIFYTIME)
       fi->SetDate(a->mtime,0);
-   return fi;
+   return fi.borrow();
 }
 
 
@@ -2069,7 +2022,7 @@ int SFtpDirList::Do()
    if(b==0) // eof
    {
       if(!fset && session->IsOpen())
-	 fset=((SFtp*)session)->GetFileSet();
+	 fset=session.Cast<SFtp>()->GetFileSet();
       FileAccess::cache->Add(session,dir,FA::LONG_LIST,FA::OK,ubuf,fset);
       if(use_file_set)
       {
@@ -2080,10 +2033,9 @@ int SFtpDirList::Do()
 	    buf->Put(fi->GetLongName());
 	    buf->Put("\n");
 	 }
-	 delete fset;
 	 fset=0;
       }
-      Delete(ubuf); ubuf=0;
+      ubuf=0;
       dir=args->getnext();
       if(!dir)
 	 buf->PutEOF();
@@ -2108,13 +2060,10 @@ int SFtpDirList::Do()
    return m;
 }
 
-SFtpDirList::SFtpDirList(ArgV *a,FileAccess *fa)
-   : DirList(a)
+SFtpDirList::SFtpDirList(SFtp *s,ArgV *a)
+   : DirList(s,a)
 {
-   session=fa;
-   ubuf=0;
    use_file_set=true;
-   fset=0;
    args->rewind();
    int opt;
    while((opt=args->getopt("fCFl"))!=EOF)
@@ -2142,21 +2091,11 @@ SFtpDirList::SFtpDirList(ArgV *a,FileAccess *fa)
       buf->Format("%s:\n",dir);
 }
 
-SFtpDirList::~SFtpDirList()
-{
-   Delete(ubuf);
-   delete fset;
-}
-
 const char *SFtpDirList::Status()
 {
-   static char s[256];
    if(ubuf && !ubuf->Eof() && session->IsOpen())
-   {
-      sprintf(s,_("Getting file list (%lld) [%s]"),
+      return xstring::format(_("Getting file list (%lld) [%s]"),
 		     (long long)session->GetPos(),session->CurrentStatus());
-      return s;
-   }
    return "";
 }
 
@@ -2212,7 +2151,7 @@ int SFtpListInfo::Do()
    if(b==0) // eof
    {
       if(!result && session->IsOpen())
-	 result=((SFtp*)session)->GetFileSet();
+	 result=session.Cast<SFtp>()->GetFileSet();
       FileAccess::cache->Add(session,"",FA::LONG_LIST,FA::OK,ubuf,result);
       result->ExcludeDots();
       result->Exclude(exclude_prefix,exclude);
@@ -2233,13 +2172,9 @@ int SFtpListInfo::Do()
 }
 const char *SFtpListInfo::Status()
 {
-   static char s[256];
    if(ubuf && !ubuf->Eof() && session->IsOpen())
-   {
-      sprintf(s,_("Getting file list (%lld) [%s]"),
+      return xstring::format(_("Getting file list (%lld) [%s]"),
 		     (long long)session->GetPos(),session->CurrentStatus());
-      return s;
-   }
    return "";
 }
 
