@@ -18,19 +18,18 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: SFtp.h,v 1.21 2006/06/13 14:35:41 lav Exp $ */
+/* $Id: SFtp.h,v 1.27 2008/02/26 15:46:57 lav Exp $ */
 
 #ifndef SFTP_H
 #define SFTP_H
 
-#include "NetAccess.h"
+#include "SSH_Access.h"
 #include "StatusLine.h"
-#include "PtyShell.h"
 #include <sys/types.h>
 #include <sys/stat.h>
 #include "FileSet.h"
 
-class SFtp : public NetAccess
+class SFtp : public SSH_Access
 {
    int	 protocol_version;
 
@@ -219,28 +218,20 @@ private:
    };
 
    state_t state;
-   bool received_greeting;
-   int  password_sent;
    unsigned ssh_id;
-   char *handle;
-   int handle_len;
+   xstring handle;
 
    void Init();
 
    void	 SendMethod();
    void	 SendArrayInfoRequests();
 
-   IOBuffer *send_buf;
-   IOBuffer *recv_buf;
-   IOBuffer *pty_send_buf;
-   IOBuffer *pty_recv_buf;
-   DirectedBuffer *send_translate;
-   DirectedBuffer *recv_translate;
+   Ref<DirectedBuffer> send_translate;
+   Ref<DirectedBuffer> recv_translate;
 
-   Buffer   *file_buf;
-   FileSet  *file_set;
-
-   PtyShell *ssh;
+   Ref<Buffer> file_buf;
+   Ref<FileSet> file_set;
+   Timer flush_timer;
 
    void Disconnect();
    int IsConnected() const
@@ -294,16 +285,16 @@ public:
 	    if(HasID())
 	       b->PackUINT32BE(id);
 	 }
-      virtual unpack_status_t Unpack(Buffer *b);
+      virtual unpack_status_t Unpack(const Buffer *b);
       virtual ~Packet() {}
       int GetLength() { return length; }
       packet_type GetPacketType() { return type; }
       const char *GetPacketTypeText();
-      unsigned GetID() { return id; }
+      unsigned GetID() const { return id; }
       void SetID(unsigned new_id) { id=new_id; }
       void DropData(Buffer *b) { b->Skip(4+(length>0?length:0)); }
       bool TypeIs(packet_type t) const { return type==t; }
-      static unpack_status_t UnpackString(Buffer *b,int *offset,int limit,char **str_out,int *len_out=0);
+      static unpack_status_t UnpackString(const Buffer *b,int *offset,int limit,xstring *str_out);
       static void PackString(Buffer *b,const char *str,int len=-1);
    };
 private:
@@ -314,7 +305,7 @@ private:
       unsigned data;
       PacketUINT32(packet_type t,unsigned d=0) : Packet(t)
 	 { data=d; length+=4; }
-      unpack_status_t Unpack(Buffer *b)
+      unpack_status_t Unpack(const Buffer *b)
 	 {
 	    unpack_status_t res;
 	    res=Packet::Unpack(b);
@@ -330,34 +321,25 @@ private:
    class PacketSTRING : public Packet
    {
    protected:
-      int string_len;
-      char *string;
+      xstring string;
       PacketSTRING(packet_type t) : Packet(t)
 	 {
-	    string_len=0;
-	    string=0;
 	    length=4;
 	 }
-      PacketSTRING(packet_type t,const char *s,int l=-1) : Packet(t)
+      PacketSTRING(packet_type t,const xstring &s) : Packet(t)
 	 {
-	    if(l==-1)
-	       l=strlen(s);
-	    string_len=l;
-	    string=(char*)xmalloc(l+1);
-	    memcpy(string,s,l);
-	    string[l]=0;
-	    length+=4+l;
+	    string.set(s);
+	    length+=4+string.length();
 	 }
-      ~PacketSTRING();
-      unpack_status_t Unpack(Buffer *b);
-      void ComputeLength() { Packet::ComputeLength(); length+=4+string_len; }
+      unpack_status_t Unpack(const Buffer *b);
+      void ComputeLength() { Packet::ComputeLength(); length+=4+string.length(); }
       void Pack(Buffer *b)
 	 {
 	    Packet::Pack(b);
-	    Packet::PackString(b,string,string_len);
+	    Packet::PackString(b,string,string.length());
 	 }
       const char *GetString() { return string; }
-      int GetStringLength() { return string_len; }
+      int GetStringLength() { return string.length(); }
    };
    class Request_INIT : public PacketUINT32
    {
@@ -370,7 +352,7 @@ private:
       char **extension_data;
    public:
       Reply_VERSION() : PacketUINT32(SSH_FXP_VERSION) {}
-      unpack_status_t Unpack(Buffer *b)
+      unpack_status_t Unpack(const Buffer *b)
 	 {
 	    unpack_status_t res;
 	    res=PacketUINT32::Unpack(b);
@@ -391,7 +373,7 @@ private:
       unsigned flags;
       int protocol_version;
    public:
-      Request_FSTAT(const char *h,int len,unsigned f,int pv) : PacketSTRING(SSH_FXP_FSTAT,h,len)
+      Request_FSTAT(const xstring &h,unsigned f,int pv) : PacketSTRING(SSH_FXP_FSTAT,h)
 	 {
 	    flags=f;
 	    protocol_version=pv;
@@ -412,7 +394,7 @@ private:
    class Request_STAT : public Request_FSTAT
    {
    public:
-      Request_STAT(const char *p,unsigned f,int pv) : Request_FSTAT(p,strlen(p),f,pv)
+      Request_STAT(const char *p,unsigned f,int pv) : Request_FSTAT(p,f,pv)
 	 {
 	    type=SSH_FXP_STAT;
 	 }
@@ -423,11 +405,9 @@ public:
    {
       struct ExtFileAttr
       {
-	 char *extended_type;
-	 char *extended_data;
-	 ExtFileAttr() { extended_type=extended_data=0; }
-	 ~ExtFileAttr() { xfree(extended_type); xfree(extended_data); }
-	 unpack_status_t Unpack(Buffer *b,int *offset,int limit);
+	 xstring extended_type;
+	 xstring extended_data;
+	 unpack_status_t Unpack(const Buffer *b,int *offset,int limit);
 	 void Pack(Buffer *b);
       };
       struct FileACE
@@ -435,18 +415,17 @@ public:
 	 unsigned ace_type;
 	 unsigned ace_flag;
 	 unsigned ace_mask;
-	 char     *who;
-	 FileACE() { ace_type=ace_flag=ace_mask=0; who=0; }
-	 ~FileACE() { xfree(who); }
-	 unpack_status_t Unpack(Buffer *b,int *offset,int limit);
+	 xstring who;
+	 FileACE() { ace_type=ace_flag=ace_mask=0; }
+	 unpack_status_t Unpack(const Buffer *b,int *offset,int limit);
 	 void Pack(Buffer *b);
       };
 
       unsigned flags;
       int      type;		    // v>=4
       off_t    size;		    // present only if flag SIZE
-      char     *owner;		    // present only if flag OWNERGROUP // v>=4
-      char     *group;		    // present only if flag OWNERGROUP // v>=4
+      xstring  owner;		    // present only if flag OWNERGROUP // v>=4
+      xstring  group;		    // present only if flag OWNERGROUP // v>=4
       uid_t    uid;		    // present only if flag UIDGID // v<=3
       gid_t    gid;		    // present only if flag UIDGID // v<=3
       unsigned permissions;	    // present only if flag PERMISSIONS
@@ -463,36 +442,34 @@ public:
       unsigned attrib_bits;         // if flag BITS		  // v>=5
       unsigned attrib_bits_valid;   // if flag BITS		  // v>=6
       unsigned char text_hint;      // if flag TEXT_HINT	  // v>=6
-      char     *mime_type;          // if flag MIME_TYPE	  // v>=6
+      xstring  mime_type;	    // if flag MIME_TYPE	  // v>=6
       unsigned link_count;          // if flag LINK_COUNT	  // v>=6
-      char     *untranslated_name;  // if flag UNTRANSLATED_NAME  // v>=6
+      xstring  untranslated_name;   // if flag UNTRANSLATED_NAME  // v>=6
       unsigned extended_count;	    // present only if flag EXTENDED
       ExtFileAttr *extended_attrs;
 
       FileAttrs()
       {
-	 flags=0; type=0; size=NO_SIZE; owner=group=0; uid=gid=0;
+	 flags=0; type=0; size=NO_SIZE; uid=gid=0;
 	 permissions=0;
 	 atime=createtime=mtime=ctime=NO_DATE;
 	 atime_nseconds=createtime_nseconds=mtime_nseconds=ctime_nseconds=0;
 	 extended_count=0; extended_attrs=0;
 	 ace_count=0; ace=0;
 	 attrib_bits=attrib_bits_valid=0; text_hint=0;
-	 mime_type=untranslated_name=0; link_count=0;
+	 link_count=0;
       }
       ~FileAttrs();
-      unpack_status_t Unpack(Buffer *b,int *offset,int limit,int proto_version);
+      unpack_status_t Unpack(const Buffer *b,int *offset,int limit,int proto_version);
       void Pack(Buffer *b,int proto_version);
       int ComputeLength(int v);
    };
    struct NameAttrs
    {
-      char *name;
-      char *longname;
+      xstring name;
+      xstring longname;
       FileAttrs attrs;
-      NameAttrs() { name=0; longname=0; }
-      ~NameAttrs() { xfree(name); xfree(longname); }
-      unpack_status_t Unpack(Buffer *b,int *offset,int limit,int proto_version);
+      unpack_status_t Unpack(const Buffer *b,int *offset,int limit,int proto_version);
    };
 private:
    class Reply_NAME : public Packet
@@ -504,7 +481,7 @@ private:
    public:
       Reply_NAME(int pv) : Packet(SSH_FXP_NAME) { protocol_version=pv; eof=false; }
       ~Reply_NAME() { delete[] names; }
-      unpack_status_t Unpack(Buffer *b);
+      unpack_status_t Unpack(const Buffer *b);
       int GetCount() { return count; }
       const NameAttrs *GetNameAttrs(int index)
 	 {
@@ -520,7 +497,7 @@ private:
       FileAttrs attrs;
    public:
       Reply_ATTRS(int pv) : Packet(SSH_FXP_ATTRS) { protocol_version=pv; }
-      unpack_status_t Unpack(Buffer *b);
+      unpack_status_t Unpack(const Buffer *b);
       const FileAttrs *GetAttrs() { return &attrs; }
    };
    class PacketSTRING_ATTRS : public PacketSTRING
@@ -529,8 +506,8 @@ private:
       int protocol_version;
    public:
       FileAttrs attrs;
-      PacketSTRING_ATTRS(packet_type type,const char *s,int len,int pv)
-       : PacketSTRING(type,s,len)
+      PacketSTRING_ATTRS(packet_type type,const xstring &h,int pv)
+       : PacketSTRING(type,h)
 	 {
 	    protocol_version=pv;
 	 }
@@ -548,8 +525,8 @@ private:
    class Request_FSETSTAT : public PacketSTRING_ATTRS
    {
    public:
-      Request_FSETSTAT(const char *h,int h_len,int pv)
-       : PacketSTRING_ATTRS(SSH_FXP_FSETSTAT,h,h_len,pv) {}
+      Request_FSETSTAT(const xstring &h,int pv)
+       : PacketSTRING_ATTRS(SSH_FXP_FSETSTAT,h,pv) {}
    };
    class Request_OPEN : public PacketSTRING_ATTRS
    {
@@ -558,7 +535,7 @@ private:
       unsigned flags;		 // v>=5
    public:
       Request_OPEN(const char *fn,unsigned pf,unsigned da,unsigned f,int pv)
-       : PacketSTRING_ATTRS(SSH_FXP_OPEN,fn,strlen(fn),pv)
+       : PacketSTRING_ATTRS(SSH_FXP_OPEN,fn,pv)
 	 {
 	    pflags=pf;
 	    desired_access=da;
@@ -575,18 +552,12 @@ private:
    {
    public:
       Reply_HANDLE() : PacketSTRING(SSH_FXP_HANDLE) {}
-      char *GetHandle(int *out_len=0)
-	 {
-	    char *out=(char*)xmemdup(string,string_len+1);
-	    if(out_len)
-	       *out_len=string_len;
-	    return out;
-	 }
+      const xstring &GetHandle() { return string; }
    };
    class Request_CLOSE : public PacketSTRING
    {
    public:
-      Request_CLOSE(const char *h,int len) : PacketSTRING(SSH_FXP_CLOSE,h,len) {}
+      Request_CLOSE(const xstring &h) : PacketSTRING(SSH_FXP_CLOSE,h) {}
    };
    class Request_OPENDIR : public PacketSTRING
    {
@@ -596,18 +567,17 @@ private:
    class Request_READDIR : public PacketSTRING
    {
    public:
-      Request_READDIR(const char *h,int len) : PacketSTRING(SSH_FXP_READDIR,h,len) {}
+      Request_READDIR(const xstring &h) : PacketSTRING(SSH_FXP_READDIR,h) {}
    };
    class Reply_STATUS : public Packet
    {
       int protocol_version;
       unsigned code;
-      char *message;
-      char *language;
+      xstring message;
+      xstring language;
    public:
-      Reply_STATUS(int pv) { protocol_version=pv; code=0; message=0; language=0; }
-      ~Reply_STATUS() { xfree(message); xfree(language); }
-      unpack_status_t Unpack(Buffer *b);
+      Reply_STATUS(int pv) { protocol_version=pv; code=0; }
+      unpack_status_t Unpack(const Buffer *b);
       int GetCode() { return code; }
       const char *GetCodeText();
       const char *GetMessage() { return message; }
@@ -617,8 +587,8 @@ private:
    public:
       off_t pos;
       unsigned len;
-      Request_READ(const char *h,int hlen,off_t p,unsigned l)
-       : PacketSTRING(SSH_FXP_READ,h,hlen) { pos=p; len=l; }
+      Request_READ(const xstring &h,off_t p,unsigned l)
+       : PacketSTRING(SSH_FXP_READ,h) { pos=p; len=l; }
       void ComputeLength() { PacketSTRING::ComputeLength(); length+=8+4; }
       void Pack(Buffer *b);
    };
@@ -627,33 +597,31 @@ private:
       bool eof;
    public:
       Reply_DATA() : PacketSTRING(SSH_FXP_DATA) { eof=false; }
-      void GetData(const char **b,int *s) { *b=string; *s=string_len; }
-      unpack_status_t Unpack(Buffer *b);
+      void GetData(const char **b,int *s) { *b=string; *s=string.length(); }
+      unpack_status_t Unpack(const Buffer *b);
       bool Eof() { return eof; }
    };
    class Request_WRITE : public PacketSTRING
    {
    public:
       off_t pos;
-      unsigned len;
-      char *data;
-      Request_WRITE(const char *h,int hlen,off_t p,const char *d,unsigned l)
-       : PacketSTRING(SSH_FXP_WRITE,h,hlen) { pos=p; len=l; data=(char*)xmemdup(d,l); }
-      ~Request_WRITE() { xfree(data); }
-      void ComputeLength() { PacketSTRING::ComputeLength(); length+=8+4+len; }
+      xstring data;
+      Request_WRITE(const xstring &h,off_t p,const char *d,unsigned l)
+       : PacketSTRING(SSH_FXP_WRITE,h) { pos=p; data.nset(d,l); }
+      void ComputeLength() { PacketSTRING::ComputeLength(); length+=8+4+data.length(); }
       void Pack(Buffer *b);
    };
    class Request_MKDIR : public PacketSTRING_ATTRS
    {
    public:
       Request_MKDIR(const char *fn,int pv)
-       : PacketSTRING_ATTRS(SSH_FXP_MKDIR,fn,strlen(fn),pv) {}
+       : PacketSTRING_ATTRS(SSH_FXP_MKDIR,fn,pv) {}
    };
    class Request_SETSTAT : public PacketSTRING_ATTRS
    {
    public:
       Request_SETSTAT(const char *fn,int pv)
-       : PacketSTRING_ATTRS(SSH_FXP_SETSTAT,fn,strlen(fn),pv) {}
+       : PacketSTRING_ATTRS(SSH_FXP_SETSTAT,fn,pv) {}
    };
    class Request_RMDIR : public PacketSTRING
    {
@@ -668,22 +636,16 @@ private:
    class Request_RENAME : public Packet
    {
       int protocol_version;
-      char *oldpath;
-      char *newpath;
+      xstring oldpath;
+      xstring newpath;
       unsigned flags;
    public:
-      Request_RENAME(const char *o,const char *n,unsigned f,int pv) : Packet(SSH_FXP_RENAME)
+      Request_RENAME(const char *o,const char *n,unsigned f,int pv)
+      : Packet(SSH_FXP_RENAME), oldpath(o), newpath(n)
 	 {
 	    protocol_version=pv;
-	    oldpath=xstrdup(o);
-	    newpath=xstrdup(n);
 	    flags=f;
 	 }
-      ~Request_RENAME()
-	 {
-	    xfree(oldpath);
-	    xfree(newpath);
-	 }
       void ComputeLength()
 	 {
 	    Packet::ComputeLength();
@@ -710,13 +672,12 @@ private:
 	 IGNORE
       };
 
-      Packet *request;
-      Packet *reply;
+      Ref<Packet> request;
+      Ref<Packet> reply;
       Expect *next;
       int i;
       expect_t tag;
-      Expect(Packet *req,expect_t t,int j=0) { request=req; tag=t; reply=0; i=j; }
-      ~Expect() { delete request; delete reply; }
+      Expect(Packet *req,expect_t t,int j=0) : request(req), i(j), tag(t) {}
    };
 
    void PushExpect(Expect *);
@@ -762,6 +723,7 @@ private:
    int max_packets_in_flight_slow_start;
    int size_read;
    int size_write;
+   bool use_full_path;
 
 protected:
    void SetError(int code,const Packet *reply);
@@ -806,21 +768,19 @@ public:
    void Cleanup();
    void CleanupThis();
 
-   FileSet *GetFileSet() { FileSet *fset=file_set; file_set=0; return fset?fset:new FileSet; }
+   FileSet *GetFileSet() { FileSet *fset=file_set.borrow(); return fset?fset:new FileSet; }
 };
 
 class SFtpDirList : public DirList
 {
-   FileAccess *session;
-   IOBuffer *ubuf;
+   Ref<IOBuffer> ubuf;
    const char *dir;
    bool use_file_set;
-   FileSet *fset;
+   Ref<FileSet> fset;
    LsOptions ls_options;
 
 public:
-   SFtpDirList(ArgV *a,FileAccess *fa);
-   ~SFtpDirList();
+   SFtpDirList(SFtp *s,ArgV *a);
    const char *Status();
    int Do();
 
@@ -830,14 +790,11 @@ public:
 
 class SFtpListInfo : public ListInfo
 {
-   IOBuffer *ubuf;
+   SMTaskRef<IOBuffer> ubuf;
+
 public:
    SFtpListInfo(SFtp *session,const char *dir)
-      : ListInfo(session,dir)
-      {
-	 ubuf=0;
-      }
-   ~SFtpListInfo() { Delete(ubuf); }
+      : ListInfo(session,dir) {}
    int Do();
    const char *Status();
 };
