@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2005 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2008 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: SMTask.cc,v 1.48 2006/09/27 07:03:38 lav Exp $ */
+/* $Id: SMTask.cc,v 1.53 2008/01/21 14:07:01 lav Exp $ */
 
 #include <config.h>
 #include <assert.h>
@@ -35,10 +35,8 @@
 
 SMTask	 *SMTask::chain=0;
 SMTask	 *SMTask::current=0;
-SMTask	 **SMTask::stack=0;
-int      SMTask::stack_ptr=0;
-int      SMTask::stack_size=0;
-PollVec	 SMTask::sched_total;
+xarray<SMTask*>	SMTask::stack;
+PollVec	 SMTask::block;
 TimeDate SMTask::now;
 
 static int task_count=0;
@@ -112,7 +110,7 @@ SMTask::~SMTask()
    {
       fprintf(stderr,"SMTask(%p).running=%d\n",this,running);
       fprintf(stderr,"SMTask stack:");
-      for(int i=0; i<stack_ptr; i++)
+      for(int i=0; i<stack.count(); i++)
 	 fprintf(stderr," %p",stack[i]);
       fprintf(stderr,"; current=%p\n",current);
       abort();
@@ -142,9 +140,7 @@ void SMTask::Delete(SMTask *task)
 
 void SMTask::Enter(SMTask *task)
 {
-   if(stack_size<=stack_ptr)
-      stack=(SMTask**)xrealloc(stack,(stack_size+=16)*sizeof(*stack));
-   stack[stack_ptr++]=current;
+   stack.append(current);
    current=task;
    current->running++;
 }
@@ -152,7 +148,9 @@ void SMTask::Leave(SMTask *task)
 {
    assert(current==task);
    current->running--;
-   current=stack[--stack_ptr];
+   assert(stack.count()>0);
+   current=stack.last();
+   stack.chop();
 }
 
 int SMTask::Roll(SMTask *task)
@@ -171,7 +169,7 @@ void SMTask::RollAll(const TimeInterval &max_time)
 {
    Timer limit_timer(max_time);
    do { Schedule(); }
-   while(sched_total.GetTimeout()==0 && !limit_timer.Stopped());
+   while(block.GetTimeout()==0 && !limit_timer.Stopped());
 }
 
 int SMTask::CollectGarbage()
@@ -192,7 +190,7 @@ int SMTask::CollectGarbage()
 	 count++;
 	 if(scan->next)
 	 {
-	    Enter(scan->next); // protect it from deleting
+	    Enter(scan->next); // protect it from deleting (in scan's dtor)
 	    delete scan;
 	    Leave(scan=current);
 	 }
@@ -210,67 +208,26 @@ void SMTask::Schedule()
 {
    SMTask *scan;
 
-   for(scan=chain; scan; scan=scan->next)
-   {
-      if(!scan->running)
-	 scan->block.Empty();
-   }
-   sched_total.Empty();
+   block.Empty();
 
    // get time once and assume Do() don't take much time
    UpdateNow();
 
    int timer_timeout=Timer::GetTimeout();
    if(timer_timeout>=0)
-      sched_total.SetTimeout(timer_timeout);
+      block.SetTimeout(timer_timeout);
 
-   bool repeat=false;
-   scan=chain;
-   while(scan)
-   {
-      if(scan->running || scan->IsSuspended())
-      {
-	 scan=scan->next;
-	 continue;
-      }
-
-      if(repeat)
-	 scan->block.SetTimeout(0); // little optimization
-
-      Enter(scan);	   // mark it current and running.
-      int res=scan->Do();  // let it run.
-      Leave(scan);	   // unmark it running and change current.
-
-      if(res==MOVED || scan->block.GetTimeout()==0)
-	 repeat=true;
-      scan=scan->next;	   // move to a next task.
-   }
-   if(CollectGarbage())
-      repeat=true;
-   if(repeat)
-   {
-      sched_total.SetTimeout(0);
-      return;
-   }
-
-   // now collect all wake up conditions; excluding suspended and
-   // already running tasks, because they can't run again on this
-   // level of recursion, and thus could cause spinning by their wake up
-   // conditions.
+   int res=STALL;
    for(scan=chain; scan; scan=scan->next)
    {
-      if(!scan->IsSuspended() && !scan->running && !scan->block.IsEmpty())
-	 sched_total.Merge(scan->block);
+      if(scan->running || scan->IsSuspended())
+	 continue;
+      Enter(scan);	   // mark it current and running.
+      res|=scan->Do();	   // let it run.
+      Leave(scan);	   // unmark it running and change current.
    }
-}
-
-void SMTask::ReconfigAll(const char *name)
-{
-   UpdateNow();
-   Timer::ReconfigAll(name);
-   for(SMTask *scan=chain; scan; scan=scan->next)
-      scan->Reconfig(name);
-   sched_total.SetTimeout(0);  // for new values handling
+   if(CollectGarbage() || res)
+      block.NoWait();
 }
 
 int SMTaskInit::Do()
@@ -343,7 +300,7 @@ void SMTask::PrintTasks()
    {
       const char *c=scan->GetLogContext();
       if(!c) c="";
-      printf("%p\t%c%c%c\t%s\n",scan,scan->running?'R':' ',
-	 scan->suspended?'S':' ',scan->deleting?'D':' ',c);
+      printf("%p\t%c%c%c\t%d\t%s\n",scan,scan->running?'R':' ',
+	 scan->suspended?'S':' ',scan->deleting?'D':' ',scan->ref_count,c);
    }
 }
