@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1998-2005 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1998-2008 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,23 +18,24 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: SleepJob.cc,v 1.28 2006/07/05 08:52:36 lav Exp $ */
+/* $Id: SleepJob.cc,v 1.36 2008/01/21 14:07:01 lav Exp $ */
 
 #include <config.h>
 #include <ctype.h>
+#include <stddef.h>
 #include "SleepJob.h"
 #include "CmdExec.h"
 #include "misc.h"
 #include "LocalDir.h"
 
 SleepJob::SleepJob(const TimeInterval &when,FileAccess *s,LocalDirectory *cwd,char *what)
-   : SessionJob(s), Timer(when)
+   : SessionJob(s), Timer(when), saved_cwd(cwd)
 {
-   cmd=what;
+   cmd.set_allocated(what);
    exit_code=0;
    done=false;
-   saved_cwd=cwd;
    repeat=false;
+   weak=false;
    repeat_count=0;
    max_repeat_count=0;
    exec=0;
@@ -44,8 +45,6 @@ SleepJob::SleepJob(const TimeInterval &when,FileAccess *s,LocalDirectory *cwd,ch
 SleepJob::~SleepJob()
 {
    Delete(exec);
-   xfree(cmd);
-   delete saved_cwd;
 }
 
 int SleepJob::Do()
@@ -53,7 +52,7 @@ int SleepJob::Do()
    if(Done())
       return STALL;
 
-   if(waiting_num>0)
+   if(waiting.count()>0)
    {
       Job *j=FindDoneAwaitedJob();
       if(!j)
@@ -79,13 +78,10 @@ int SleepJob::Do()
       {
 	 if(!exec)
 	 {
-	    exec=new CmdExec(session,saved_cwd);
-	    session=0;
-	    saved_cwd=0;
+	    exec=new CmdExec(session.borrow(),saved_cwd.borrow());
 	    exec->SetParentFg(this);
 	    exec->AllocJobno();
-	    exec->cmdline=(char*)xmalloc(3+strlen(cmd));
-	    sprintf(exec->cmdline,"(%s)",cmd);
+	    exec->cmdline.vset("(",cmd.get(),")",NULL);
 	 }
 	 exec->FeedCmd(cmd);
 	 exec->FeedCmd("\n");
@@ -110,7 +106,7 @@ void SleepJob::PrintStatus(int,const char *prefix)
 
 void SleepJob::lftpMovesToBackground()
 {
-   if(IsInfty() || (repeat && cmd[0]==0))
+   if(weak || IsInfty() || (repeat && cmd[0]==0))
    {
       // terminate
       done=true;
@@ -149,6 +145,7 @@ Job *cmd_repeat(CmdExec *parent)
    const char *delay_str=0;
    bool while_ok=false;
    bool until_ok=false;
+   bool weak=false;
    int opt;
 
    static struct option repeat_opts[]=
@@ -157,6 +154,7 @@ Job *cmd_repeat(CmdExec *parent)
       {"count",required_argument,0,'c'},
       {"while-ok",no_argument,0,'o'},
       {"until-ok",no_argument,0,'O'},
+      {"weak",no_argument,0,'w'},
       {0},
    };
 
@@ -177,6 +175,9 @@ Job *cmd_repeat(CmdExec *parent)
       case('O'):
 	 until_ok=true;
 	 break;
+      case('w'):
+	 weak=true;
+	 break;
       case('?'):
 	 eprintf(_("Try `help %s' for more information.\n"),args->a0());
 	 return 0;
@@ -207,6 +208,7 @@ Job *cmd_repeat(CmdExec *parent)
 	        ? args->Combine(cmd_start) : args->CombineQuoted(cmd_start));
    SleepJob *s=new SleepJob(delay,session->Clone(),parent->cwd->Clone(),cmd);
    s->Repeat(max_count);
+   s->SetWeak(weak);
    if(while_ok)
       s->ContinueCode(0);
    if(until_ok)
@@ -236,17 +238,10 @@ Job *cmd_at(CmdExec *parent)
       count++;
    }
 
-#if 0
-   char **av=(char**)xmemdup(args->GetV(),(count+1)*sizeof(char**));
-   av[count]=0;
-   time_t when=parsetime(count-1,av+1);
-   xfree(av);
-#endif
-   char *date=args->Combine(1);
-   date[date_len]=0;
+   xstring_ca date(args->Combine(1));
+   date.truncate(date_len);
    time_t now=SMTask::now;
    time_t when=get_date(date,&now);
-   xfree(date);
 
    if(when==0 || when==(time_t)-1)
    {
