@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 2001 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 2001-2007 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,12 +21,12 @@
 #ifndef STRINGPOOL_H
 #define STRINGPOOL_H
 
+#include "xarray.h"
+
 // maybe it is better to have many separate pools with dtors.
 class StringPool
 {
-   static char **strings;
-   static int n_strings;
-   static int allocated;
+   static xarray_p<char> strings;
 
 public:
    static const char *Get(const char *);
