@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1998-2005 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1998-2008 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,7 +21,6 @@
 #include <config.h>
 #include <errno.h>
 #include "buffer.h"
-#include "xmalloc.h"
 #include "FileAccess.h"
 #include "misc.h"
 #include "trio.h"
@@ -30,19 +29,17 @@
 
 #define BUFFER_INC	   (8*1024) // should be power of 2
 
-void Buffer::Get(const char **buf,int *size)
+const char *Buffer::Get() const
 {
-   if(in_buffer==0)
-   {
-      *size=0;
-      if(eof)
-	 *buf=0;
-      else
-	 *buf="";
-      return;
-   }
-   *buf=buffer+buffer_ptr;
-   *size=in_buffer;
+   if(Size()==0)
+      return eof?0:"";
+   return buffer+buffer_ptr;
+}
+
+void Buffer::Get(const char **buf,int *size) const
+{
+   *size=Size();
+   *buf=Get();
 }
 
 void Buffer::GetSaved(const char **buf,int *size) const
@@ -54,45 +51,45 @@ void Buffer::GetSaved(const char **buf,int *size) const
       return;
    }
    *buf=buffer;
-   *size=buffer_ptr+in_buffer;
+   *size=buffer.length();
 }
 
-void Buffer::SaveRollback(int p)
+void Buffer::SaveRollback(off_t p)
 {
+   pos=p;
    if(buffer_ptr<p)
       save=false;
    if(!save)
-   {
-      buffer_ptr=0;
-      in_buffer=0;
-   }
-   else
-   {
-      buffer_ptr=p;
-      in_buffer=0;
-   }
+      p=0;
+   buffer.truncate(buffer_ptr=p);
 }
 
 void Buffer::Allocate(int size)
 {
-   if(in_buffer==0 && !save)
+   if(buffer_ptr>0 && Size()==0 && !save)
+   {
+      buffer.truncate(0);
       buffer_ptr=0;
+   }
 
-   int in_buffer_real=in_buffer;
-   if(save)
+   size_t in_buffer_real=Size();
+   /* disable data movement to beginning of the buffer, if:
+      1. we save the data explicitly;
+      2. we add more data than there is space in the beginning of the buffer
+	 (because the probability of realloc is high anyway);
+      3. the gap at beginning is smaller than the amount of data in the buffer
+	 (because the penalty of data movement is high). */
+   if(save || buffer_ptr<size || buffer_ptr<Size())
       in_buffer_real+=buffer_ptr;
 
-   if(buffer_allocated<in_buffer_real+size)
-   {
-      buffer_allocated=(in_buffer_real+size+(BUFFER_INC-1)) & ~(BUFFER_INC-1);
-      buffer=(char*)xrealloc(buffer,buffer_allocated);
-   }
    // could be round-robin, but this is easier
-   if(!save && buffer_ptr+in_buffer+size>buffer_allocated)
+   if(buffer.length()>in_buffer_real)
    {
-      memmove(buffer,buffer+buffer_ptr,in_buffer);
+      buffer.nset(buffer+buffer_ptr,Size());
       buffer_ptr=0;
    }
+
+   buffer.get_space(in_buffer_real+size,BUFFER_INC);
 }
 
 void Buffer::SaveMaxCheck(int size)
@@ -101,35 +98,47 @@ void Buffer::SaveMaxCheck(int size)
       save=false;
 }
 
-void Buffer::Put(const char *buf,int size)
+void Buffer::Append(const char *buf,int size)
 {
-   SaveMaxCheck(size);
-
-   if(in_buffer==0 && !save)
-   {
-      buffer_ptr=0;
-
-      if(size>=PUT_LL_MIN)
-      {
-	 int res=Put_LL(buf,size);
-	 if(res>=0)
-	 {
-	    buf+=res;
-	    size-=res;
-	    pos+=res;
-	 }
-      }
-   }
-
    if(size==0)
       return;
 
-   Allocate(size);
+   SaveMaxCheck(size);
+   if(Size()==0 && buffer_ptr>0 && !save)
+   {
+      buffer.truncate(0);
+      buffer_ptr=0;
+   }
 
-   memcpy(buffer+buffer_ptr+in_buffer,buf,size);
-   in_buffer+=size;
+   memmove(GetSpace(size),buf,size);
+   SpaceAdd(size);
+}
+void Buffer::Put(const char *buf,int size)
+{
+   Append(buf,size);
    pos+=size;
 }
+void Buffer::Prepend(const char *buf,int size)
+{
+   if(size==0)
+      return;
+   save=false;
+   if(Size()==0)
+   {
+      memmove(GetSpace(size),buf,size);
+      SpaceAdd(size);
+      return;
+   }
+   if(buffer_ptr<size)
+   {
+      Allocate(size-buffer_ptr);
+      memmove(buffer.get_non_const()+size,buffer+buffer_ptr,Size());
+      SpaceAdd(size-buffer_ptr);
+      buffer_ptr=size;
+   }
+   memmove(buffer.get_non_const()+buffer_ptr-size,buf,size);
+   buffer_ptr-=size;
+}
 
 void Buffer::Format(const char *f,...)
 {
@@ -144,14 +153,13 @@ void Buffer::vFormat(const char *f, va_list v)
    int size=64;
    for(;;)
    {
-      Allocate(size);
       va_list tmp;
       VA_COPY(tmp,v);
-      int res=vsnprintf(buffer+buffer_ptr+in_buffer, size, f, v);
+      int res=vsnprintf(GetSpace(size), size, f, tmp);
       va_end(tmp);
       if(res>=0 && res<size)
       {
-	 in_buffer+=res;
+	 SpaceAdd(res);
 	 return;
       }
       if(res>size)   // some vsnprintf's return desired buffer size.
@@ -163,9 +171,8 @@ void Buffer::vFormat(const char *f, va_list v)
 
 void Buffer::Skip(int len)
 {
-   if(len>in_buffer)
-      len=in_buffer;
-   in_buffer-=len;
+   if(len>Size())
+      len=Size();
    buffer_ptr+=len;
    pos+=len;
 }
@@ -173,14 +180,13 @@ void Buffer::UnSkip(int len)
 {
    if(len>buffer_ptr)
       len=buffer_ptr;
-   in_buffer+=len;
    buffer_ptr-=len;
    pos-=len;
 }
 
 void Buffer::Empty()
 {
-   in_buffer=0;
+   buffer.truncate(0);
    buffer_ptr=0;
    if(save_max>0)
       save=true;
@@ -188,25 +194,17 @@ void Buffer::Empty()
 
 Buffer::Buffer()
 {
-   error_text=0;
    saved_errno=0;
    error_fatal=false;
-   buffer=0;
-   buffer_allocated=0;
-   in_buffer=0;
    buffer_ptr=0;
    eof=false;
    broken=false;
    save=false;
    save_max=0;
    pos=0;
-   rate=0;
 }
 Buffer::~Buffer()
 {
-   xfree(error_text);
-   xfree(buffer);
-   SMTask::Delete(rate);
 }
 
 const char *Buffer::GetRateStrS()
@@ -224,39 +222,91 @@ void Buffer::RateAdd(int n)
 
 void Buffer::SetError(const char *e,bool fatal)
 {
-   xfree(error_text);
-   error_text=xstrdup(e);
+   error_text.set(e);
    error_fatal=fatal;
 }
 void Buffer::SetErrorCached(const char *e)
 {
-   xfree(error_text);
-   error_text=xasprintf(_("%s [cached]"),e);
-   error_fatal=false;
+   SetError(e,false);
+   error_text.append(_(" [cached]"));
 }
-#if 0 // unused
-void Buffer::SetError2(const char *e1,const char *e2)
+
+void DataTranslator::AppendTranslated(Buffer *target,const char *put_buf,int size)
 {
-   xfree(error_text);
-   int len1=strlen(e1);
-   int len2=strlen(e2);
-   error_text=(char*)xmalloc(len1+len2+1);
-   strcpy(error_text,e1);
-   strcat(error_text,e2);
+   off_t old_pos=target->GetPos();
+   PutTranslated(target,put_buf,size);
+   target->SetPos(old_pos);
 }
-#endif
 
 #ifdef HAVE_ICONV
-void DirectedBuffer::SetTranslation(const char *enc,bool translit)
+void DataRecoder::PutTranslated(Buffer *target,const char *put_buf,int size)
+{
+   if(!backend_translate)
+   {
+      target->Put(put_buf,size);
+      return;
+   }
+   bool from_untranslated=false;
+   if(Size()>0)
+   {
+      Put(put_buf,size);
+      Get(&put_buf,&size);
+      from_untranslated=true;
+   }
+   if(size<=0)
+      return;
+   size_t put_size=size;
+
+   int size_coeff=6;
+try_again:
+   if(put_size==0)
+      return;
+   size_t store_size=size_coeff*put_size;
+   char *store_space=target->GetSpace(store_size);
+   char *store_buf=store_space;
+   const char *base_buf=put_buf;
+   // do the translation
+   ICONV_CONST char **put_buf_ptr=const_cast<ICONV_CONST char**>(&put_buf);
+   size_t res=iconv(backend_translate,put_buf_ptr,&put_size,&store_buf,&store_size);
+   target->SpaceAdd(store_buf-store_space);
+   if(from_untranslated)
+      Skip(put_buf-base_buf);
+   if(res==(size_t)-1)
+   {
+      switch(errno)
+      {
+      case EINVAL: // incomplete character
+	 if(!from_untranslated)
+	    Put(put_buf,put_size);
+	 break;
+      case EILSEQ: // invalid character
+	 target->Put("?");
+	 put_buf++;
+	 put_size--;
+	 goto try_again;
+      case E2BIG:  // no room to store result, allocate more.
+	 size_coeff*=2;
+	 goto try_again;
+      default:
+	 break;
+      }
+   }
+   return;
+}
+void DataRecoder::ResetTranslation()
+{
+   Empty();
+   if(!backend_translate)
+      return;
+   iconv(backend_translate,0,0,0,0);
+}
+DataRecoder::~DataRecoder()
 {
    if(backend_translate)
       iconv_close(backend_translate);
-   backend_translate=0;
-   if(!enc)
-      return;
-   const char *local_code=ResMgr::Query("file:charset",0);
-   const char *from_code=(mode==GET?enc:local_code);
-   const char *to_code  =(mode==GET?local_code:enc);
+}
+DataRecoder::DataRecoder(const char *from_code,const char *to_code,bool translit)
+{
    if(translit)
    {
       const char *add="//TRANSLIT";
@@ -272,102 +322,78 @@ void DirectedBuffer::SetTranslation(const char *enc,bool translit)
       backend_translate=0;
    }
 }
-DirectedBuffer::~DirectedBuffer()
-{
-   SetTranslation(0);
-   delete untranslated;
-}
-void DirectedBuffer::ResetTranslation()
-{
-   if(!backend_translate)
-      return;
-   iconv(backend_translate,0,0,0,0);
-   delete untranslated;
-   untranslated=0;
-}
-void DirectedBuffer::PutTranslated(const char *put_buf,int size)
-{
-   if(!backend_translate)
-   {
-      Buffer::Put(put_buf,size);
-      return;
-   }
-   bool from_untranslated=false;
-   if(untranslated && untranslated->Size()>0)
-   {
-      untranslated->Put(put_buf,size);
-      untranslated->Get(&put_buf,&size);
-      from_untranslated=true;
-   }
-   if(size<=0)
-      return;
-   size_t put_size=size;
 
-try_again:
-   if(put_size==0)
+void DirectedBuffer::SetTranslation(const char *enc,bool translit)
+{
+   if(!enc)
       return;
-   size_t store_size=6*put_size;
-   Allocate(store_size);
-   char *store_buf=buffer+buffer_ptr+in_buffer;
-   store_size=buffer_allocated-(store_buf-buffer);
-   const char *base_buf=put_buf;
-   // do the translation
-   ICONV_CONST char **put_buf_ptr=const_cast<ICONV_CONST char**>(&put_buf);
-   size_t res=iconv(backend_translate,put_buf_ptr,&put_size,&store_buf,&store_size);
-   in_buffer+=store_buf-(buffer+buffer_ptr+in_buffer);
-   if(from_untranslated)
-      untranslated->Skip(put_buf-base_buf);
-   if(res==(size_t)-1)
-   {
-      switch(errno)
-      {
-      case EINVAL: // incomplete character
-	 if(!from_untranslated)
-	 {
-	    if(!untranslated)
-	       untranslated=new Buffer;
-	    untranslated->Put(put_buf,put_size);
-	 }
-	 break;
-      case EILSEQ: // invalid character
-	 Buffer::Put("?");
-	 put_buf++;
-	 put_size--;
-	 goto try_again;
-      case E2BIG:  // no room to store result
-	 if(store_size>=6*put_size)
-	    Allocate(store_size*2);
-	 goto try_again;
-      default:
-	 break;
-      }
-   }
-   return;
+   const char *local_code=ResMgr::Query("file:charset",0);
+   const char *from_code=(mode==GET?enc:local_code);
+   const char *to_code  =(mode==GET?local_code:enc);
+   SetTranslator(new DataRecoder(from_code,to_code,translit));
 }
 #endif //HAVE_ICONV
 
+void DirectedBuffer::ResetTranslation()
+{
+   if(translator)
+      translator->ResetTranslation();
+}
+void DirectedBuffer::Put(const char *buf,int size)
+{
+   if(mode==PUT && translator)
+      translator->PutTranslated(this,buf,size);
+   else
+      Buffer::Put(buf,size);
+}
+void DirectedBuffer::PutTranslated(const char *buf,int size)
+{
+   if(translator)
+      translator->PutTranslated(this,buf,size);
+   else
+      Buffer::Put(buf,size);
+}
+
 void DirectedBuffer::EmbraceNewData(int len)
 {
    if(len<=0)
       return;
    RateAdd(len);
-#ifdef HAVE_ICONV
-   if(backend_translate)
+   if(translator)
    {
-      if(!untranslated)
-	 untranslated=new Buffer;
       // copy the data to free room for translated data
-      untranslated->Put(buffer+buffer_ptr+in_buffer,len);
-      PutTranslated(0,0);
+      translator->Put(buffer+buffer.length(),len);
+      translator->AppendTranslated(this,0,0);
    }
    else
-#endif
-   {
-      in_buffer+=len;
-   }
+      SpaceAdd(len);
    SaveMaxCheck(0);
 }
 
+
+IOBuffer::IOBuffer(dir_t m)
+   : DirectedBuffer(m), event_time(now)
+{
+}
+IOBuffer::~IOBuffer()
+{
+}
+
+void IOBuffer::Put(const char *buf,int size)
+{
+   if(size>=PUT_LL_MIN && Size()==0 && mode==PUT && !save)
+   {
+      int res=Put_LL(buf,size);
+      if(res>=0)
+      {
+	 buf+=res;
+	 size-=res;
+	 pos+=res;
+      }
+   }
+   DirectedBuffer::Put(buf,size);
+}
+
 int IOBuffer::Do()
 {
    if(Done() || Error())
@@ -376,13 +402,12 @@ int IOBuffer::Do()
    switch(mode)
    {
    case PUT:
-      if(in_buffer==0)
+      if(Size()==0)
 	 return STALL;
-      res=Put_LL(buffer+buffer_ptr,in_buffer);
+      res=Put_LL(buffer+buffer_ptr,Size());
       if(res>0)
       {
 	 RateAdd(res);
-	 in_buffer-=res;
 	 buffer_ptr+=res;
 	 event_time=now;
 	 return MOVED;
@@ -436,12 +461,11 @@ int IOBufferStacked::Do()
 	 SetError(down->ErrorText(),down->ErrorFatal());
 	 m=MOVED;
       }
-      if(in_buffer==0)
+      if(Size()==0)
 	 return m;
-      res=Put_LL(buffer+buffer_ptr,in_buffer);
+      res=Put_LL(buffer+buffer_ptr,Size());
       if(res>0)
       {
-	 in_buffer-=res;
 	 buffer_ptr+=res;
 	 m=MOVED;
 	 down->Do();
@@ -494,8 +518,7 @@ int IOBufferStacked::Get_LL(int)
       return 0;
    }
 
-   Allocate(size);
-   memcpy(buffer+buffer_ptr+in_buffer,b,size);
+   memcpy(GetSpace(size),b,size);
    down->Skip(size);
    return size;
 }
@@ -514,7 +537,7 @@ bool IOBufferStacked::Done()
 #define super IOBuffer
 int IOBufferFDStream::Put_LL(const char *buf,int size)
 {
-   if(put_ll_timer && !eof && in_buffer<PUT_LL_MIN
+   if(put_ll_timer && !eof && Size()<PUT_LL_MIN
    && !put_ll_timer->Stopped())
       return 0;
    if(stream->broken())
@@ -576,9 +599,7 @@ int IOBufferFDStream::Get_LL(int size)
       return 0;
    }
 
-   Allocate(size);
-
-   res=read(fd,buffer+buffer_ptr+in_buffer,size);
+   res=read(fd,GetSpace(size),size);
    if(res==-1)
    {
       saved_errno=errno;
@@ -621,18 +642,11 @@ bool IOBufferFDStream::Done()
 // IOBufferFileAccess implementation
 #undef super
 #define super IOBuffer
-IOBufferFileAccess::~IOBufferFileAccess()
-{
-   session->Resume();
-   session->Close();
-}
 int IOBufferFileAccess::Get_LL(int size)
 {
    int res=0;
 
-   Allocate(size);
-
-   res=session->Read(buffer+buffer_ptr+in_buffer,size);
+   res=session->Read(GetSpace(size),size);
    if(res<0)
    {
       if(res==FA::DO_AGAIN)
@@ -658,47 +672,47 @@ const char *IOBufferFileAccess::Status()
    return session->CurrentStatus();
 }
 
-unsigned long long Buffer::UnpackUINT64BE(int offset)
+unsigned long long Buffer::UnpackUINT64BE(int offset) const
 {
-   if(in_buffer-offset<8)
+   if(Size()-offset<8)
       return 0;
    unsigned long long res=UnpackUINT32BE(offset);
    res=(res<<32)|UnpackUINT32BE(offset+4);
    return res;
 }
-long long Buffer::UnpackINT64BE(int offset)
+long long Buffer::UnpackINT64BE(int offset) const
 {
    unsigned long long n=UnpackUINT64BE(offset);
    if(n&0x8000000000000000ULL)
       return -(long long)(n^0xFFFFFFFFFFFFFFFFULL)-1;
    return (long long)n;
 }
-unsigned Buffer::UnpackUINT32BE(int offset)
+unsigned Buffer::UnpackUINT32BE(int offset) const
 {
-   if(in_buffer-offset<4)
+   if(Size()-offset<4)
       return 0;
-   unsigned char *b=(unsigned char*)buffer+buffer_ptr+offset;
+   unsigned char *b=(unsigned char*)buffer.get()+buffer_ptr+offset;
    return (b[0]<<24)|(b[1]<<16)|(b[2]<<8)|b[3];
 }
-int Buffer::UnpackINT32BE(int offset)
+int Buffer::UnpackINT32BE(int offset) const
 {
    unsigned n=UnpackUINT32BE(offset);
    if(n&0x80000000U)
       return -(int)(n^0xFFFFFFFFU)-1;
    return (int)n;
 }
-unsigned Buffer::UnpackUINT16BE(int offset)
+unsigned Buffer::UnpackUINT16BE(int offset) const
 {
-   if(in_buffer-offset<2)
+   if(Size()-offset<2)
       return 0;
-   unsigned char *b=(unsigned char*)buffer+buffer_ptr+offset;
+   unsigned char *b=(unsigned char*)buffer.get()+buffer_ptr+offset;
    return (b[0]<<8)|b[1];
 }
-unsigned Buffer::UnpackUINT8(int offset)
+unsigned Buffer::UnpackUINT8(int offset) const
 {
-   if(in_buffer-offset<1)
+   if(Size()-offset<1)
       return 0;
-   unsigned char *b=(unsigned char*)buffer+buffer_ptr+offset;
+   unsigned char *b=(unsigned char*)buffer.get()+buffer_ptr+offset;
    return b[0];
 }
 void Buffer::PackUINT64BE(unsigned long long data)
@@ -724,13 +738,12 @@ void Buffer::PackUINT32BE(unsigned data)
 #ifndef NDEBUG
    Log::global->Format(11,"PackUINT32BE(0x%08X)\n",data);
 #endif
-   Allocate(4);
-   char *b=buffer+buffer_ptr+in_buffer;
+   char *b=GetSpace(4);
    b[0]=(data>>24)&255;
    b[1]=(data>>16)&255;
    b[2]=(data>>8)&255;
    b[3]=(data)&255;
-   in_buffer+=4;
+   SpaceAdd(4);
 }
 void Buffer::PackINT32BE(int data)
 {
@@ -743,19 +756,17 @@ void Buffer::PackINT32BE(int data)
 }
 void Buffer::PackUINT16BE(unsigned data)
 {
-   Allocate(2);
-   char *b=buffer+buffer_ptr+in_buffer;
+   char *b=GetSpace(2);
    b[0]=(data>>8)&255;
    b[1]=(data)&255;
-   in_buffer+=2;
+   SpaceAdd(2);
 }
 void Buffer::PackUINT8(unsigned data)
 {
 #ifndef NDEBUG
    Log::global->Format(11,"PackUINT8(0x%02X)\n",data);
 #endif
-   Allocate(1);
-   char *b=buffer+buffer_ptr+in_buffer;
+   char *b=GetSpace(1);
    b[0]=(data)&255;
-   in_buffer+=1;
+   SpaceAdd(1);
 }
