@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1998-2005 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1998-2007 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: buffer.h,v 1.40.2.2 2007/06/15 07:16:39 lav Exp $ */
+/* $Id: buffer.h,v 1.54 2007/09/10 11:45:28 lav Exp $ */
 
 #ifndef BUFFER_H
 #define BUFFER_H
@@ -28,30 +28,27 @@
 #include "Timer.h"
 #include "fg.h"
 #include "xstring.h"
+#include "Speedometer.h"
 
 #include <stdarg.h>
 
 #ifdef HAVE_ICONV
 CDECL_BEGIN
-#include <iconv.h>
+# include <iconv.h>
 CDECL_END
 #endif
 
 #define GET_BUFSIZE 0x10000
 #define PUT_LL_MIN  0x2000
 
-class Speedometer;
-
 class Buffer
 {
 protected:
-   char *error_text;
+   xstring error_text;
    int  saved_errno;
    bool error_fatal;
 
-   char *buffer;
-   int buffer_allocated;
-   int in_buffer;
+   xstring buffer;
    int buffer_ptr;
    bool eof;	  // no reads possible (except from mem buffer)
    bool broken;	  // no writes possible
@@ -61,11 +58,7 @@ protected:
 
    off_t pos;
 
-   // low-level for derived classes
-   virtual int Put_LL(const char *buf,int size) { return 0; }
-   virtual int PutEOF_LL() { return 0; }
-
-   Speedometer *rate;
+   SMTaskRef<Speedometer> rate;
    void RateAdd(int n);
 
    void Allocate(int size);
@@ -76,35 +69,46 @@ public:
    bool Error() const { return error_text!=0; }
    bool ErrorFatal() const { return error_fatal; }
    void SetError(const char *e,bool fatal=false);
-   void SetError2(const char *e1,const char *e2,bool fatal=false);
    void SetErrorCached(const char *e);
    const char *ErrorText() const { return error_text; }
-   int Size() const { return in_buffer; }
+   int Size() const { return buffer.length()-buffer_ptr; }
    bool Eof() const { return eof; }
    bool Broken() const { return broken; }
 
-   void Get(const char **buf,int *size);
+   const char *Get() const;
+   void Get(const char **buf,int *size) const;
    void Skip(int len); // Get(); consume; Skip()
    void UnSkip(int len); // this only works if there were no Put's.
+   void Append(const char *buf,int size);
    void Put(const char *buf,int size);
    void Put(const char *buf) { Put(buf,strlen(buf)); }
+   void Put(const xstring &s) { Put(s.get(),s.length()); }
    void Format(const char *f,...) PRINTF_LIKE(2,3);
    void vFormat(const char *f, va_list v);
-   void PutEOF() { eof=true; PutEOF_LL(); }
+   void PutEOF() { eof=true; }
+   char *GetSpace(int size) {
+      Allocate(size);
+      return buffer.get_non_const()+buffer.length();
+   }
+   void SpaceAdd(int size) {
+      buffer.set_length(buffer.length()+size);
+   }
+   void Prepend(const char *buf,int size);
+   void Prepend(const char *buf) { Prepend(buf,strlen(buf)); }
 
-   unsigned long long UnpackUINT64BE(int offset=0);
-   unsigned UnpackUINT32BE(int offset=0);
-   unsigned UnpackUINT16BE(int offset=0);
-   unsigned UnpackUINT8(int offset=0);
+   unsigned long long UnpackUINT64BE(int offset=0) const;
+   unsigned UnpackUINT32BE(int offset=0) const;
+   unsigned UnpackUINT16BE(int offset=0) const;
+   unsigned UnpackUINT8(int offset=0) const;
    void PackUINT64BE(unsigned long long data);
    void PackUINT32BE(unsigned data);
    void PackUINT16BE(unsigned data);
    void PackUINT8(unsigned data);
 
-   long long UnpackINT64BE(int offset=0);
-   int UnpackINT32BE(int offset=0);
-   int UnpackINT16BE(int offset=0);
-   int UnpackINT8(int offset=0);
+   long long UnpackINT64BE(int offset=0) const;
+   int UnpackINT32BE(int offset=0) const;
+   int UnpackINT16BE(int offset=0) const;
+   int UnpackINT8(int offset=0) const;
    void PackINT64BE(long long data);
    void PackINT32BE(int data);
    void PackINT16BE(int data);
@@ -114,7 +118,7 @@ public:
    void Save(int m) { save=true; save_max=m; }
    bool IsSaving() const { return save; }
    void GetSaved(const char **buf,int *size) const;
-   void SaveRollback(int p);
+   void SaveRollback(off_t p);
 
    void SetPos(off_t p) { pos=p; }
    off_t GetPos() const { return pos; }
@@ -125,54 +129,74 @@ public:
    void Empty();
 
    Buffer();
-   virtual ~Buffer();
+   ~Buffer();
 };
 
+class DataTranslator : public Buffer
+{
+public:
+   virtual void PutTranslated(Buffer *dst,const char *buf,int size)=0;
+   virtual void ResetTranslation() { Empty(); }
+   virtual ~DataTranslator() {}
+
+   // same as PutTranslated, but does not advance pos.
+   void AppendTranslated(Buffer *dst,const char *buf,int size);
+};
+
+#ifdef HAVE_ICONV
+class DataRecoder : public DataTranslator
+{
+   iconv_t backend_translate;
+public:
+   void PutTranslated(Buffer *dst,const char *buf,int size);
+   void ResetTranslation();
+   DataRecoder(const char *from_code,const char *to_code,bool translit=true);
+   ~DataRecoder();
+};
+#endif //HAVE_ICONV
+
 class DirectedBuffer : public Buffer
 {
 public:
    enum dir_t { GET, PUT };
 
 protected:
-#ifdef HAVE_ICONV
-   iconv_t backend_translate;
-   Buffer *untranslated;
-#endif
+   Ref<DataTranslator> translator;
    dir_t mode;
    void EmbraceNewData(int len);
 
 public:
+   DirectedBuffer(dir_t m) : mode(m) {}
+   void SetTranslator(DataTranslator *t) { translator=t; }
+   void SetTranslation(const char *be_encoding,bool translit=true)
 #ifdef HAVE_ICONV
-   DirectedBuffer(dir_t m) { mode=m; backend_translate=0; untranslated=0; }
-   ~DirectedBuffer();
-   void SetTranslation(const char *be_encoding,bool translit=true);
-   virtual void PutTranslated(const char *buf,int size);
+      ;
+#else
+      {}
+#endif //HAVE_ICONV
+   void PutTranslated(const char *buf,int size);
    void PutTranslated(const char *buf) { PutTranslated(buf,strlen(buf)); }
    void ResetTranslation();
-   void Put(const char *buf,int size)
-      {
-	 if(mode==PUT)
-	    PutTranslated(buf,size);
-	 else
-	    Buffer::Put(buf,size);
-      }
+   void PutRaw(const char *buf,int size) { Buffer::Put(buf,size); }
+   void PutRaw(const char *buf) { Buffer::Put(buf); }
+   void Put(const char *buf,int size);
    void Put(const char *buf) { Put(buf,strlen(buf)); }
-#else
-   DirectedBuffer(dir_t m) { mode=m; }
-   void SetTranslation(const char *be_encoding,bool translit=true) {}
-   virtual void PutTranslated(const char *buf,int size) { Put(buf,size); }
-   void ResetTranslation() {}
-#endif
    dir_t GetDirection() { return mode; }
 };
 
 class IOBuffer : public DirectedBuffer, public SMTask
 {
 protected:
+   // low-level for derived classes
+   virtual int Get_LL(int size) { return 0; }
+   virtual int Put_LL(const char *buf,int size) { return 0; }
+   virtual int PutEOF_LL() { return 0; }
+
    Time event_time; // used to detect timeouts
 
 public:
-   IOBuffer(dir_t m) : DirectedBuffer(m), event_time(now) {}
+   IOBuffer(dir_t m);
+   virtual ~IOBuffer();
    virtual const Time& EventTime()
       {
 	 if(IsSuspended())
@@ -181,26 +205,31 @@ public:
       }
    virtual bool Done()
       {
-	 return(broken || Error() || (eof && (mode==GET || in_buffer==0)));
+	 return(broken || Error() || (eof && (mode==GET || Size()==0)));
       }
-   virtual int Get_LL(int size) { return 0; }
    virtual int Do();
 
    virtual FgData *GetFgData(bool) { return 0; }
    virtual const char *Status() { return ""; }
    virtual int Buffered() { return Size(); }
+
+   // Put method with Put_LL shortcut
+   void Put(const char *,int);
+   void Put(const char *buf) { Put(buf,strlen(buf)); }
+   void Put(const xstring &s) { Put(s.get(),s.length()); }
+   // anchor to PutEOF_LL
+   void PutEOF() { DirectedBuffer::PutEOF(); PutEOF_LL(); }
 };
 
 class IOBufferStacked : public IOBuffer
 {
-   IOBuffer *down;
+   SMTaskRef<IOBuffer> down;
 
    int Get_LL(int size);
    int Put_LL(const char *buf,int size);
 
 public:
-   IOBufferStacked(IOBuffer *b) : IOBuffer(b->GetDirection()) { down=b; }
-   ~IOBufferStacked() { Delete(down); }
+   IOBufferStacked(IOBuffer *b) : IOBuffer(b->GetDirection()), down(b) {}
    const Time& EventTime() { return down->EventTime(); }
    int Do();
    bool Done();
@@ -208,36 +237,32 @@ public:
 
 class IOBufferFDStream : public IOBuffer
 {
-   FDStream *stream;
-   class Timer *put_ll_timer;
+   Ref<FDStream> stream;
+   Ref<Timer> put_ll_timer;
 
    int Get_LL(int size);
    int Put_LL(const char *buf,int size);
 
-protected:
-   ~IOBufferFDStream() { delete stream; delete put_ll_timer; }
-
 public:
-   IOBufferFDStream(FDStream *o,dir_t m,Timer *t=0)
-      : IOBuffer(m) { stream=o; put_ll_timer=t; }
+   IOBufferFDStream(FDStream *o,dir_t m)
+      : IOBuffer(m), stream(o) {}
+   IOBufferFDStream(FDStream *o,dir_t m,Timer *t)
+      : IOBuffer(m), stream(o), put_ll_timer(t) {}
    bool Done();
    FgData *GetFgData(bool fg);
    const char *Status() { return stream->status; }
 };
 
-class FileAccess;
+#include <FileAccess.h>
 
 class IOBufferFileAccess : public IOBuffer
 {
-   FileAccess *session;
+   const FileAccessRef& session;
 
    int Get_LL(int size);
 
-protected:
-   ~IOBufferFileAccess();
-
 public:
-   IOBufferFileAccess(FileAccess *i) : IOBuffer(GET) { session=i; }
+   IOBufferFileAccess(const FileAccessRef& i) : IOBuffer(GET), session(i) {}
 
    void SuspendInternal();
    void ResumeInternal();
