@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2006 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2008 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: commands.cc,v 1.261.2.2 2007/02/26 15:09:54 lav Exp $ */
+/* $Id: commands.cc,v 1.286 2008/02/06 07:11:23 lav Exp $ */
 
 #include <config.h>
 
@@ -87,7 +87,7 @@ History	 cwd_history;
 CMD(alias); CMD(anon);   CMD(cd);      CMD(debug);	CMD(du);
 CMD(exit);  CMD(get);    CMD(help);    CMD(jobs);
 CMD(kill);  CMD(lcd);    CMD(ls);      CMD(cls);
-CMD(open);  CMD(pwd);    CMD(set);
+CMD(open);  CMD(pwd);    CMD(set);     CMD(eval);
 CMD(shell); CMD(source); CMD(user);    CMD(rm);
 CMD(wait);  CMD(subsh);  CMD(mirror);
 CMD(mv);    CMD(cat);    CMD(cache);
@@ -235,6 +235,7 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 " -s, --summarize       display only a total for each argument\n"
 	 "     --exclude=PAT     exclude files that match PAT\n")},
    {"echo",    cmd_echo,   0},
+   {"eval",    cmd_eval,   0},
    {"exit",    cmd_exit,   N_("exit [<code>|bg]"),
 	 N_("exit - exit from lftp or move to background if jobs are active\n\n"
 	 "If no jobs active, the code is passed to operating system as lftp\n"
@@ -359,6 +360,7 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 " -e <cmd>            execute the command just after selecting\n"
 	 " -u <user>[,<pass>]  use the user/password for authentication\n"
 	 " -p <port>           use the port for connection\n"
+	 " -s <slot>           assign the connection to this slot\n"
 	 " <site>              host name, URL or bookmark name\n")},
    {"pget",    cmd_get,    N_("pget [OPTS] <rfile> [-o <lfile>]"),
 	 N_("Gets the specified file using several connections. This can speed up transfer,\n"
@@ -452,7 +454,10 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 "Sleep for given amount of time. The time argument can be optionally\n"
 	 "followed by unit specifier: d - days, h - hours, m - minutes, s - seconds.\n"
 	 "By default time is assumed to be seconds.\n")},
-   {"slot",    cmd_slot,   0,0},
+   {"slot",    cmd_slot, 0,
+        N_("Usage: slot [<label>]\n"
+   	"List assigned slots.\n"
+	"If <label> is specified, switch to the slot named <label>.\n")},
    {"source",  cmd_source, N_("source <file>"),
 	 N_("Execute commands recorded in file <file>\n")},
    {"suspend", cmd_suspend},
@@ -553,8 +558,7 @@ Job *CmdExec::builtin_lcd()
       return 0;
    }
 
-   xfree(old_lcwd);
-   old_lcwd=xstrdup(cwd->GetName());
+   old_lcwd.set(cwd->GetName());
 
    SaveCWD();
 
@@ -613,13 +617,19 @@ Job *CmdExec::builtin_cd()
 	 is_file=(last_char(dir)!='/');
    }
 
+   int cache_is_dir=FileAccess::cache->IsDirectory(session,dir);
+   if(cache_is_dir==1)
+      is_file=false;
+   else if(cache_is_dir==0)
+      is_file=true;
+
    old_cwd=session->GetCwd();
-   FileAccess::Path new_cwd=old_cwd;
+   FileAccess::Path new_cwd(old_cwd);
    new_cwd.Change(dir,is_file);
    if(url)
       new_cwd.SetURL(url);
    if(!verify_path || background
-   || (!verify_path_cached && FileAccess::cache->IsDirectory(session,dir)==1))
+   || (!verify_path_cached && cache_is_dir==1))
    {
       cwd_history.Set(session,old_cwd);
       session->SetCwd(new_cwd);
@@ -629,7 +639,7 @@ Job *CmdExec::builtin_cd()
       return 0;
    }
    session->PathVerify(new_cwd);
-   Roll(session);
+   session->Roll();
    builtin=BUILTIN_CD;
    return this;
 }
@@ -649,7 +659,7 @@ Job *CmdExec::builtin_exit()
       if(!strcmp(a,"top") || !strcmp(a,"bg"))
       {
 	 if(top)
-	    exec=top;
+	    exec=top.get_non_const();
       }
       else if(!strcmp(a,"kill"))
       {
@@ -697,8 +707,7 @@ CmdFeeder *lftp_feeder=0;
 Job *CmdExec::builtin_lftp()
 {
    int c;
-   const char *cmd=0;
-   char *acmd;
+   xstring cmd;
    bool debug=false;
    static struct option lftp_options[]=
    {
@@ -708,34 +717,23 @@ Job *CmdExec::builtin_lftp()
    };
 
    opterr=false;
-   while((c=args->getopt_long("+f:c:vhd",lftp_options,0))!=EOF)
+   while((c=args->getopt_long("+f:c:vhd",lftp_options))!=EOF)
    {
       switch(c)
       {
       case('h'):
-	 cmd="help lftp;";
+	 cmd.set("help lftp;");
 	 break;
       case('v'):
-	 cmd="version;";
+	 cmd.set("version;");
 	 break;
       case('f'):
-	 acmd=string_alloca(20+2*strlen(optarg));
-	 strcpy(acmd,"source \"");
-	 unquote(acmd+strlen(acmd),optarg);
-	 strcat(acmd,"\";");
-	 cmd=acmd;
+	 cmd.vset("source \"",unquote(optarg),"\";",NULL);
 	 break;
       case('c'):
-      {
-	 int cmd_start=args->getindex()-1;
-	 char *c = (args->count()==cmd_start+1
-		      ? args->Combine(cmd_start) : args->CombineQuoted(cmd_start));
-	 acmd=string_alloca(4+strlen(c));
-	 sprintf(acmd,"%s\n\n",c);
-	 cmd=acmd;
-	 xfree(c);
+	 cmd.set_allocated(args->CombineCmd(args->getindex()-1));
+	 cmd.append("\n\n");
 	 break;
-      }
       case('d'):
 	 debug=true;
 	 break;
@@ -773,11 +771,11 @@ Job *CmdExec::builtin_open()
    ReuseSavedSession();
 
    bool	 debug=false;
-   char	 *port=NULL;
+   const char *port=NULL;
    const char *host=NULL;
-   char  *path=NULL;
-   char	 *user=NULL;
-   char	 *pass=NULL;
+   const char *path=NULL;
+   const char *user=NULL;
+   const char *pass=NULL;
    int	 c;
    NetRC::Entry *nrc=0;
    char  *cmd_to_exec=0;
@@ -797,30 +795,36 @@ Job *CmdExec::builtin_open()
       {"execute",required_argument,0,'e'},
       {"debug",optional_argument,0,'d'},
       {"no-bookmark",no_argument,0,'B'},
+      {"slot",required_argument,0,'s'},
       {"help",no_argument,0,'h'},
       {0,0,0,0}
    };
 
-   while((c=args->getopt_long("u:p:e:dBh",open_options,0))!=EOF)
+   while((c=args->getopt_long("u:p:e:s:dBh",open_options))!=EOF)
    {
       switch(c)
       {
+      case('s'):
+      	if (*optarg) ChangeSlot(optarg);
+        break;
       case('p'):
 	 port=optarg;
 	 break;
       case('u'):
+      {
          user=optarg;
-         pass=strchr(optarg,',');
-	 if(pass==NULL)
-	    pass=strchr(optarg,' ');
-	 if(pass==NULL)
-	    pass=strchr(optarg,':');
-	 if(pass==NULL)
+         char *sep=strchr(optarg,',');
+	 if(sep==NULL)
+	    sep=strchr(optarg,' ');
+	 if(sep==NULL)
+	    sep=strchr(optarg,':');
+	 if(sep==NULL)
    	    break;
-	 *pass=0;
-	 pass++;
+	 *sep=0;
+	 pass=sep+1;
 	 insecure=true;
          break;
+      }
       case(OPT_USER):
 	 user=optarg;
 	 break;
@@ -855,7 +859,7 @@ Job *CmdExec::builtin_open()
    if(optind<args->count())
       host=args->getarg(optind++);
 
-   ParsedURL *url=0;
+   Ref<ParsedURL> url;
 
    const char *bm=0;
 
@@ -906,24 +910,19 @@ Job *CmdExec::builtin_open()
 	       p="ftp";
 	    char *u=string_alloca(strlen(p)+3+strlen(host)+1);
 	    sprintf(u,"%s://%s",p,host);
-	    delete url;
 	    url=new ParsedURL(u);
 	 }
 	 if(user || port)
 	 {
-	    url->user=user;
-	    url->pass=pass;
 	    if(user)
 	    {
-	       url->user=user;
-	       url->pass=pass;
+	       url->user.set(user);
+	       url->pass.set(pass);
 	    }
 	    if(port)
-	       url->port=port;
-	    char *host1=url->Combine();
-	    delete url;
+	       url->port.set(port);
+	    xstring_ca host1(url->Combine());
 	    url=new ParsedURL(host1);
-	    xfree(host1);
 	 }
 
 	 const ParsedURL &uc=*url;
@@ -939,7 +938,7 @@ Job *CmdExec::builtin_open()
 	       insecure=true;
 	    }
 	    host=uc.host;
-	    if(uc.port!=0 && port==0)
+	    if(uc.port && !port)
 	       port=uc.port;
 	    if(uc.path && !path)
 	       path=uc.path;
@@ -947,13 +946,12 @@ Job *CmdExec::builtin_open()
 	    FileAccess *new_session=FileAccess::New(uc.proto,host,port);
 	    if(!new_session)
 	    {
-	       eprintf("%s: %s%s\n",args->a0(),uc.proto,
+	       eprintf("%s: %s%s\n",args->a0(),uc.proto.get(),
 			_(" - not supported protocol"));
 	       return 0;
 	    }
 
-	    saved_session=session;
-	    session=0;
+	    saved_session=session.borrow();
 	    ChangeSession(new_session);
 	 }
 
@@ -964,9 +962,9 @@ Job *CmdExec::builtin_open()
 	    if(nrc)
 	    {
 	       if(!user)
-		  Log::global->Format(3,"---- using user `%s' and password from ~/.netrc\n",nrc->user);
+		  FA::LogNote(3,"using user `%s' and password from ~/.netrc",nrc->user.get());
 	       else
-		  Log::global->Format(3,"---- using password from ~/.netrc\n");
+		  FA::LogNote(3,"using password from ~/.netrc");
 	       user=nrc->user;
 	       pass=nrc->pass;
 	    }
@@ -1028,7 +1026,7 @@ Job *CmdExec::builtin_open()
 	 session->SetCwd(FileAccess::Path(old,is_file,url));
       }
 
-      const char *cd_arg=(url && url->orig_url)?url->orig_url:path;
+      const char *cd_arg=(url && url->orig_url)?url->orig_url.get():path;
       char *s=string_alloca(strlen(cd_arg)*2+40);
       strcpy(s,"&& cd \"");
       unquote(s+strlen(s),cd_arg);
@@ -1042,12 +1040,6 @@ Job *CmdExec::builtin_open()
    if(debug)
       PrependCmd("debug\n");
 
-   if(url)
-   {
-      delete url;
-      url=0;
-   }
-
    if(slot)
       ConnectionSlot::Set(slot,session);
 
@@ -1106,7 +1098,6 @@ Job *CmdExec::builtin_glob()
    const char *pat=args->getnext();
    if(!pat)
    {
-      delete args_glob;
       args_glob=0;
       args->rewind();
       return cmd_command(this);
@@ -1135,7 +1126,7 @@ Job *CmdExec::builtin_queue()
    int verbose = -1; /* default */
 
    int opt;
-   while((opt=args->getopt_long("+dm:n:qvQw",queue_options,0))!=EOF)
+   while((opt=args->getopt_long("+dm:n:qvQw",queue_options))!=EOF)
    {
       switch(opt)
       {
@@ -1210,11 +1201,7 @@ Job *CmdExec::builtin_queue()
 	 if(!queue)
 	    queue=GetQueue(true);
 
-	 char *cmd=NULL;
-	 if(args_remaining == 1)
-	    cmd=args->Combine(args->getindex());
-	 else
-	    cmd=args->CombineQuoted(args->getindex());
+	 xstring_ca cmd(args->CombineCmd(args->getindex()));
 
 	 if(!strcasecmp(cmd,"stop"))
 	    queue->Suspend();
@@ -1223,7 +1210,6 @@ Job *CmdExec::builtin_queue()
 	 else
 	    queue->queue_feeder->QueueCmd(cmd, session->GetCwd(),
 					  cwd?cwd->GetName():0, pos, verbose);
-	 xfree(cmd);
 
 	 last_bg=queue->jobno;
 	 exit_code=0;
@@ -1303,23 +1289,20 @@ Job *CmdExec::builtin_queue_edit()
    const char *home=getenv("HOME");
    if(home==0)
       home="";
-   char *q_file = (char *) xmalloc(strlen(home)+32);
    /* if we have $HOME, use $HOME/.lftp/file; otherwise just file */
-   sprintf(q_file, "%s%slftp-queue-%i", home, home? "/.lftp/":"", (int) getpid());
+   xstring_ca q_file(xasprintf("%s%slftp-queue-%i", home, home? "/.lftp/":"", (int) getpid()));
 
    int q_fd=open(q_file,O_RDWR|O_CREAT,0600);
    if(q_fd==-1) {
-      eprintf(_("Couldn't create temporary file `%s': %s.\n"), q_file, strerror(errno));
-      xfree(q_file);
+      eprintf(_("Couldn't create temporary file `%s': %s.\n"), q_file.get(), strerror(errno));
       return 0;
    }
 
    if(!queue->WriteCmds(q_fd)) {
-      eprintf(_("%s: error writing %s: %s\n"), args->a0(), q_file, strerror(errno));
+      eprintf(_("%s: error writing %s: %s\n"), args->a0(), q_file.get(), strerror(errno));
 
       /* abort without clearing the queue */
       close(q_fd);
-      xfree(q_file);
 
       return 0;
    }
@@ -1332,14 +1315,10 @@ Job *CmdExec::builtin_queue_edit()
     * this also effectively "suspends" this queue until it's done editing. */
 
    {
-      char *editcmd = (char *) xmalloc(strlen(q_file)*3+128);
-      sprintf(editcmd, "shell \"/bin/sh -c 'exec ${EDITOR:-vi} %s'\"; queue -s %s; shell rm '%s'\n", q_file, q_file, q_file);
+      xstring_ca editcmd(xasprintf("shell \"/bin/sh -c 'exec ${EDITOR:-vi} %s'\"; queue -s %s; shell rm '%s'\n", q_file.get(), q_file.get(), q_file.get()));
       PrependCmd(editcmd);
-      xfree(editcmd);
    }
 
-   xfree(q_file);
-
    return 0;
 }
 #endif
@@ -1381,25 +1360,29 @@ CMD(ls)
 	 args->insarg(1,"SITE");
    }
 
-   char *a=args->Combine(nlist?1:0);
+   xstring_ca a(args->Combine(nlist?1:0));
 
    const char *var_ls=ResMgr::Query("cmd:ls-default",session->GetConnectURL(FA::NO_PATH));
    if(!nlist && args->count()==1 && var_ls[0])
       args->Append(var_ls);
 
+   bool use_color=false;
+   if(!nlist)
+   {
+      ResValue color=ResMgr::Query("color:use-color",0);
+      if(!strcasecmp(color,"auto"))
+	 use_color=!output && isatty(1);
+      else
+	 use_color=color.to_bool();
+   }
+   bool no_status=(!output || output->usesfd(1));
+
    FileCopyPeer *src_peer=0;
    if(!nlist)
    {
-      FileCopyPeerDirList *dir_list=new FileCopyPeerDirList(session->Clone(),args);
-
+      FileCopyPeerDirList *dir_list=new FileCopyPeerDirList(session->Clone(),args.borrow());
+      dir_list->UseColor(use_color);
       src_peer=dir_list;
-      ResValue color=ResMgr::Query("color:use-color",0);
-      bool use;
-      if(!strcasecmp(color,"auto"))
-	 use=!output && isatty(1);
-      else
-	 use=color.to_bool();
-      dir_list->UseColor(use);
    }
    else
       src_peer=new FileCopyPeerFA(session->Clone(),a,mode);
@@ -1408,7 +1391,7 @@ CMD(ls)
       src_peer->NoCache();
    src_peer->SetDate(NO_DATE);
    src_peer->SetSize(NO_SIZE);
-   FileCopyPeer *dst_peer=new FileCopyPeerFDStream(output,FileCopyPeer::PUT);
+   FileCopyPeer *dst_peer=new FileCopyPeerFDStream(output.borrow(),FileCopyPeer::PUT);
 
    FileCopy *c=FileCopy::New(src_peer,dst_peer,false);
    c->DontCopyDate();
@@ -1417,18 +1400,14 @@ CMD(ls)
       c->Ascii();
 
    CopyJob *j=new CopyJob(c,a,op);
-   if(!output || output->usesfd(1))
+   if(no_status)
       j->NoStatusOnWrite();
-   xfree(a);
-   output=0;
-   if(!nlist)
-      args=0;  // `ls' consumes args itself.
 
    return j;
 }
 
 /* this seems to belong here more than in FileSetOutput.cc ... */
-const char *FileSetOutput::parse_argv(ArgV *a)
+const char *FileSetOutput::parse_argv(const Ref<ArgV>& a)
 {
    enum {
       OPT_BLOCK_SIZE,
@@ -1472,8 +1451,8 @@ const char *FileSetOutput::parse_argv(ArgV *a)
 
    const char *time_style=ResMgr::Query("cmd:time-style",0);
 
-   int opt, longopt;
-   while((opt=a->getopt_long(":1BdFhiklqsDISrt", cls_options, &longopt))!=EOF)
+   int opt;
+   while((opt=a->getopt_long(":1BdFhiklqsDISrt", cls_options))!=EOF)
    {
       switch(opt) {
       case OPT_SORT:
@@ -1564,30 +1543,23 @@ const char *FileSetOutput::parse_argv(ArgV *a)
 	 break;
 
       default:
-	 /* silly getopt won't give us its error instead of printing it, oh well.
-	  * we only need to never print errors for completion and validation; for
-	  * cls itself we could do it, but that'd be a special case ... */
-	 return _("invalid option");
+	 return a->getopt_error_message(opt);
       }
    }
 
-   while(a->getindex()>1)
-      a->delarg(1);
-   a->rewind();
-
-   xfree(time_fmt); time_fmt=0;
+   time_fmt.set(0);
    if(time_style && time_style[0]) {
       if(time_style[0]=='+')
-	 time_fmt=xstrdup(time_style+1);
+	 time_fmt.set(time_style+1);
       else if(!strcmp(time_style,"full-iso"))
-//	 time_fmt=xstrdup("%Y-%m-%d %H:%M:%S.%N %z"); // %N and %z are GNU extensions
-	 time_fmt=xstrdup("%Y-%m-%d %H:%M:%S");
+//	 time_fmt.set("%Y-%m-%d %H:%M:%S.%N %z"); // %N and %z are GNU extensions
+	 time_fmt.set("%Y-%m-%d %H:%M:%S");
       else if(!strcmp(time_style,"long-iso"))
-	 time_fmt=xstrdup("%Y-%m-%d %H:%M");
+	 time_fmt.set("%Y-%m-%d %H:%M");
       else if(!strcmp(time_style,"iso"))
-	 time_fmt=xstrdup("%Y-%m-%d \n%m-%d %H:%M");
+	 time_fmt.set("%Y-%m-%d \n%m-%d %H:%M");
       else
-	 time_fmt=xstrdup(time_style);
+	 time_fmt.set(time_style);
       need_exact_time=false;
       if(time_fmt) {
 	 static const char exact_fmts[][3]={"%H","%M","%S","%N",""};
@@ -1604,6 +1576,11 @@ const char *FileSetOutput::parse_argv(ArgV *a)
       }
    }
 
+   // remove parsed options.
+   while(a->getindex()>1)
+      a->delarg(1);
+   a->rewind();
+
    return NULL;
 }
 
@@ -1614,31 +1591,25 @@ CMD(cls)
    const char *op=args->a0();
    bool re=false;
 
-   OutputJob *out=new OutputJob(output, args->a0());
-   output=0;
-
-   FileSetOutput fso;
-   fso.config(out);
+   JobRef<OutputJob> out(new OutputJob(output.borrow(), args->a0()));
+   Ref<FileSetOutput> fso(new FileSetOutput);
+   fso->config(out);
 
    if(!strncmp(op,"re",2))
       re=true;
 
-   ArgV arg("", ResMgr::Query("cmd:cls-default", 0));
-   fso.parse_argv(&arg);
+   fso->parse_res(ResMgr::Query("cmd:cls-default", 0));
 
-   if(const char *err = fso.parse_argv(args)) {
-      eprintf("%s: %s.\n", op, err);
+   if(const char *err = fso->parse_argv(args)) {
+      eprintf("%s: %s\n", op, err);
       eprintf(_("Try `help %s' for more information.\n"),op);
-      delete out;
       return 0;
    }
 
-   clsJob *j = new clsJob(session->Clone(), args, fso, out);
+   clsJob *j = new clsJob(session->Clone(), args.borrow(), fso.borrow(), out.borrow());
    if(re)
       j->UseCache(false);
 
-   args=0;
-
    return j;
 }
 
@@ -1674,8 +1645,8 @@ CMD(cat)
       eprintf(_("Usage: %s [OPTS] files...\n"),op);
       return 0;
    }
-   OutputJob *out=new OutputJob(output, args->a0());
-   CatJob *j=new CatJob(session->Clone(),out,args);
+   OutputJob *out=new OutputJob(output.borrow(), args->a0());
+   CatJob *j=new CatJob(session->Clone(),out,args.borrow());
    if(!auto_ascii)
    {
       if(ascii)
@@ -1683,8 +1654,6 @@ CMD(cat)
       else
 	 j->Binary();
    }
-   output=0;
-   args=0;
    return j;
 }
 
@@ -1695,14 +1664,14 @@ CMD(get)
    const char *opts="+cEeuaO:";
    const char *op=args->a0();
    ArgV	 *get_args=new ArgV(op);
-   int n_conn=0;
+   int n_conn=1;
    bool del=false;
    bool del_target=false;
    bool ascii=false;
    bool glob=false;
    bool make_dirs=false;
    bool reverse=false;
-   const char *output_dir=0;
+   xstring_c output_dir;
 
    if(!strncmp(op,"re",2))
    {
@@ -1712,7 +1681,7 @@ CMD(get)
    if(!strcmp(op,"pget"))
    {
       opts="+n:ceuO:";
-      n_conn=-1;
+      n_conn=0; // default, which means to take pget:default-n
    }
    else if(!strcmp(op,"put") || !strcmp(op,"reput"))
    {
@@ -1731,9 +1700,9 @@ CMD(get)
    }
    if(!reverse)
    {
-      output_dir=ResMgr::Query("xfer:destination-directory",session->GetHostName());
-      if(!*output_dir)
-	 output_dir=0;
+      const char *od=ResMgr::Query("xfer:destination-directory",session->GetHostName());
+      if(od && *od)
+	 output_dir.set(od);
    }
    while((opt=args->getopt(opts))!=EOF)
    {
@@ -1763,7 +1732,7 @@ CMD(get)
 	 make_dirs=true;
 	 break;
       case('O'):
-	 output_dir=optarg;
+	 output_dir.set(optarg);
 	 break;
       case('?'):
       err:
@@ -1782,21 +1751,18 @@ CMD(get)
 	 goto err;
       }
       delete get_args;
-      char *output_dir_dup=xstrdup(output_dir);  // save it from deleting.
-      output_dir=0;
       // remove options
       while(args->getindex()>1)
 	 args->delarg(1);
-      mgetJob *j=new mgetJob(session->Clone(),args,cont,make_dirs);
+      mgetJob *j=new mgetJob(session->Clone(),args.borrow(),cont,make_dirs);
       if(reverse)
 	 j->Reverse();
       if(del)
 	 j->DeleteFiles();
       if(ascii)
 	 j->Ascii();
-      if(output_dir_dup)
-	 j->OutputDir(output_dir_dup);
-      args=0;
+      if(output_dir)
+	 j->OutputDir(output_dir.borrow());
       return j;
    }
    args->back();
@@ -1820,28 +1786,18 @@ CMD(get)
       get_args->Append(dst);
    }
 
-   if(n_conn==0)
-   {
-      GetJob *j=new GetJob(session->Clone(),get_args,cont);
-      if(del)
-	 j->DeleteFiles();
-      if(del_target)
-	 j->RemoveTargetFirst();
-      if(ascii)
-	 j->Ascii();
-      if(reverse)
-	 j->Reverse();
-      return j;
-   }
-   else
-   {
-      pgetJob *j=new pgetJob(session->Clone(),get_args,cont);
-      if(n_conn!=-1)
-	 j->SetMaxConn(n_conn);
-      if(del_target)
-	 j->RemoveTargetFirst();
-      return j;
-   }
+   GetJob *j=new GetJob(session->Clone(),get_args,cont);
+   if(del)
+      j->DeleteFiles();
+   if(del_target)
+      j->RemoveTargetFirst();
+   if(ascii)
+      j->Ascii();
+   if(reverse)
+      j->Reverse();
+   if(n_conn!=1)
+      j->SetCopyJobCreator(new pCopyJobCreator(n_conn));
+   return j;
 }
 
 CMD(shell)
@@ -1851,9 +1807,8 @@ CMD(shell)
       j=new SysCmdJob(0);
    else
    {
-      char *a=args->Combine(1);
+      xstring_ca a(args->Combine(1));
       j=new SysCmdJob(a);
-      xfree(a);
    }
    return j;
 }
@@ -1898,15 +1853,13 @@ CMD(rm)
    if(args->getcurr()==0)
       goto print_usage;
 
-   rmJob *j=new rmJob(session->Clone(),args);
+   rmJob *j=new rmJob(session->Clone(),args.borrow());
 
    if(recursive)
       j->Recurse();
    if(rmdir)
       j->Rmdir();
 
-   args=0;
-
    if(silent)
       j->BeQuiet();
 
@@ -1914,9 +1867,7 @@ CMD(rm)
 }
 CMD(mkdir)
 {
-   Job *j=new mkdirJob(session->Clone(),args);
-   args=0;
-   return j;
+   return new mkdirJob(session->Clone(),args.borrow());
 }
 
 #ifndef O_ASCII
@@ -1946,9 +1897,8 @@ CMD(source)
    FDStream *f=0;
    if(e)
    {
-      char *cmd=args->Combine(args->getindex());
+      xstring_ca cmd(args->Combine(args->getindex()));
       f=new InputFilter(cmd);
-      xfree(cmd);
    }
    else
       f=new FileStream(args->getarg(1),O_RDONLY|O_ASCII);
@@ -1958,7 +1908,7 @@ CMD(source)
    {
       if(f->error())
       {
-	 fprintf(stderr,"%s: %s\n",args->a0(),f->error_text);
+	 fprintf(stderr,"%s: %s\n",args->a0(),f->error_text.get());
 	 delete f;
 	 return 0;
       }
@@ -2017,9 +1967,8 @@ CMD(pwd)
    int len=strlen(url_c);
    url[len++]='\n';  // replaces \0
 
-   OutputJob *out=new OutputJob(output, args->a0());
+   OutputJob *out=new OutputJob(output.borrow(), args->a0());
    Job *j=new echoJob(url,len,out);
-   output=0;
 
    return j;
 }
@@ -2158,7 +2107,7 @@ CMD(user)
       }
       else
       {
-	 eprintf("%s: %s%s\n",args->a0(),u.proto,
+	 eprintf("%s: %s%s\n",args->a0(),u.proto.get(),
 		  _(" - not supported protocol"));
 	 return 0;
       }
@@ -2274,11 +2223,9 @@ CMD(set)
 
    if(a==0)
    {
-      char *s=ResMgr::Format(with_defaults,only_defaults);
-      OutputJob *out=new OutputJob(output, args->a0());
+      xstring_ca s(ResMgr::Format(with_defaults,only_defaults));
+      OutputJob *out=new OutputJob(output.borrow(), args->a0());
       Job *j=new echoJob(s,out);
-      xfree(s);
-      output=0;
       return j;
    }
 
@@ -2300,17 +2247,14 @@ CMD(set)
    }
 
    args->getnext();
-   char *val=(args->getcurr()==0?0:args->Combine(args->getindex()));
+   xstring_ca val(args->getcurr()==0?0:args->Combine(args->getindex()));
    msg=ResMgr::Set(a,closure,val);
 
    if(msg)
    {
-      eprintf("%s: %s.\n",val,msg);
-      xfree(val);
+      eprintf("%s: %s.\n",val.get(),msg);
       return 0;
    }
-   xfree(val);
-
    exit_code=0;
    return 0;
 }
@@ -2319,11 +2263,9 @@ CMD(alias)
 {
    if(args->count()<2)
    {
-      char *list=Alias::Format();
-      OutputJob *out=new OutputJob(output, args->a0());
+      xstring_ca list(Alias::Format());
+      OutputJob *out=new OutputJob(output.borrow(), args->a0());
       Job *j=new echoJob(list,out);
-      xfree(list);
-      output=0;
       return j;
    }
    else if(args->count()==2)
@@ -2332,9 +2274,8 @@ CMD(alias)
    }
    else
    {
-      char *val=args->Combine(2);
+      xstring_ca val(args->Combine(2));
       Alias::Add(args->getarg(1),val);
-      xfree(val);
    }
    exit_code=0;
    return 0;
@@ -2355,8 +2296,7 @@ CMD(wait)
       if(!strcasecmp(jn,"all"))
       {
 	 parent->WaitForAllChildren();
-	 for(int i=0; i<parent->waiting_num; i++)
-	    parent->waiting[i]->Fg();
+	 parent->AllWaitingFg();
 	 exit_code=0;
 	 return 0;
       }
@@ -2405,8 +2345,7 @@ CMD(subsh)
    const char *c=args->getarg(1);
    e->FeedCmd(c);
    e->FeedCmd("\n");
-   e->cmdline=(char*)xmalloc(strlen(c)+3);
-   sprintf(e->cmdline,"(%s)",c);
+   e->cmdline.vset("(",c,")",NULL);
    return e;
 }
 
@@ -2560,8 +2499,9 @@ void CmdExec::print_cmd_index()
 {
    int i=0;
    const char *c1;
-   const cmd_rec *cmd_table=dyn_cmd_table?dyn_cmd_table:static_cmd_table;
-   while(cmd_table[i].name)
+   const cmd_rec *cmd_table=dyn_cmd_table?dyn_cmd_table.get():static_cmd_table;
+   const int count=dyn_cmd_table?dyn_cmd_table.count():1024;
+   while(i<count && cmd_table[i].name)
    {
       while(cmd_table[i].name && !cmd_table[i].short_desc)
 	 i++;
@@ -2604,7 +2544,7 @@ CMD(help)
 CMD(ver)
 {
    printf(
-      _("LFTP | Version %s | Copyright (c) 1996-%d Alexander V. Lukyanov\n"),VERSION,2006);
+      _("LFTP | Version %s | Copyright (c) 1996-%d Alexander V. Lukyanov\n"),VERSION,2008);
    printf("\n");
    printf(
       _("LFTP is free software, covered by the GNU General Public License, and you are\n"
@@ -2751,11 +2691,9 @@ CMD(bookmark)
 
    if(!strcasecmp(op,"list") || !strcasecmp(op,"list-p"))
    {
-      char *list=op[4]?lftp_bookmarks.Format():lftp_bookmarks.FormatHidePasswords();
-      OutputJob *out=new OutputJob(output, args->a0());
+      xstring_ca list(op[4]?lftp_bookmarks.Format():lftp_bookmarks.FormatHidePasswords());
+      OutputJob *out=new OutputJob(output.borrow(), args->a0());
       Job *j=new echoJob(list,out);
-      xfree(list);
-      output=0;
       return j;
    }
    else if(!strcasecmp(op,"add"))
@@ -2837,28 +2775,24 @@ CMD(bookmark)
 
 CMD(echo)
 {
-   char *s=args->Combine(1);
-   int len=strlen(s);
+   xstring s;
+   s.set_allocated(args->Combine(1));
    if(args->count()>1 && !strcmp(args->getarg(1),"-n"))
    {
-      if(len<=3)
+      if(s.length()<=3)
       {
 	 exit_code=0;
-	 xfree(s);
 	 return 0;
       }
-      memmove(s,s+3,len-=3);
+      s.set_substr(0,3,"");
    }
    else
    {
-      s[len++]='\n'; // replaces \0 char
+      s.append('\n');
    }
 
-   OutputJob *out=new OutputJob(output, args->a0());
-   Job *j=new echoJob(s,len,out);
-
-   xfree(s);
-   output=0;
+   OutputJob *out=new OutputJob(output.borrow(), args->a0());
+   Job *j=new echoJob(s,s.length(),out);
    return j;
 }
 
@@ -2875,11 +2809,11 @@ CMD(find)
       {"maxdepth",required_argument,0,'d'},
       {0,0,0,0}
    };
-   int opt, longopt;
+   int opt;
    int maxdepth = -1;
    const char *op=args->a0();
 
-   while((opt=args->getopt_long("+d:",find_options,&longopt))!=EOF)
+   while((opt=args->getopt_long("+d:",find_options))!=EOF)
    {
       switch(opt)
       {
@@ -2899,10 +2833,8 @@ CMD(find)
 
    if(!args->getcurr())
       args->Append(".");
-   FinderJob_List *j=new class FinderJob_List(session->Clone(),args,output);
+   FinderJob_List *j=new class FinderJob_List(session->Clone(),args.borrow(),output.borrow());
    j->set_maxdepth(maxdepth);
-   args=0;
-   output=0;
    return j;
 }
 
@@ -2947,8 +2879,8 @@ CMD(du)
 
    const char *op=args->a0();
 
-   int opt, longopt;
-   while((opt=args->getopt_long("+abcd:FhHkmsS",du_options,&longopt))!=EOF)
+   int opt;
+   while((opt=args->getopt_long("+abcd:FhHkmsS",du_options))!=EOF)
    {
       switch(opt)
       {
@@ -3036,8 +2968,7 @@ CMD(du)
 
    if(!args->getcurr())
       args->Append(".");
-   FinderJob_Du *j=new class FinderJob_Du(session->Clone(),args,output);
-   args=0;
+   FinderJob_Du *j=new class FinderJob_Du(session->Clone(),args.borrow(),output.borrow());
    j->PrintDepth(maxdepth);
    j->SetBlockSize(blocksize,human_opts);
    if(print_totals)
@@ -3058,8 +2989,6 @@ CMD(du)
       p->Add(p->EXCLUDE,new PatternSet::Glob(exclude));
       j->SetExclude(p);
    }
-
-   output=0;
    return j;
 }
 
@@ -3103,8 +3032,7 @@ CMD(lpwd)
    const char *name=parent->cwd->GetName();
    char *buf=alloca_strdup2(name,2);
    sprintf(buf,"%s\n",name?name:"?");
-   Job *j=new echoJob(buf,new OutputJob(output, args->a0()));
-   output=0;
+   Job *j=new echoJob(buf,new OutputJob(output.borrow(), args->a0()));
    return j;
 }
 
@@ -3130,7 +3058,7 @@ CMD(chmod)
    int opt;
    int modeind = 0;
 
-   while((opt=args->getopt_long("vcRfrwxXstugoa,+-=",chmod_options,0))!=EOF)
+   while((opt=args->getopt_long("vcRfrwxXstugoa,+-=",chmod_options))!=EOF)
    {
       switch(opt)
       {
@@ -3182,14 +3110,13 @@ CMD(chmod)
       return 0;
    }
 
-   ChmodJob *j=new ChmodJob(session->Clone(),args);
+   ChmodJob *j=new ChmodJob(session->Clone(),args.borrow());
    j->SetVerbosity(verbose);
    j->SetMode(m);
    if(quiet)
       j->BeQuiet(); /* does not affect messages from Verbosity */
    if(recurse)
       j->Recurse();
-   args=0;
    return j;
 }
 
@@ -3220,7 +3147,7 @@ CMD(get1)
    long long target_region_begin=0,target_region_end=FILE_END;
    int n,p;
 
-   while((opt=args->getopt_long("arco:",get1_options,0))!=EOF)
+   while((opt=args->getopt_long("arco:",get1_options))!=EOF)
    {
       switch(opt)
       {
@@ -3313,7 +3240,7 @@ CMD(get1)
    FileCopyPeer *src_peer=0;
    FileCopyPeer *dst_peer=0;
 
-   src_peer=FileCopyPeerFA::New(session->Clone(),src,FA::RETRIEVE,true);
+   src_peer=FileCopyPeerFA::New(session->Clone(),src,FA::RETRIEVE);
    if(!cont && (source_region_begin>0 || source_region_end!=FILE_END))
       src_peer->SetRange(source_region_begin,source_region_end);
 
@@ -3344,10 +3271,8 @@ CMD(slot)
    }
    else
    {
-      char *slots=ConnectionSlot::Format();
-      Job *j=new echoJob(slots,new OutputJob(output,args->a0()));
-      xfree(slots);
-      output=0;
+      xstring_ca slots(ConnectionSlot::Format());
+      Job *j=new echoJob(slots,new OutputJob(output.borrow(),args->a0()));
       return j;
    }
 }
@@ -3359,3 +3284,64 @@ CMD(tasks)
    exit_code=0;
    return 0;
 }
+
+CMD(eval)
+{
+   int opt;
+   const char *fmt=0;
+   const char *op=args->getarg(0);
+   while((opt=args->getopt("+f:"))!=EOF)
+   {
+      switch(opt)
+      {
+      case 'f':
+	 fmt=optarg;
+	 break;
+      default:
+	 eprintf(_("Try `%s --help' for more information\n"),op);
+	 return 0;
+      }
+   }
+   int base=optind;
+   xstring cmd;
+   if(!fmt)
+      cmd.set_allocated(args->Combine(optind));
+   else
+   {
+      while(*fmt)
+      {
+	 if(*fmt=='\\' && (fmt[1]=='$' || fmt[1]=='\\'))
+	 {
+	    cmd.append(fmt[1]);
+	    fmt+=2;
+	    continue;
+	 }
+	 if(*fmt=='$' && fmt[1]>='0' && fmt[1]<='9')
+	 {
+	    int n=fmt[1]-'0';
+	    if(n+base<args->count())
+	       cmd.append(args->getarg(n+base));
+	    fmt+=2;
+	    continue;
+	 }
+	 if(*fmt=='$' && fmt[1]=='@')
+	 {
+	    xstring_ca c(args->CombineQuoted(base));
+	    cmd.append(c);
+	    fmt+=2;
+	    continue;
+	 }
+	 if(*fmt=='$' && fmt[1]=='$')
+	 {
+	    cmd.appendf("%d",(int)getpid());
+	    fmt+=2;
+	    continue;
+	 }
+	 cmd.append(*fmt++);
+      }
+   }
+   cmd.append(";\n\n");
+   parent->PrependCmd(cmd);
+   exit_code=parent->prev_exit_code;
+   return 0;
+}
