@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2005 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2007 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: ftpclass.h,v 1.130 2006/02/06 11:00:41 lav Exp $ */
+/* $Id: ftpclass.h,v 1.155 2008/02/17 08:19:25 lav Exp $ */
 
 #ifndef FTPCLASS_H
 #define FTPCLASS_H
@@ -32,11 +32,21 @@
 # include "lftp_ssl.h"
 #endif
 
+class TelnetEncode : public DataTranslator {
+   void PutTranslated(Buffer *target,const char *buf,int size);
+};
+class TelnetDecode : public DataTranslator {
+   void PutTranslated(Buffer *target,const char *buf,int size);
+};
 class IOBufferTelnet : public IOBufferStacked
 {
-   void PutTranslated(const char *,int);
 public:
-   IOBufferTelnet(IOBuffer *b) : IOBufferStacked(b) {}
+   IOBufferTelnet(IOBuffer *b) : IOBufferStacked(b) {
+      if(mode==PUT)
+	 SetTranslator(new TelnetEncode());
+      else
+	 SetTranslator(new TelnetDecode());
+   }
 };
 
 class Ftp : public NetAccess
@@ -53,20 +63,22 @@ class Ftp : public NetAccess
       DATA_OPEN_STATE,	   // data connection is open, for read or write
       CWD_CWD_WAITING_STATE,  // waiting until 'CWD $cwd' finishes
       USER_RESP_WAITING_STATE,// waiting until 'USER $user' finishes
-      DATASOCKET_CONNECTING_STATE  // waiting for data_sock to connect
+      DATASOCKET_CONNECTING_STATE, // waiting for data_sock to connect
+      WAITING_150_STATE,   // waiting for 150 message
+      WAITING_CCC_SHUTDOWN // waiting for the server to shutdown SSL connection
    };
 
-
    class Connection
    {
+      xstring_c closure;
    public:
       int control_sock;
-      IOBuffer *control_recv;
-      IOBuffer *control_send;
+      SMTaskRef<IOBuffer> control_recv;
+      SMTaskRef<IOBuffer> control_send;
       IOBufferTelnet *telnet_layer_send;
-      Buffer *send_cmd_buffer; // holds unsent commands.
+      DirectedBuffer send_cmd_buffer; // holds unsent commands.
       int data_sock;
-      IOBuffer *data_iobuf;
+      SMTaskRef<IOBuffer> data_iobuf;
       int aborted_data_sock;
       sockaddr_u peer_sa;
       sockaddr_u data_sa; // address for data accepting
@@ -82,6 +94,7 @@ class Ftp : public NetAccess
       bool try_feat_after_login;
       bool tune_after_login;
       bool utf8_activated; // server is switched to UTF8 mode.
+      bool received_150;
 
       char type;  // type of transfer: 'A'scii or 'I'mage
 
@@ -104,27 +117,36 @@ class Ftp : public NetAccess
       off_t last_rest;	// last successful REST position.
       off_t rest_pos;	// the number sent with REST command.
 
+      Timer abor_close_timer;	 // timer for closing aborted connection.
+      Timer stat_timer;		 // timer for sending periodic STAT commands.
+      Timer waiting_150_timer;   // limit time to wait for 150 reply.
+      Timer waiting_ssl_timer;	 // limit time to wait for ssl shutdown
+
+      time_t nop_time;
+      off_t  nop_offset;
+      int    nop_count;
+
 #if USE_SSL
-      lftp_ssl *control_ssl;
+      Ref<lftp_ssl> control_ssl;
       char prot;  // current data protection scheme 'C'lear or 'P'rivate
       bool auth_sent;
       bool auth_supported;
       bool cpsv_supported;
       bool sscn_supported;
       bool sscn_on;
-      char *auth_args_supported;
+      xstring auth_args_supported;
       bool ssl_is_activated() { return control_ssl!=0; }
+      Timer waiting_ssl_shutdown;
 #else
       bool ssl_is_activated() { return false; }
 #endif
 
-      char *mlst_attr_supported;
+      xstring_c mlst_attr_supported;
 
-      Connection();
+      Connection(const char *c);
       ~Connection();
 
-      bool data_address_ok(sockaddr_u *d,bool verify_address,bool verify_port);
-      void SavePeerAddress();
+      bool data_address_ok(const sockaddr_u *d,bool verify_address,bool verify_port);
 
       void MakeBuffers();
       void MakeSSLBuffers(const char *h);
@@ -136,15 +158,17 @@ class Ftp : public NetAccess
       void AbortDataConnection();
       void CloseAbortedDataConnection();
 
-      void Send(const char *cmd,int len);
+      void Send(const char *cmd);
+      void SendEncoded(const char *url);
       void SendCmd(const char *cmd);
-      void SendCmd2(const char *cmd,const char *f);
+      void SendCmd2(const char *cmd,const char *f,const char *u=0,const char *home=0);
       void SendCmd2(const char *cmd,int v);
       void SendCmdF(const char *fmt,...) PRINTF_LIKE(2,3);
       int FlushSendQueueOneCmd(); // sends single command from send_cmd_buffer
    };
 
-   Connection *conn;
+   Ref<Connection> conn;
+   bool last_connection_failed;
 
    struct Expect
    {
@@ -174,6 +198,8 @@ class Ftp : public NetAccess
 	 USER_PROXY,	// check response for USER sent to proxy
 	 PASS,		// check response for PASS
 	 PASS_PROXY,	// check response for PASS sent to proxy
+	 OPEN_PROXY,	// special UserGate proxy command
+	 ACCT_PROXY,	// special ACCT with proxy password
 	 TRANSFER,	// generic check for transfer
 	 TRANSFER_CLOSED, // check for transfer complete when Close()d.
 	 FEAT,
@@ -187,24 +213,14 @@ class Ftp : public NetAccess
       };
 
       expect_t check_case;
-      char *arg;
+      xstring_c arg;
       Expect *next;
 
-      Expect(expect_t e,const char *a=0)
+      Expect(expect_t e,const char *a=0) : check_case(e), arg(a) {}
+      Expect(expect_t e,char c) : check_case(e)
 	 {
-	    check_case=e;
-	    arg=xstrdup(a);
-	 }
-      Expect(expect_t e,char c)
-	 {
-	    check_case=e;
-	    arg=(char*)xmalloc(2);
-	    arg[0]=c;
-	    arg[1]=0;
-	 }
-      ~Expect()
-	 {
-	    xfree(arg);
+	    char cc[2]={c,0};
+	    arg.set(cc);
 	 }
    };
    class ExpectQueue;
@@ -230,7 +246,7 @@ class Ftp : public NetAccess
       void Close();
    };
 
-   ExpectQueue *expect;
+   Ref<ExpectQueue> expect;
 
    void  CheckResp(int resp);
    int	 ReplyLogPriority(int code);
@@ -249,6 +265,7 @@ class Ftp : public NetAccess
    void	 CatchDATE_opt(int);
    void	 CatchSIZE(int);
    void	 CatchSIZE_opt(int);
+   void	 TurnOffStatForList();
 
    enum pasv_state_t
    {
@@ -277,31 +294,18 @@ private:
    static const bool ftps; // for convenience
 #endif
 
-   char  *line;
-   int	 line_len;
-   char  *all_lines;
+   void	DataAbort();
+   void DataClose();
+   void ControlClose();
 
-   bool	 eof;
-
-   time_t   nop_time;
-   off_t    nop_offset;
-   int	    nop_count;
-
-   time_t   stat_time;
-   time_t   retry_time;
-
-   void	 DataAbort();
-   void  DataClose();
-   void  ControlClose();
-
-   void  SendUrgentCmd(const char *cmd);
-   int	 FlushSendQueueOneCmd();
-   int	 FlushSendQueue(bool all=false);
-   void	 SendArrayInfoRequests();
-   void	 SendSiteIdle();
-   void	 SendAcct();
-   void	 SendSiteGroup();
-   void	 SendUTimeRequest();
+   void SendUrgentCmd(const char *cmd);
+   int	FlushSendQueueOneCmd();
+   int	FlushSendQueue(bool all=false);
+   void	SendArrayInfoRequests();
+   void	SendSiteIdle();
+   void	SendAcct();
+   void	SendSiteGroup();
+   void	SendUTimeRequest();
    void SendAuth(const char *auth);
    void TuneConnectionAfterFEAT();
    void SendOPTS_MLST();
@@ -309,6 +313,7 @@ private:
 
    const char *QueryStringWithUserAtHost(const char *);
 
+   int	 ReceiveOneLine();
    // If a response is received, it checks it for accordance with
    // response_queue and switch to a state if necessary
    int	 ReceiveResp();
@@ -325,46 +330,39 @@ private:
    void	 HttpProxySendConnect();
    void	 HttpProxySendConnectData();
    // Send http proxy auth.
-   void	 HttpProxySendAuth(IOBuffer *);
+   void	 HttpProxySendAuth(const SMTaskRef<IOBuffer>&);
    // Check if proxy returned a reply, returns true if reply is ok.
    // May disconnect.
-   bool	 HttpProxyReplyCheck(IOBuffer *buf);
+   bool	 HttpProxyReplyCheck(const SMTaskRef<IOBuffer>&);
 
    bool	 AbsolutePath(const char *p);
 
    void MoveConnectionHere(Ftp *o);
+   bool GetBetterConnection(int level,bool limit_reached);
+   bool SameConnection(const Ftp *o) const;
 
+   // state
    automate_state state;
+   int flags;
+   bool eof;
+   Timer retry_timer;
 
-   char	 *anon_user;
-   char	 *anon_pass;
-
-   char	 *charset;
-
-   static const char *DefaultAnonPass();
-
-   int	 flags;
+   xstring line;	// last line of last server reply
+   xstring all_lines;   // all lines of last server reply
 
    void	 SetError(int code,const char *mess=0);
 
-   bool  verify_data_address;
-   bool  verify_data_port;
-   bool	 rest_list;
-   char  *list_options;
+   // settings
+   xstring_c anon_user;
+   xstring_c anon_pass;
+   xstring_c charset;
+   xstring_c list_options;
+   int nop_interval;
+   bool verify_data_address;
+   bool verify_data_port;
+   bool	rest_list;
 
-   bool	 GetBetterConnection(int level,bool limit_reached);
-   bool  SameConnection(const Ftp *o);
-
-   int	 nop_interval;
-
-   void set_idle_start()
-      {
-	 idle_start=now;
-	 if(conn && idle>0)
-	    TimeoutS(idle);
-      }
-
-   char *skey_pass;
+   xstring_c skey_pass;
    bool allow_skey;
    bool force_skey;
    const char *make_skey_reply();
@@ -385,6 +383,8 @@ private:
    bool copy_allow_store;
    bool copy_failed;
 
+   bool use_stat;
+   bool use_stat_for_list;
    bool use_mdtm;
    bool use_size;
    bool use_pret;
@@ -393,10 +393,7 @@ private:
 
    bool use_telnet_iac;
 
-   bool use_stat;
-   int  stat_interval;
-
-   const char *encode_eprt(sockaddr_u *);
+   const char *encode_eprt(const sockaddr_u *);
 
    typedef FileInfo *(*FtpLineParser)(char *line,int *err,const char *tz);
    static FtpLineParser line_parsers[];
@@ -410,15 +407,15 @@ public:
    Ftp();
    Ftp(const Ftp *);
 
-   const char *GetProto() { return "ftp"; }
+   const char *GetProto() const { return "ftp"; }
 
-   FileAccess *Clone() { return new Ftp(this); }
+   FileAccess *Clone() const { return new Ftp(this); }
    static FileAccess *New();
 
    const char *ProtocolSubstitution(const char *host);
 
-   bool	 SameLocationAs(FileAccess *);
-   bool	 SameSiteAs(FileAccess *);
+   bool	 SameLocationAs(const FileAccess *) const;
+   bool	 SameSiteAs(const FileAccess *) const;
 
    void	 ResetLocationData();
 
@@ -431,8 +428,8 @@ public:
       CL_LOGGED_IN,
       CL_JUST_BEFORE_DISCONNECT
    };
-   ConnectLevel GetConnectLevel();
-   int IsConnected()
+   ConnectLevel GetConnectLevel() const;
+   int IsConnected() const
    {
       return GetConnectLevel()!=CL_NOT_CONNECTED;
    }
@@ -478,7 +475,7 @@ public:
    ListInfo *MakeListInfo(const char *path);
    Glob *MakeGlob(const char *pattern);
    DirList *MakeDirList(ArgV *args);
-   FileSet *ParseLongList(const char *buf,int len,int *err=0);
+   FileSet *ParseLongList(const char *buf,int len,int *err=0) const;
 
    void SetCopyMode(copy_mode_t cm,bool rp,bool prot,bool sscn,int rnum,time_t tt)
       {
@@ -489,7 +486,7 @@ public:
 	 retries=rnum;
 	 try_time=tt;
       }
-   bool SetCopyAddress(Ftp *o)
+   bool SetCopyAddress(const Ftp *o)
       {
 	 if(copy_addr_valid || !o->copy_addr_valid)
 	    return false;
@@ -516,11 +513,12 @@ public:
 	    return copy_addr_valid && expect->FirstIs(Expect::TRANSFER);
 	 return state==WAITING_STATE && expect->IsEmpty();
       }
-   void CopyCheckTimeout(Ftp *o)
+   void CopyCheckTimeout(const Ftp *o)
       {
-	 BumpEventTime(o->event_time);
+	 timeout_timer.Reset(o->timeout_timer);
 	 CheckTimeout();
       }
+   bool AnonymousQuietMode();
 };
 
 class FtpS : public Ftp
@@ -530,9 +528,9 @@ public:
    FtpS(const FtpS *);
    ~FtpS();
 
-   const char *GetProto() { return "ftps"; }
+   const char *GetProto() const { return "ftps"; }
 
-   FileAccess *Clone() { return new FtpS(this); }
+   FileAccess *Clone() const { return new FtpS(this); }
    static FileAccess *New();
 };
 
