@@ -18,12 +18,12 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: keyvalue.h,v 1.9 2002/09/26 06:50:49 lav Exp $ */
+/* $Id: keyvalue.h,v 1.11 2007/08/02 11:35:32 lav Exp $ */
 
 #ifndef KEYVALUE_H
 #define KEYVALUE_H
 
-#include "xmalloc.h"
+#include "xstring.h"
 
 class StringMangler
 {
@@ -40,29 +40,17 @@ public:
    class Pair
    {
    public:
-      char *key;
-      char *value;
+      xstring_c key;
+      xstring_c value;
       Pair *next;
       Pair(const char *k,const char *v)
-	 {
-	    key=xstrdup(k);
-	    value=xstrdup(v);
-	    next=0;
-	 }
-      virtual ~Pair()
-	 {
-	    xfree(key);
-	    xfree(value);
-	 }
+	 : key(k), value(v), next(0) {}
+      virtual ~Pair() {}
       int KeyCompare(const char *s) const
 	 {
 	    return strcmp(s,key);
 	 }
-      void SetValue(const char *v)
-	 {
-	    xfree(value);
-	    value=xstrdup(v);
-	 }
+      void SetValue(const char *v) { value.set(v); }
    };
 
 protected:
