@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: lftp.cc,v 1.71 2006/04/19 09:48:07 lav Exp $ */
+/* $Id: lftp.cc,v 1.79 2007/08/16 15:39:52 lav Exp $ */
 
 #include <config.h>
 
@@ -45,7 +45,6 @@
 #include "DummyProto.h"
 #include "ResMgr.h"
 #include "LsCache.h"
-#include "DirColors.h"
 #include "IdNameCache.h"
 #include "LocalDir.h"
 #include "ConnectionSlot.h"
@@ -78,9 +77,9 @@ class ReadlineFeeder : public CmdFeeder
    bool tty:1;
    bool ctty:1;
    bool add_newline:1;
-   char *to_free;
    int eof_count;
-   char *for_history;
+   xstring_c cmd_buf;
+   xstring_c for_history;
 
    static bool readline_inited;
    void readline_init()
@@ -93,8 +92,7 @@ class ReadlineFeeder : public CmdFeeder
       if(for_history)
       {
 	 lftp_add_history_nodups(for_history);
-	 xfree(for_history);
-	 for_history=0;
+	 for_history.set(0);
       }
    }
 
@@ -104,11 +102,9 @@ public:
       tty=isatty(0);
       ctty=(tcgetpgrp(0)!=(pid_t)-1);
       add_newline=false;
-      to_free=0;
       eof_count=0;
-      for_history=0;
       if(args && args->count()>1)
-	 for_history=args->CombineQuoted();
+	 for_history.set_allocated(args->CombineQuoted());
    }
    virtual ~ReadlineFeeder()
    {
@@ -119,8 +115,6 @@ public:
 	 if(res_save_rl_history.QueryBool(0))
 	    lftp_rl_write_history();
       }
-      xfree(to_free);
-      xfree(for_history);
    }
    bool RealEOF()
    {
@@ -129,9 +123,6 @@ public:
 
    const char *NextCmd(class CmdExec *exec,const char *prompt)
    {
-      xfree(to_free);
-      to_free=0;
-
       if(add_newline)
       {
 	 add_newline=false;
@@ -141,7 +132,6 @@ public:
       ::completion_shell=exec;
       ::remote_completion=exec->remote_completion;
 
-      char *cmd_buf;
       if(tty)
       {
 	 readline_init();
@@ -158,36 +148,37 @@ public:
 	 }
 
 	 SignalHook::ResetCount(SIGINT);
-	 cmd_buf=lftp_readline(prompt);
+	 cmd_buf.set_allocated(lftp_readline(prompt));
+	 xmalloc_register_block(cmd_buf.get_non_const());
 	 if(cmd_buf && *cmd_buf)
 	 {
 	    if(exec->csh_history)
 	    {
-	       char *history_value;
-	       int expanded = lftp_history_expand (cmd_buf, &history_value);
-
+	       char *history_value0=0;
+	       int expanded = lftp_history_expand (cmd_buf, &history_value0);
 	       if (expanded)
 	       {
+		  if(history_value0)
+		     xmalloc_register_block(history_value0);
+		  xstring_ca history_value(history_value0);
+
 		  if (expanded < 0);
-		     fprintf (stderr, "%s\n", history_value);
+		     fprintf (stderr, "%s\n", history_value.get());
 
 		  /* If there was an error, return nothing. */
 		  if (expanded < 0 || expanded == 2)	/* 2 == print only */
 		  {
 		     exec->Timeout(0);  // and retry immediately
-		     xfree(history_value);
 		     return "";
 		  }
 
-		  to_free=history_value;
-		  cmd_buf=history_value;
+		  cmd_buf.set_allocated(history_value.borrow());
 	       }
 	    }
 	    lftp_add_history_nodups(cmd_buf);
 	 }
 	 else if(cmd_buf==0 && exec->interactive)
 	    puts("exit");
-	 xmalloc_register_block(cmd_buf);
 
 	 if(cmd_buf==0)
 	    eof_count++;
@@ -206,9 +197,8 @@ public:
 	    }
 	    fflush(stdout);
 	 }
-	 cmd_buf=readline_from_file(stdin);
+	 cmd_buf.set_allocated(readline_from_file(0));
       }
-      to_free=cmd_buf;
 
       ::completion_shell=0;
 
@@ -326,7 +316,7 @@ static void move_to_background()
    // notify jobs
    Job::lftpMovesToBackground_ToAll();
    // wait they do something, but no more than 1 sec.
-   SMTask::RollAll(1);
+   SMTask::RollAll(TimeInterval(1,0));
    // if all jobs terminated, don't really move to bg.
    if(Job::NumberOfJobs()==0)
       return;
@@ -410,13 +400,13 @@ int lftp_slot(int count,int key)
    return 0;
 }
 
-void  source_if_exist(CmdExec *exec,const char *rc)
+void  source_if_exist(const char *rc)
 {
    if(access(rc,R_OK)!=-1)
    {
-      exec->FeedCmd("source ");
-      exec->FeedCmd(rc);
-      exec->FeedCmd("\n");
+      top_exec->FeedCmd("source ");
+      top_exec->FeedCmd(rc);
+      top_exec->FeedCmd("\n");
    }
 }
 
@@ -426,10 +416,14 @@ static void tty_clear()
       top_exec->pre_stdout();
 }
 
+char *program_name;
+
 int   main(int argc,char **argv)
 {
+   program_name=argv[0];
+
 #ifdef SOCKS4
-   SOCKSinit(argv[0]);
+   SOCKSinit(program_name);
 #endif
 
    setlocale (LC_ALL, "");
@@ -451,40 +445,25 @@ int   main(int argc,char **argv)
    top_exec->SetStatusLine(new StatusLine(1));
    Log::global->SetCB(tty_clear);
 
-   source_if_exist(top_exec,SYSCONFDIR"/lftp.conf");
-
+   source_if_exist(SYSCONFDIR"/lftp.conf");
    const char *home=getenv("HOME");
    if(home)
-   {
-      char *rc=(char*)alloca(strlen(home)+9+1);
-
-      sprintf(rc,"%s/.lftprc",home);
-      source_if_exist(top_exec,rc);
-   }
-
+      source_if_exist(dir_file(home,".lftprc"));
    home=get_lftp_home();
    if(home)
-   {
-      char *rc=(char*)alloca(strlen(home)+3+1);
-
-      sprintf(rc,"%s/rc",home);
-      source_if_exist(top_exec,rc);
-   }
-
+      source_if_exist(dir_file(home,"rc"));
    top_exec->WaitDone();
 
    top_exec->SetTopLevel();
    top_exec->Fg();
 
-   ArgV *args=new ArgV(argc,argv);
+   Ref<ArgV> args(new ArgV(argc,argv));
    args->setarg(0,"lftp");
 
    lftp_feeder=new ReadlineFeeder(args);
 
-   top_exec->ExecParsed(args);
-
+   top_exec->ExecParsed(args.borrow());
    top_exec->WaitDone();
-
    int exit_code=top_exec->ExitCode();
 
    top_exec->AtExit();
@@ -496,23 +475,17 @@ int   main(int argc,char **argv)
       move_to_background();
    }
    top_exec->KillAll();
-   SMTask::Delete(top_exec);
    top_exec=0;
+
    Job::Cleanup();
    ConnectionSlot::Cleanup();
    SessionPool::ClearAll();
-   LsCache::Flush();
+   FileAccess::ClassCleanup();
    ProcWait::DeleteAll();
-   DirColors::DeleteInstance();
    IdNameCacheCleanup();
    SignalHook::Cleanup();
    Log::Cleanup();
    SMTask::Cleanup();
 
-   // the tasks left: LsCache::ExpireHelper, lftp_ssl_instance(if created)
-   int task_count=SMTask::TaskCount();
-   if(task_count>2)
-      printf("WARNING: task_count=%d\n",task_count);
-
    return exit_code;
 }
