@@ -30,18 +30,19 @@
 #  include <openssl/rand.h>
 # endif
 
+#include "xstring.h"
+
 class lftp_ssl_base
 {
 public:
    bool handshake_done;
    int fd;
-   char *hostname;
+   xstring_c hostname;
    enum handshake_mode_t { CLIENT, SERVER } handshake_mode;
-   char *error;
+   xstring error;
    bool fatal;
 
    lftp_ssl_base(int fd,handshake_mode_t m,const char *host=0);
-   ~lftp_ssl_base();
 
    enum code { RETRY=-2, ERROR=-1, DONE=0 };
 
@@ -105,6 +106,7 @@ public:
    bool want_in();
    bool want_out();
    void copy_sid(const lftp_ssl_gnutls *);
+   void load_keys();
 };
 typedef lftp_ssl_gnutls lftp_ssl;
 #elif USE_OPENSSL
@@ -137,6 +139,7 @@ public:
    bool want_in();
    bool want_out();
    void copy_sid(const lftp_ssl_openssl *);
+   void load_keys();
 };
 typedef lftp_ssl_openssl lftp_ssl;
 #endif
