@@ -18,18 +18,17 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: log.cc,v 1.16 2004/09/06 09:19:57 lav Exp $ */
+/* $Id: log.cc,v 1.21 2007/08/24 16:22:12 lav Exp $ */
 
 #include <config.h>
 
 #include <stdarg.h>
 #include "trio.h"
-#include "xmalloc.h"
 #include "xstring.h"
 #include "log.h"
 #include "SMTask.h"
 
-Log *Log::global=new Log;
+SMTaskRef<Log> Log::global(new Log);
 
 void Log::Init()
 {
@@ -45,22 +44,29 @@ void Log::Init()
    at_line_start=true;
 }
 
-void Log::Write(int l,const char *s)
+bool Log::WillOutput(int l)
 {
-   if(!enabled || l>level)
-      return;
-   if(output==-1)
-      return;
-   if(!*s)
-      return;
+   if(!enabled || l>level || output==-1)
+      return false;
    if(tty)
    {
       pid_t pg=tcgetpgrp(output);
-      if(pg==(pid_t)-1)
-	 tty=false;
-      else if(pg!=getpgrp())
-	 return;
+      if(pg!=(pid_t)-1 && pg!=getpgrp())
+	 return false;
    }
+   return true;
+}
+
+void Log::Write(int l,const char *s)
+{
+   if(!WillOutput(l))
+      return;
+   DoWrite(s);
+}
+void Log::DoWrite(const char *s)
+{
+   if(!s || !*s)
+      return;
    if(at_line_start)
    {
       if(tty_cb && tty)
@@ -101,33 +107,26 @@ int Log::Do()
 
 void Log::Format(int l,const char *f,...)
 {
-   static char *buf=0;
-   static int buf_alloc;
+   if(!WillOutput(l))
+      return;
+
    va_list v;
+   va_start(v,f);
+   DoWrite(xstring::vformat(f,v));
+   va_end(v);
+}
 
-   if(buf==0)
-      buf=(char*)xmalloc(buf_alloc=1024);
+void Log::vFormat(int l,const char *f,va_list v)
+{
+   if(!WillOutput(l))
+      return;
 
-   for(;;)
-   {
-      va_start(v,f);
-      int res=vsnprintf(buf,buf_alloc,f,v);
-      va_end(v);
-      if(res>=0 && res<buf_alloc)
-	 break;
-      if(res==buf_alloc)
-	 res*=2;
-      if(res==-1)
-	 res=buf_alloc*2;
-      buf=(char*)xrealloc(buf,buf_alloc=res);
-   }
-
-   Write(l,buf);
+   DoWrite(xstring::vformat(f,v));
 }
 
 void Log::Cleanup()
 {
-   delete global;
+   global=0;
 }
 Log::~Log()
 {
