@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2007 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: mgetJob.h,v 1.8 2002/07/31 11:28:02 lav Exp $ */
+/* $Id: mgetJob.h,v 1.11 2007/04/13 12:18:33 lav Exp $ */
 
 #ifndef MGETJOB_H
 #define MGETJOB_H
@@ -33,22 +33,22 @@ class mgetJob : public GetJob
    ArgV *mkdir_args;
    int mkdir_base_arg;
    ArgV *m_args;
-   char *output_dir;
+   xstring_c output_dir;
 
    bool	 make_dirs;
-   void	 make_directory(char *d);
+   void	 make_directory(const char *d);
 
    void LocalGlob(const char *p);
 
 public:
    int	 Do();
    void	 PrintStatus(int,const char *);
-   void	 ShowRunStatus(StatusLine *s);
+   void	 ShowRunStatus(const SMTaskRef<StatusLine>&);
 
    mgetJob(FileAccess *session,ArgV *args,bool c,bool md);
    ~mgetJob();
 
-   void OutputDir(char *o) { output_dir=o; }
+   void OutputDir(char *o) { output_dir.set_allocated(o); }
 };
 
 #endif // MGETJOB_H
